# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# * The software is provided "as is" without any warranty, neither express
#   nor implied.
# * In no event will the author be held liable for any damage(s) arising
#   from the use of the software.
# * Redistribution of the software is allowed only in unmodified form.
# * Permission is granted to use the software solely for personal private
#   and non-commercial purposes.
# * An individuals use of the software in his or her capacity or function
#   as an agent, (independent) contractor, employee, member or officer of
#   a business, corporation or organization (commercial or non-commercial)
#   does not qualify as personal private and non-commercial purpose.
# * Without written approval from the author the software must not be used
#   for a business, for commercial, corporate, governmental, military or
#   organizational purposes of any kind, or in a commercial, corporate,
#   governmental, military or organizational environment of any kind.

.SUFFIXES:

all: I386-MSC.LIB

clean:
	ERASE I386-MSC.DOS I386-MSC.EXP I386-MSC.OBJ
	ERASE I386-CRT.OBJ I386-DLG.OBJ I386-DLH.OBJ I386-DLI.OBJ I386-FMT.OBJ I386-FPU.OBJ I386-GS.OBJ I386-I64.OBJ I386-MEM.OBJ I386-RTC.OBJ I386-STK.OBJ I386-STR.OBJ I386-TLS.OBJ I386-WCS.OBJ

clobber: clean
	ERASE I386-MSC.EXE I386-MSC.LIB

test: I386-MSC.DOS I386-MSC.EXE
	"$(MAKEDIR)\I386-MSC.EXE"

I386-MSC.IMP: I386-MSC.MAK
	LINK.EXE /LIB /BREPRO /DEF /EXPORT:_CIcos /EXPORT:_CIlog /EXPORT:_CIpow /EXPORT:_CIsin /EXPORT:_CIsqrt /EXPORT:_alldiv /EXPORT:_alldvrm /EXPORT:_allmul /EXPORT:_alloca_probe /EXPORT:_alloca_probe_8 /EXPORT:_alloca_probe_16 /EXPORT:_allrem /EXPORT:_allshl /EXPORT:_allshr /EXPORT:_aulldiv /EXPORT:_aulldvrm /EXPORT:_aullrem /EXPORT:_aullshr /EXPORT:_chkstk /EXPORT:_fltused /EXPORT:_ftol /EXPORT:memchr /EXPORT:memcmp /EXPORT:memcpy /EXPORT:memmove /EXPORT:memset /MACHINE:I386 /NAME:NTDLL /NODEFAULTLIB /NOLOGO /OUT:$@

I386-MSC.LIB: I386-CRT.OBJ I386-DLI.OBJ I386-FMT.OBJ I386-FPU.OBJ I386-GS.OBJ I386-I64.OBJ I386-MEM.OBJ I386-STK.OBJ I386-STR.OBJ I386-TLS.OBJ I386-WCS.OBJ # I386-RTC.OBJ
	LINK.EXE /LIB /BREPRO /MACHINE:I386 /NODEFAULTLIB /NOLOGO /OUT:$@ $**

I386-MSC.EXE: I386-MSC.OBJ # I386-MSC.LIB # I386-MSC.DOS
	LINK.EXE /LINK /BREPRO /DEFAULTLIB:KERNEL32.LIB /DEFAULTLIB:USER32.LIB /ENTRY:wmainCRTStartup /FIXED /LARGEADDRESSAWARE /MACHINE:I386 /MERGE:.rdata=.const /MERGE:.text=.code /NOCOFFGRPINFO /NOLOGO /NXCOMPAT /OPT:ICF,REF /OSVERSION:4.0 /OUT:$@ /RELEASE /STUB:I386-MSC.DOS /SUBSYSTEM:CONSOLE /SWAPRUN:CD,NET /TEST /VERSION:0.815 $** I386-MSC.LIB

I386-MSC.DOS: I386-MSC.MAK
	CERTUTIL.EXE /DecodeHex /F /V << $@
4D 5A 90 00 01 00 00 00 04 00 00 00 FF FF 00 00   MZ..............
E0 00 00 00 43 00 00 00 40 00 00 00 00 00 00 00   ....C...@.......
00 00 00 00 19 57 04 27 00 00 00 00 00 00 00 00   .....W.'........
00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00   ................
28 43 29 6F 70 79 72 69 67 68 74 20 32 30 30 34   (C)opyright 2004
2D 32 30 32 34 2C 20 53 74 65 66 61 6E 20 4B 61   -2024, Stefan Ka
6E 74 68 61 6B 20 3C 73 74 65 66 61 6E 2E 6B 61   nthak <stefan.ka
6E 74 68 61 6B 40 6E 65 78 67 6F 2E 64 65 3E 0D   nthak@nexgo.de>.
0A 07 24 0E 1F 33 D2 B4 09 CD 21 B8 01 4C CD 21   .....3....!..L.!
<<NOKEEP

I386-MSC.OBJ: I386-MSC.MAK
	CL.EXE /Brepro /c /DCYCLES /Fo$@ /GAFwy /nologo /O2txy /Tc<< /W4 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

#ifndef _M_IX86
#pragma message("For I386 platform only!")
#endif

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

#define STRICT
#define UNICODE
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

typedef	LONGLONG	SQWORD;
typedef	ULONGLONG	QWORD;

#define _(DIVIDEND, DIVISOR)	{(DIVIDEND), (DIVISOR), (DIVIDEND) / (DIVISOR), (DIVIDEND) % (DIVISOR)}

const	struct	_ull
{
	QWORD	ullDividend, ullDivisor, ullQuotient, ullRemainder;
} ullTable[] = {_(0x0000000000000000ULL, 0x0000000000000001ULL),
                _(0x0000000000000001ULL, 0x0000000000000001ULL),
                _(0x0000000000000002ULL, 0x0000000000000001ULL),
                _(0x0000000000000002ULL, 0x0000000000000002ULL),
                _(0x0000000000000000ULL, 0xFFFFFFFFFFFFFFFFULL),
                _(0x0000000000000001ULL, 0xFFFFFFFFFFFFFFFFULL),
                _(0x0000000000000001ULL, 0xFFFFFFFFFFFFFFFEULL),
                _(0x0000000000000002ULL, 0xFFFFFFFFFFFFFFFEULL),
                _(0x0000000000000003ULL, 0xFFFFFFFFFFFFFFFEULL),
                _(0x0000000000000003ULL, 0xFFFFFFFFFFFFFFFDULL),
                _(0x000000000FFFFFFFULL, 0x0000000000000001ULL),
                _(0x0000000FFFFFFFFFULL, 0x000000000000000FULL),
                _(0x0000000FFFFFFFFFULL, 0x0000000000000010ULL),
                _(0x0000000000000100ULL, 0x000000000FFFFFFFULL),
                _(0x00FFFFFFF0000000ULL, 0x0000000010000000ULL),
                _(0x07FFFFFF80000000ULL, 0x0000000080000000ULL),
                _(0x7FFFFFFEFFFFFFF0ULL, 0xFFFFFFFFFFFFFFFEULL),
                _(0x7FFFFFFEFFFFFFF0ULL, 0x0000FFFFFFFFFFFEULL),
                _(0x7FFFFFFEFFFFFFF0ULL, 0x7FFFFFFEFFFFFFF0ULL),
                _(0x7FFFFFFFFFFFFFFFULL, 0x8000000000000000ULL),
                _(0x7FFFFFFFFFFFFFFFULL, 0xFFFFFFFFFFFFFFFDULL),
                _(0x7FFFFFFFFFFFFFFFULL, 0xFFFFFFFFFFFFFFFEULL),
                _(0x7FFFFFFFFFFFFFFFULL, 0xFFFFFFFFFFFFFFFFULL),
                _(0x8000000000000000ULL, 0x0000000000000001ULL),
                _(0x8000000000000000ULL, 0x0000000000000002ULL),
                _(0x8000000000000000ULL, 0x0000000000000003ULL),
                _(0x8000000000000000ULL, 0x00000000FFFFFFFDULL),
                _(0x8000000000000000ULL, 0x00000000FFFFFFFEULL),
                _(0x8000000000000000ULL, 0x00000000FFFFFFFFULL),
                _(0x8000000000000000ULL, 0x0000000100000000ULL),
                _(0x8000000000000000ULL, 0x0000000100000001ULL),
                _(0x8000000000000000ULL, 0x0000000100000002ULL),
                _(0x8000000000000000ULL, 0x0000000100000003ULL),
                _(0x8000000000000000ULL, 0xFFFFFFFF00000000ULL),
                _(0x8000000000000000ULL, 0xFFFFFFFFFFFFFFFDULL),
                _(0x8000000000000000ULL, 0xFFFFFFFFFFFFFFFEULL),
                _(0x8000000000000000ULL, 0xFFFFFFFFFFFFFFFFULL),
                _(0x8000000080000000ULL, 0x0000000080000000ULL),
                _(0x8000000080000001ULL, 0x0000000080000001ULL),
                _(0xFFFFFFFEFFFFFFF0ULL, 0xFFFFFFFFFFFFFFFEULL),
                _(0xFFFFFFFFFFFFFFFCULL, 0x00000000FFFFFFFEULL),
                _(0xFFFFFFFFFFFFFFFCULL, 0x0000000100000002ULL),
                _(0xFFFFFFFFFFFFFFFEULL, 0x0000000080000000ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x0000000000000001ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x0000000000000002ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x0000000000000003ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x00000000FFFFFFFDULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x00000000FFFFFFFEULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x00000000FFFFFFFFULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x0000000100000001ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x0000000100000002ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x0000000100000003ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x00000001C0000001ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x0000000380000003ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x8000000000000000ULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0x7FFFFFFFFFFFFFFFULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0xFFFFFFFFFFFFFFFEULL),
                _(0xFFFFFFFFFFFFFFFFULL, 0xFFFFFFFFFFFFFFFFULL)};

const	struct	_ll
{
	SQWORD	llDividend, llDivisor, llQuotient, llRemainder;
} llTable[] = {_(0x0000000000000000LL, 0x0000000000000001LL),	// 0, 1
               _(0x0000000000000001LL, 0x0000000000000001LL),	// 1, 1
               _(0x0000000000000000LL, 0xFFFFFFFFFFFFFFFFLL),	// 0, -1
               _(0x0000000000000001LL, 0xFFFFFFFFFFFFFFFFLL),	// 1, -1
               _(0x0000000000000001LL, 0xFFFFFFFFFFFFFFFELL),	// 1, -2
               _(0x0000000000000002LL, 0xFFFFFFFFFFFFFFFELL),	// 2, -2
               _(0x000000000FFFFFFFLL, 0x0000000000000001LL),
               _(0x0000000FFFFFFFFFLL, 0x000000000000000FLL),
               _(0x0000000FFFFFFFFFLL, 0x0000000000000010LL),
               _(0x0000000000000100LL, 0x000000000FFFFFFFLL),
               _(0x00FFFFFFF0000000LL, 0x0000000010000000LL),
               _(0x07FFFFFF80000000LL, 0x0000000080000000LL),
               _(0x7FFFFFFEFFFFFFF0LL, 0xFFFFFFFFFFFFFFFELL),
               _(0x7FFFFFFEFFFFFFF0LL, 0x0000FFFFFFFFFFFELL),
               _(0x7FFFFFFEFFFFFFF0LL, 0x7FFFFFFEFFFFFFF0LL),
               _(0x7FFFFFFFFFFFFFFFLL, 0x8000000000000000LL),	// llmax, llmin
               _(0x7FFFFFFFFFFFFFFFLL, 0xFFFFFFFFFFFFFFFDLL),	// llmax, -3
               _(0x7FFFFFFFFFFFFFFFLL, 0xFFFFFFFFFFFFFFFELL),	// llmax, -2
               _(0x7FFFFFFFFFFFFFFFLL, 0xFFFFFFFFFFFFFFFFLL),	// llmax, -1
               _(0x8000000000000000LL, 0x0000000000000001LL),	// llmin, 1
               _(0x8000000000000000LL, 0x0000000000000002LL),	// llmin, 2
               _(0x8000000000000000LL, 0x0000000000000003LL),	// llmin, 3
               _(0x8000000000000000LL, 0x00000000FFFFFFFELL),
               _(0x8000000000000000LL, 0x00000000FFFFFFFFLL),
               _(0x8000000000000000LL, 0x0000000100000000LL),
               _(0x8000000000000000LL, 0x0000000100000001LL),
               _(0x8000000000000000LL, 0x0000000100000002LL),
               _(0x8000000000000000LL, 0x8000000000000000LL),	// llmin, llmin
               _(0x8000000000000000LL, 0xFFFFFFFF00000000LL),
               _(0x8000000000000000LL, 0xFFFFFFFFFFFFFFFDLL),	// llmin, -3
               _(0x8000000000000000LL, 0xFFFFFFFFFFFFFFFELL),	// llmin, -2
               _(0x8000000000000000LL, 0xFFFFFFFFFFFFFFFFLL),	// llmin, -1
               _(0x8000000080000000LL, 0x0000000080000000LL),
               _(0x8000000080000001LL, 0x0000000080000001LL),
               _(0xFFFFFFFEFFFFFFF0LL, 0xFFFFFFFFFFFFFFFELL),
               _(0xFFFFFFFFFFFFFFFELL, 0x0000000080000000LL),
               _(0xFFFFFFFFFFFFFFFELL, 0x0000000000000001LL),	// -2, 1
               _(0xFFFFFFFFFFFFFFFELL, 0x0000000000000002LL),	// -2, 2
               _(0xFFFFFFFFFFFFFFFELL, 0xFFFFFFFFFFFFFFFELL),	// -2, -2
               _(0xFFFFFFFFFFFFFFFELL, 0xFFFFFFFFFFFFFFFFLL),	// -2, -1
               _(0xFFFFFFFFFFFFFFFFLL, 0x0000000000000001LL),	// -1, 1
               _(0xFFFFFFFFFFFFFFFFLL, 0x0000000000000002LL),	// -1, 2
               _(0xFFFFFFFFFFFFFFFFLL, 0xFFFFFFFFFFFFFFFELL),	// -1, -2
               _(0xFFFFFFFFFFFFFFFFLL, 0xFFFFFFFFFFFFFFFFLL)};	// -1, -1

#undef _

// https://msdn.microsoft.com/en-us/library/hh916382.aspx

__declspec(safebuffers)
BOOL	CDECL	PrintConsole(HANDLE hConsole, [SA_FormatString(Style="printf")] LPCWSTR lpFormat, ...)
{
	WCHAR	szOutput[1024];
	DWORD	dwOutput;
	DWORD	dwConsole;

	// https://msdn.microsoft.com/en-us/library/kb57fad8.aspx

	va_list	vaInput;
	va_start(vaInput, lpFormat);

	// https://msdn.microsoft.com/en-us/library/ms647551.aspx

	dwOutput = wvsprintf(szOutput, lpFormat, vaInput);

	va_end(vaInput);

	if (dwOutput == 0UL)
		return FALSE;

	// https://msdn.microsoft.com/en-us/library/ms687401.aspx

	if (!WriteConsole(hConsole, szOutput, dwOutput, &dwConsole, NULL))
		return FALSE;

	return dwConsole == dwOutput;
}

// https://msdn.microsoft.com/en-us/library/f9t8842e.aspx

__declspec(noreturn)
VOID	CDECL	wmainCRTStartup(VOID)
{
	DWORD	dwError = ERROR_SUCCESS;
	DWORD	dw;
	QWORD	ullQuotient, ullRemainder;
	SQWORD	llQuotient, llRemainder;

	// https://msdn.microsoft.com/en-us/library/ms683231.aspx

	HANDLE	hConsole = GetStdHandle(STD_ERROR_HANDLE);

	if (hConsole == INVALID_HANDLE_VALUE)
		dwError = GetLastError();
	else
	{
		PrintConsole(hConsole, L"\nTesting _aulldvrm() ...\n");

		for (dw = 0UL; dw < sizeof(ullTable) / sizeof(*ullTable); dw++)
		{
			PrintConsole(hConsole, L"\r%lu", dw);

			ullQuotient = ullTable[dw].ullDividend / ullTable[dw].ullDivisor;
			ullRemainder = ullTable[dw].ullDividend % ullTable[dw].ullDivisor;

			if (ullQuotient != ullTable[dw].ullQuotient)
				PrintConsole(hConsole,
				             L"\t%I64u / %I64u:\a quotient %I64u not equal %I64u\n",
				             ullTable[dw].ullDividend, ullTable[dw].ullDivisor, ullQuotient, ullTable[dw].ullQuotient);

			if (ullQuotient > ullTable[dw].ullDividend)
				PrintConsole(hConsole,
				             L"\t%I64u / %I64u:\a quotient %I64u greater dividend\n",
				             ullTable[dw].ullDividend, ullTable[dw].ullDivisor, ullQuotient);

			if (ullRemainder != ullTable[dw].ullRemainder)
				PrintConsole(hConsole,
				             L"\t%I64u %% %I64u:\a remainder %I64u not equal %I64u\n",
				             ullTable[dw].ullDividend, ullTable[dw].ullDivisor, ullRemainder, ullTable[dw].ullRemainder);

			if (ullRemainder >= ullTable[dw].ullDivisor)
				PrintConsole(hConsole,
				             L"\t%I64u %% %I64u:\a remainder %I64u not less divisor\n",
				             ullTable[dw].ullDividend, ullTable[dw].ullDivisor, ullRemainder);
		}

		PrintConsole(hConsole, L"\nTesting _aulldiv() and _aullrem() ...\n");

		for (dw = 0UL; dw < sizeof(ullTable) / sizeof(*ullTable); dw++)
		{
			PrintConsole(hConsole, L"\r%lu", dw);

			ullQuotient = ullTable[dw].ullDividend / ullTable[dw].ullDivisor;

			if (ullQuotient != ullTable[dw].ullQuotient)
				PrintConsole(hConsole,
				             L"\t%I64u / %I64u:\a quotient %I64u not equal %I64u\n",
				             ullTable[dw].ullDividend, ullTable[dw].ullDivisor, ullQuotient, ullTable[dw].ullQuotient);

			if (ullQuotient > ullTable[dw].ullDividend)
				PrintConsole(hConsole,
				             L"\t%I64u / %I64u:\a quotient %I64u greater dividend\n",
				             ullTable[dw].ullDividend, ullTable[dw].ullDivisor, ullQuotient);

			ullRemainder = ullTable[dw].ullDividend - ullTable[dw].ullDivisor * ullQuotient;

			if (ullRemainder != ullTable[dw].ullRemainder)
				PrintConsole(hConsole,
				             L"\t%I64u / %I64u:\a remainder %I64u not equal %I64u\n",
				             ullTable[dw].ullDividend, ullTable[dw].ullDivisor, ullRemainder, ullTable[dw].ullRemainder);

			ullRemainder = ullTable[dw].ullDividend % ullTable[dw].ullDivisor;

			if (ullRemainder != ullTable[dw].ullRemainder)
				PrintConsole(hConsole,
				             L"\t%I64u %% %I64u:\a remainder %I64u not equal %I64u\n",
				             ullTable[dw].ullDividend, ullTable[dw].ullDivisor, ullRemainder, ullTable[dw].ullRemainder);

			if (ullRemainder >= ullTable[dw].ullDivisor)
				PrintConsole(hConsole,
				             L"\t%I64u %% %I64u:\a remainder %I64u not less divisor\n",
				             ullTable[dw].ullDividend, ullTable[dw].ullDivisor, ullRemainder);
		}

		PrintConsole(hConsole, L"\nTesting _alldvrm() ...\n");

		for (dw = 0UL; dw < sizeof(llTable) / sizeof(*llTable); dw++)
		{
			PrintConsole(hConsole, L"\r%lu", dw);

			llQuotient = llTable[dw].llDividend / llTable[dw].llDivisor;
			llRemainder = llTable[dw].llDividend % llTable[dw].llDivisor;

			if (llQuotient != llTable[dw].llQuotient)
				PrintConsole(hConsole,
				             L"\t%I64d / %I64d:\a quotient %I64d not equal %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llQuotient, llTable[dw].llQuotient);

			if ((llTable[dw].llDividend < 0LL) && (llQuotient < llTable[dw].llDividend)
			 || (llTable[dw].llDividend >= 0LL) && (llQuotient > llTable[dw].llDividend))
				PrintConsole(hConsole,
				             L"\t%I64d / %I64d:\a quotient %I64d greater dividend\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llQuotient);

			if (llRemainder != llTable[dw].llRemainder)
				PrintConsole(hConsole,
				             L"\t%I64d %% %I64d:\a remainder %I64d not equal %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder, llTable[dw].llRemainder);

			if ((llTable[dw].llDivisor < 0LL) && (llRemainder <= llTable[dw].llDivisor)
			 || (llTable[dw].llDivisor > 0LL) && (llRemainder >= llTable[dw].llDivisor))
				PrintConsole(hConsole,
				             L"\t%I64d %% %I64d:\a remainder %I64d not less divisor\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder);

			if ((llRemainder != 0LL) && ((llRemainder < 0LL) != (llTable[dw].llDividend < 0LL)))
				PrintConsole(hConsole,
				             L"\t%I64d %% %I64d:\a sign of remainder %I64d not equal sign of quotient %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder, llTable[dw].llDividend);

			llRemainder = llTable[dw].llDividend - llTable[dw].llDivisor * llQuotient;

			if (llRemainder != llTable[dw].llRemainder)
				PrintConsole(hConsole,
				             L"\t%I64d / %I64d:\a remainder %I64d not equal %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder, llTable[dw].llRemainder);

			if ((llRemainder != 0LL) && ((llRemainder < 0LL) != (llTable[dw].llDividend < 0LL)))
				PrintConsole(hConsole,
				             L"\t%I64d / %I64d:\a sign of remainder %I64d not equal sign of quotient %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder, llTable[dw].llDividend);
		}

		PrintConsole(hConsole, L"\nTesting _alldiv() and _allrem() ...\n");

		for (dw = 0UL; dw < sizeof(llTable) / sizeof(*llTable); dw++)
		{
			PrintConsole(hConsole, L"\r%lu", dw);

			llQuotient = llTable[dw].llDividend / llTable[dw].llDivisor;

			if (llQuotient != llTable[dw].llQuotient)
				PrintConsole(hConsole,
				             L"\t%I64d / %I64d:\a quotient %I64d not equal %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llQuotient, llTable[dw].llQuotient);

			if ((llTable[dw].llDividend < 0LL) && (llQuotient < llTable[dw].llDividend)
			 || (llTable[dw].llDividend >= 0LL) && (llQuotient > llTable[dw].llDividend))
				PrintConsole(hConsole,
				             L"\t%I64d / %I64d:\a quotient %I64d greater dividend\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llQuotient);

			llRemainder = llTable[dw].llDividend - llTable[dw].llDivisor * llQuotient;

			if (llRemainder != llTable[dw].llRemainder)
				PrintConsole(hConsole,
				             L"\t%I64d / %I64d:\a remainder %I64d not equal %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder, llTable[dw].llRemainder);

			if ((llRemainder != 0LL) && ((llRemainder < 0LL) != (llTable[dw].llDividend < 0LL)))
				PrintConsole(hConsole,
				             L"\t%I64d / %I64d:\a sign of remainder %I64d not equal sign of quotient %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder, llTable[dw].llDividend);

			llRemainder = llTable[dw].llDividend % llTable[dw].llDivisor;

			if (llRemainder != llTable[dw].llRemainder)
				PrintConsole(hConsole,
				             L"\t%I64d %% %I64d:\a remainder %I64d not equal %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder, llTable[dw].llRemainder);

			if ((llTable[dw].llDivisor < 0LL) && (llRemainder <= llTable[dw].llDivisor)
			 || (llTable[dw].llDivisor > 0LL) && (llRemainder >= llTable[dw].llDivisor))
				PrintConsole(hConsole,
				             L"\t%I64d %% %I64d:\a remainder %I64d not less divisor\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder);

			if ((llRemainder != 0LL) && ((llRemainder < 0LL) != (llTable[dw].llDividend < 0LL)))
				PrintConsole(hConsole,
				             L"\t%I64d %% %I64d:\a sign of remainder %I64d not equal sign of quotient %I64d\n",
				             llTable[dw].llDividend, llTable[dw].llDivisor, llRemainder, llTable[dw].llDividend);
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms682658.aspx

	ExitProcess(dwError);
}
<<NOKEEP

I386-CRT.OBJ: I386-MSC.MAK
	CL.EXE /Brepro /c /Fo$@ /GAFdwy /J /nologo /O2txy /Tc<< /W4 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(lib, "kernel32.lib")
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

#define STRICT
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

// https://msdn.microsoft.com/en-us/library/3ze4ytsc.aspx
// https://msdn.microsoft.com/en-us/library/6wd819wh.aspx
// https://msdn.microsoft.com/en-us/library/bky3b5dh.aspx

extern	int	main(int argc, char const *argv[], char const *envp[]);
extern	int	wmain(int argc, wchar_t const *argv[], wchar_t const *envp[]);

// https://msdn.microsoft.com/en-us/library/f9t8842e.aspx

__declspec(noreturn)
__declspec(safebuffers)
VOID	CDECL	mainCRTStartup(VOID)
{
	LPSTR	lpArgument;

	// https://msdn.microsoft.com/en-us/library/ms683187.aspx

	LPCSTR	lpBlock = GetEnvironmentStringsA();

	// https://msdn.microsoft.com/en-us/library/ms683156.aspx

	LPCSTR	lpCmdLine = GetCommandLineA();

	LPCSTR	lpCount = lpCmdLine;
	UINT	uiCount = 0U;
	UINT	uiQuote = 0U;
	UINT	argc;
	LPCSTR	*argv;
	LPCSTR	*envp;

	argc = (*lpCount != ' ' && *lpCount != '\t');

	while (*lpCount != '\0')	// count arguments
	{
		if (!uiQuote		// whitespace outside double quotes?
		 && (*lpCount == ' ' || *lpCount == '\t'))
		{
			do		// skip unquoted whitespace
				lpCount++;
			while (*lpCount == ' ' || *lpCount == '\t');

			argc += (*lpCount != '\0');

			uiCount = 0U;

			continue;
		}
		else if (*lpCount == '\\')
			uiCount ^= ~0U;
		else if (!uiCount	// unescaped double quote?
		      && *lpCount == '"')
			uiQuote ^= ~0U;
		else			// regular character
			uiCount = 0U;

		lpCount++;
	}

	if (uiQuote)			// unpaired double quote?

		// https://msdn.microsoft.com/en-us/library/ms680627.aspx

		SetLastError(ERROR_BAD_ARGUMENTS);

	// https://msdn.microsoft.com/en-us/library/wb1s57t5.aspx

	argv = (LPCSTR *) _alloca((argc + 1U) * sizeof(*argv) + (lpCount + 1U - lpCmdLine) * sizeof(**argv));

	argv[0] = lpArgument = (LPSTR) (argv + argc + 1U);

	argc = uiCount = uiQuote = 0U;

	while (*lpCmdLine != '\0')	// process arguments
	{
		if (!uiQuote		// whitespace outside double quotes?
		 && (*lpCmdLine == ' ' || *lpCmdLine == '\t'))
		{			// terminate current argument
			*lpArgument = '\0';

			do		// skip unquoted whitespace
				lpCmdLine++;
			while (*lpCmdLine == ' ' || *lpCmdLine == '\t');

			if (*lpCmdLine != '\0')
					// store address of next argument
				argv[++argc] = lpArgument = (LPSTR) lpCmdLine;

			uiCount = 0U;
		}
		else if (*lpCmdLine == '\\')
		{
			*lpArgument++ = *lpCmdLine++;
			uiCount++;	// count backslash
		}
		else if (*lpCmdLine == '"')
		{
			lpArgument -= (uiCount + 1U) / 2U;

			if (uiCount & 1U)
					// double quote preceeded by odd number
					//  of backslashes: keep half of them
					//   and the (escaped) double quote
				*lpArgument++ = *lpCmdLine++;
			else		// double quote preceeded by even number
					//  of backslashes: keep half of them
				if (*++lpCmdLine == '"' && uiQuote)
					// double quote inside double quotes and
					//  followed by another double quote:
					//   keep one double quote
					*lpArgument++ = *lpCmdLine++;
				else	// skip double quote and toggle state
					uiQuote ^= ~0U;
			uiCount = 0U;
		}
		else			// regular character
		{
			*lpArgument++ = *lpCmdLine++;
			uiCount = 0U;
		}
	}

	*lpArgument = '\0';		// terminate (last) argument

	argv[++argc] = NULL;		// store terminating NULL pointer
	envp = argv + argc;

	if (lpBlock != NULL)
	{
		for (uiCount = 0U,	// count environment strings
		     lpCount = lpBlock; *lpCount != '\0'; lpCount += strlen(lpCount) + 1U)
			if (*lpCount != '=')
				uiCount++;

		if (uiCount > 0U)	// process environment strings
		{
			// https://msdn.microsoft.com/en-us/library/wb1s57t5.aspx

			envp = (LPCSTR *) _alloca((uiCount + 1U) * sizeof(*envp));

			for (uiCount = 0U,
			     lpCount = lpBlock; *lpCount != '\0'; lpCount += strlen(lpCount) + 1U)
				if (*lpCount != '=')
					envp[uiCount++] = lpCount;

			envp[uiCount] = (LPCSTR) NULL;
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms686036.aspx
	// https://msdn.microsoft.com/en-us/library/dd318070.aspx

	if (!SetConsoleOutputCP(GetACP()))
		/* no error handling */;

	// https://msdn.microsoft.com/en-us/library/ms682658.aspx

	ExitProcess(main(argc, argv, envp));
}

// https://msdn.microsoft.com/en-us/library/f9t8842e.aspx

__declspec(noreturn)
__declspec(safebuffers)
VOID	CDECL	wmainCRTStartup(VOID)
{
	LPWSTR	lpArgument;

	// https://msdn.microsoft.com/en-us/library/ms683187.aspx

	LPCWSTR	lpBlock = GetEnvironmentStringsW();

	// https://msdn.microsoft.com/en-us/library/ms683156.aspx

	LPCWSTR	lpCmdLine = GetCommandLineW();

	LPCWSTR	lpCount = lpCmdLine;
	UINT	uiCount = 0U;
	UINT	uiQuote = 0U;
	UINT	argc;
	LPCWSTR	*argv;
	LPCWSTR	*envp;

	argc = (*lpCount != L' ' && *lpCount != L'\t');

	while (*lpCount != L'\0')	// count arguments
	{
		if (!uiQuote		// whitespace outside double quotes?
		 && (*lpCount == L' ' || *lpCount == L'\t'))
		{
			do		// skip unquoted whitespace
				lpCount++;
			while (*lpCount == L' ' || *lpCount == L'\t');

			argc += (*lpCount != L'\0');

			uiCount = 0U;

			continue;
		}
		else if (*lpCount == L'\\')
			uiCount ^= ~0U;
		else if (!uiCount	// unescaped double quote?
		      && *lpCount == L'"')
			uiQuote ^= ~0U;
		else			// regular character
			uiCount = 0U;

		lpCount++;
	}

	if (uiQuote)			// unpaired double quote?

		// https://msdn.microsoft.com/en-us/library/ms680627.aspx

		SetLastError(ERROR_BAD_ARGUMENTS);

	// https://msdn.microsoft.com/en-us/library/wb1s57t5.aspx

	argv = (LPCWSTR *) _alloca((argc + 1U) * sizeof(*argv) + (lpCount + 1U - lpCmdLine) * sizeof(**argv));

	argv[0] = lpArgument = (LPWSTR) (argv + argc + 1U);

	argc = uiCount = uiQuote = 0U;

	while (*lpCmdLine != L'\0')	// process arguments
	{
		if (!uiQuote		// whitespace outside double quotes?
		 && (*lpCmdLine == L' ' || *lpCmdLine == L'\t'))
		{			// terminate current argument
			*lpArgument = L'\0';

			do		// skip unquoted whitespace
				lpCmdLine++;
			while (*lpCmdLine == L' ' || *lpCmdLine == L'\t');

			if (*lpCmdLine != L'\0')
					// store address of next argument
				argv[++argc] = lpArgument = (LPWSTR) lpCmdLine;

			uiCount = 0U;
		}
		else if (*lpCmdLine == L'\\')
		{
			*lpArgument++ = *lpCmdLine++;
			uiCount++;	// count backslash
		}
		else if (*lpCmdLine == L'"')
		{
			lpArgument -= (uiCount + 1U) / 2U;

			if (uiCount & 1U)
					// double quote preceeded by odd number
					//  of backslashes: keep half of them
					//   and the (escaped) double quote
				*lpArgument++ = *lpCmdLine++;
			else		// double quote preceeded by even number
					//  of backslashes: keep half of them
				if (*++lpCmdLine == L'"' && uiQuote)
					// double quote inside double quotes and
					//  followed by another double quote:
					//   keep one double quote
					*lpArgument++ = *lpCmdLine++;
				else	// skip double quote and toggle state
					uiQuote ^= ~0U;
			uiCount = 0U;
		}
		else			// regular character
		{
			*lpArgument++ = *lpCmdLine++;
			uiCount = 0U;
		}
	}

	*lpArgument = L'\0';		// terminate (last) argument

	argv[++argc] = NULL;		// store terminating NULL pointer
	envp = argv + argc;

	if (lpBlock != NULL)
	{
		for (uiCount = 0U,	// count environment strings
		     lpCount = lpBlock; *lpCount != L'\0'; lpCount += wcslen(lpCount) + 1U)
			if (*lpCount != L'=')
				uiCount++;

		if (uiCount > 0U)	// process environment strings
		{
			// https://msdn.microsoft.com/en-us/library/wb1s57t5.aspx

			envp = (LPCWSTR *) _alloca((uiCount + 1U) * sizeof(*envp));

			for (uiCount = 0U,
			     lpCount = lpBlock; *lpCount != L'\0'; lpCount += wcslen(lpCount) + 1U)
				if (*lpCount != L'=')
					envp[uiCount++] = lpCount;

			envp[uiCount] = (LPCWSTR) NULL;
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms682658.aspx

	ExitProcess(wmain(argc, argv, envp));
}
<<NOKEEP

I386-FMT.OBJ: I386-MSC.MAK
	CL.EXE /analyze /Brepro /c /DOPTIMIZE /Fo$@ /GAFwyz /J /nologo /O2txy /Tc<< /W4 /wd4100 /wd4213 /wd4554 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(lib, "kernel32.lib")
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

typedef	unsigned long	size_t;
typedef	unsigned short	wchar_t;

// https://msdn.microsoft.com/en-us/library/k64h826d.aspx

typedef	char	*va_list;

#define va_arg(ptr, type)	(*(type *) ((ptr += ((sizeof(type) + sizeof(int) - 1) & (0 - sizeof(int)))) - ((sizeof(type) + sizeof(int) - 1) & (0 - sizeof(int)))))
#define va_copy(dst, src)	((dst) = (src))
#define va_end(ptr)		(ptr = (va_list) 0)
#define va_start(ptr, arg)	(ptr = (va_list) &arg + ((sizeof(arg) + sizeof(int) - 1) & (0 - sizeof(int))))

#define	NULL	(void *) 0

// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx
// https://msdn.microsoft.com/en-us/library/hhta9830.aspx

#define memcpy	__movsb

// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx
// https://msdn.microsoft.com/en-us/library/ywwcatsa.aspx

#define memset	__stosb

// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx
// https://msdn.microsoft.com/en-us/library/adf7db3x.aspx

#define wmemcpy	__movsw

// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx
// https://msdn.microsoft.com/en-us/library/k99wy47b.aspx

#define wmemset	__stosw

// https://msdn.microsoft.com/en-us/library/ms776446.aspx
// https://msdn.microsoft.com/en-us/library/dd317756.aspx
// https://msdn.microsoft.com/en-us/library/ms776413.aspx
// https://msdn.microsoft.com/en-us/library/dd319072.aspx

#define CP_ACP			0U
#define MB_ERR_INVALID_CHARS	1UL
#define MB_PRECOMPOSED		8UL

__declspec(dllimport)
int	MultiByteToWideChar(unsigned int  uiCodePage,
	                    unsigned long dwFlags,
	                    char const    *lpMultiByte,
	                    int           cbMultiByte,
	                    wchar_t const *lpWideChar,
	                    int           cchWideChar);

__declspec(align(256))
static
const	char	ANSI[256] = "\0\1\2\3\4\5\6\a\b\t\n\v\f\r\16\17"
		            "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F"
		            "\x20\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2A\x2B\x2C\x2D\x2E\x2F" // " !"#$$%&'()*+,-./"
		            "\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\x3A\x3B\x3C\x3D\x3E\x3F" // "0123456789:;<=>?"
		            "\x40\x41\x42\x43\x44\x45\x46\x47\x48\x49\x4A\x4B\x4C\x4D\x4E\x4F" // "@ABCDEFGHIJKLMNO"
		            "\x50\x51\x52\x53\x54\x55\x56\x57\x58\x59\x5A\x5B\x5C\x5D\x5E\x5F" // "PQRSTUVWXYZ[\\]^_"
		            "\x60\x61\x62\x63\x64\x65\x66\x67\x68\x69\x6A\x6B\x6C\x6D\x6E\x6F" // "`abcdefghijklmno"
		            "\x70\x71\x72\x73\x74\x75\x76\x77\x78\x79\x7A\x7B\x7C\x7D\x7E\x7F" // "pqrstuvwxyz{|}~ "
		            "\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8A\x8B\x8C\x8D\x8E\x8F"
		            "\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9A\x9B\x9C\x9D\x9E\x9F"
		            "\xA0\xA1\xA2\xA3\xA4\xA5\xA6\xA7\xA8\xA9\xAA\xAB\xAC\xAD\xAE\xAF"
		            "\xB0\xB1\xB2\xB3\xB4\xB5\xB6\xB7\xB8\xB9\xBA\xBB\xBC\xBD\xBE\xBF"
		            "\xC0\xC1\xC2\xC3\xC4\xC5\xC6\xC7\xC8\xC9\xCA\xCB\xCC\xCD\xCE\xCF"
		            "\xD0\xD1\xD2\xD3\xD4\xD5\xD6\xD7\xD8\xD9\xDA\xDB\xDC\xDD\xDE\xDF"
		            "\xE0\xE1\xE2\xE3\xE4\xE5\xE6\xE7\xE8\xE9\xEA\xEB\xEC\xED\xEE\xEF"
		            "\xF0\xF1\xF2\xF3\xF4\xF5\xF6\xF7\xF8\xF9\xFA\xFB\xFC\xFD\xFE\xFF";

__declspec(align(512))
static	wchar_t	UCS2[256];

__declspec(naked)
unsigned long long	__cdecl	__ull_divrem(unsigned long long dividend, unsigned long divisor, unsigned long *remainder)
{
	__asm	// returns quotient in edx:eax
	{
		mov	ecx, [esp+12]	; ecx = divisor
#if 0
		xor	edx, edx
		mov	eax, [esp+8]	; edx:eax = high dword of dividend
		cmp	eax, edx
		je	short common	; high dword of dividend = 0?
#else
		xor	eax, eax
		mov	edx, [esp+8]	; eax:edx = high dword of dividend
		cmp	edx, ecx
		jb	short common	; high dword of dividend < divisor?
					; quotient < 2**32?
		xchg	eax, edx
#endif
		div	ecx		; eax = high dword of quotient,
					; edx = high dword of dividend'
	common:
		push	eax
		mov	eax, [esp+8]	; edx:eax = dividend'
		div	ecx		; eax = low dword of quotient,
					; edx = remainder
		mov	ecx, [esp+20]	; ecx = &remainder
		mov	[ecx], edx	; *remainder = remainder
		pop	edx		; edx:eax = quotient
		ret
	}
}

__forceinline
int	__cdecl	tiebreak(char const *decimal, int start, int end)
{
	while (start < end)
		if (decimal[start++] > 0U)
			return 1;
	return 0;
}

// https://msdn.microsoft.com/en-us/library/56e442dc.aspx
// https://msdn.microsoft.com/en-us/library/8aky45ct.aspx
// https://msdn.microsoft.com/en-us/library/25366k66.aspx
// https://msdn.microsoft.com/en-us/library/0ecbz014.aspx

// '%' [{ ' ' | '#' | '\'' | '0' | '+' | '-' } ... ] [{ width | '*' }] [ '.' { precision | '*' }] [{ 'h' | 'hh' | 'l' | 'll' | 'I32' | 'I64' | 'L' | 't' | 'z' }] { 'a' | 'A' | 'c' | 'C' | 'd' | 'e' | 'E' | 'f' | 'F' | 'i' | 'n' | 'o' | 'p' | 's' | 'S' | 'u' | 'x' | 'X' }

// https://msdn.microsoft.com/en-us/library/tcxf1dw6.aspx

// signed int			d, i, n, o, x, X
// unsigned int			u
#ifdef C99 // also POSIX
// signed char		hh	d, i, n, o, x, X
// unsigned char	hh	u
#endif
// signed short		h	d, i, n, o, x, X
// unsigned short	h	u
// signed long		l	d, i, n, o, x, X
// unsigned long	l	u
#ifdef C99 // also POSIX
// signed long long	ll	d, i, n, o, x, X
// unsigned long long	ll	u
// unsigned void *	t	u
// signed void *	z	d, i, n, o, x, X
#endif

// https://msdn.microsoft.com/en-us/library/hf4y5e3w.aspx

//				d = signed integer to base 10, optionally prefixed [+- 0]
//				i = signed integer to base 10, optionally prefixed [+- 0]
//				o = unsigned integer to base 8, optionally prefixed [+-0]
//				u = unsigned integer to base 10, unsigned
//				x = unsigned integer to base 16, optionally prefixed 0x
//				X = unsigned integer to base 16, optionally prefixed 0X
//
// int				n = character count
// void *			p = unsigned pointer to base 16, optionally prefixed 0x
// void *			P = unsigned pointer to base 16, optionally prefixed 0X
//
// double			a = IEEE-754 double to base 16
// double			A = IEEE-754 double to base 16
//
// double			e = IEEE-754 double to decimal
// double			E = IEEE-754 double to decimal
//
// double		l	f, g, G			(NOT SUPPORTED)
// long double		ll	e, E, f, g, G		(NOT SUPPORTED)
// long double		L	e, E, f, g, G		(NOT SUPPORTED)
//
// char				C			(UNICODE function)
// char				c			(ANSI function)
// char			h	c, C			(ANSI & UNICODE function)
// wchar_t			C			(ANSI function)
// wchar_t			c			(UNICODE function)
// wchar_t		l	c, C			(ANSI & UNICODE function)
// char	*			S			(UNICODE function)
// char	*			s, [...]		(ANSI function)
// char	*		h	s, S			(ANSI & UNICODE function)
// wchar_t *			S			(ANSI function)
// wchar_t *			s, [...]		(UNICODE function)
// wchar_t *		l	s, S			(ANSI & UNICODE function)

// https://msdn.microsoft.com/en-us/library/1kt27hek.aspx

__declspec(safebuffers)
int	__cdecl	vsnprintf(char *output, size_t count, char const *format, va_list input)
{
	char	const * const start = format;
	char	const * const begin = output;
	char	const * const end = (count < 2) ? output : output + count - 2;
	char	const *ansi;
	wchar_t	const *unicode;
	wchar_t	cw;
	char	cb;
	char	pad;
	char	sign;
	char	size;
	char	scratch[24];		// characters from integer conversion:
					//  ~0ULL has 22 octal characters
	char	decimal[768];		// digits from floating-point conversion:
					//  2.0**-x has x fractional decimal digits;
					//  0x1p-1074 = 4.9406564584124654...e-324
					//   has 323 significant zeroes after the decimal point,
					//    followed by 751 digits;
					//  0x0.FFFFFFFFFFFFFp-1022 = 2.2250738585072008...e-308
					//   has 307 significant zeroes after the decimal point,
					//    followed by 767 digits;
					//  0x1p+1024 = 1.7976931348623159...e+308 has 309 digits
	char	digit;
	int	length;			// number of characters from conversion
	int	index;			// index of first digit from conversion
	int	left;
	int	hash;
	int	blank;
//	int	group;
	int	width;
	int	precision;
	int	exponent;
	int	point;
	int	extra;

	unsigned long		ul;
	unsigned long long	ull;

	if (UCS2['\t'] == L'\0')	// ANSI to UTF-16LE conversion table not initialized?

		// https://msdn.microsoft.com/en-us/library/ms776413.aspx
		// https://msdn.microsoft.com/en-us/library/dd319072.aspx

		if (MultiByteToWideChar(CP_ACP,
		                        MB_ERR_INVALID_CHARS | MB_PRECOMPOSED,
		                        ANSI, 256,
		                        UCS2, 256) != 256)
			return ~2147483647;

	for (;;)
	{
		while (*format != '%'	// copy format string up to next conversion specifier
		    && *format != '\0'	//  or its end to output string
		    && output < end)
			*output++ = *format++;

		if (*format == '\0')	// format string processed completely?
			break;

		if (*++format == '%')	// double '%' in format string?
		{
			if (output < end)
				*output++ = *format++;
			continue;
		}
					// conversion specifier in format string
		pad = ' ';
		sign = '\0';
		size = '\0';
		left = 0;
		hash = 0;
		blank = 0;
	//	group = 0;
		width = 0;
		length = 0;
		precision = -1;

		do			// flag(s) in format string?
			if (*format == '-')
				left = 1;
			else if (*format == '+')
				sign = '+';
			else if (*format == '0')
				pad = '0';
			else if (*format == ' ')
				blank = 1;
			else if (*format == '#')
				hash = 1;
		//	else if (*format == '\'')
		//		group = 1;
			else
				break;
		while (*++format != '\0');

		if (blank && sign == '\0')
			sign = ' ';

		if (*format >= '0'	// field width in format string?
		 && *format <= '9')
		{
			for (width = *format++ - '0';
			     *format >= '0' && *format <= '9';
			     width *= 10, width += *format++ - '0')
				continue;

			if (width < 0)
				width = 2147483647;
		}
		else if (*format == '*')
		{
			format++;

			width = va_arg(input, int);

			if (width < 0)
			{
				if (width == ~2147483647)
					width = 0;
				else
					width = -width;

				left = 1;
			}
		}

	//	if (left)
	//		pad = ' ';

		if (*format == '.')	// precision in format string?
		{
			pad = ' ';

			if (*++format == '*')
			{
				format++;

				precision = va_arg(input, int);

			//	if (precision < 0)
			//		precision = -1;
			}
			else if (*format >= '0' && *format <= '9')
			{
				for (precision = *format++ - '0';
				     *format >= '0' && *format <= '9';
				     precision *= 10, precision += *format++ - '0')
					continue;

				if (precision < 0)
					precision = 2147483647;
			}
			else
				precision = 0;
		}

		if (*format == 'j'	// size modifier in format string?
		 || *format == 't'
		 || *format == 'z')
			size = *format++;
		else if (*format == 'h'
		      || *format == 'l')
		{
			size = *format++;

			if (size == *format)
			{
				format++;
					// encode "hh" as 'H' and "ll" as 'L'
				size ^= 'a' ^ 'A';
			}
		}

		if (*format == 'd'	// conversion type in format string?
		 || *format == 'i')
		{			// signed decimal integer:
					//  [ ' ' ... | '+' | '-' ] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == 'j'
			 || size == 'L')
				(long long) ull = va_arg(input, long long);
			else if (size == 'h')
				(long long) ull = va_arg(input, short);
			else if (size == 'H')
				(long long) ull = va_arg(input, char);
			else // 'l', 't', 'z', '\0'
				(long long) ull = va_arg(input, int);

			if ((long long) ull < 0)
			{
				ull = 0 - ull;
				sign = '-';
			}
					// convert to decimal characters,
					//  least significant digit in scratch[0],
					//   most significant digit in scratch[length - 1]
#ifndef OPTIMIZE
			do
			{
				ull = __ull_divrem(ull, 10, &ul);
				scratch[length++] = '0' + (char) ul;
			} while (ull);
#else
			do		// least number of 6432-bit divisions
			{
				ull = __ull_divrem(ull, 1000000000, &ul);

				if (ull > 0)
				{
					index = length + 9;
#if 0
					do
						scratch[length++] = '0' + (char) (ul % 10);
					while (ul /= 10);

					while (length < index)
						scratch[length++] = '0';
#else
					// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

					memset(scratch + length, '0', 9);

					do
						scratch[length++] = '0' + (char) (ul % 10);
					while (ul /= 10);

					length = index;
#endif
				}
				else if (length == 0)
					do
						scratch[length++] = '0' + (char) (ul % 10);
					while (ul /= 10);
				else
					while (ul > 0)
						scratch[length++] = '0' + (char) (ul % 10), ul /= 10;
			} while (ull);
#endif // OPTIMIZE
					// index of decimal integer (leading '0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			if (sign != '\0')
				scratch[length++] = sign;

			goto NUMBER;
		}
		else if (*format == 'u')
		{			// unsigned decimal integer:
					//  [{ ' ' | '0' } ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == 'j'
			 || size == 'L')
				ull = va_arg(input, unsigned long long);
			else if (size == 'h')
				ull = va_arg(input, unsigned short);
			else if (size == 'H')
				ull = va_arg(input, unsigned char);
			else // 'l', 't', 'z', '\0'
				ull = va_arg(input, unsigned int);

					// convert to decimal characters,
					//  least significant digit in scratch[0],
					//   most significant digit in scratch[length - 1]
#ifndef OPTIMIZE
			do
			{
				ull = __ull_divrem(ull, 10, &ul);
				scratch[length++] = '0' + (char) ul;
			} while (ull);
#else
			do		// least number of 6432-bit divisions
			{
				ull = __ull_divrem(ull, 1000000000, &ul);

				if (ull > 0)
				{
					index = length + 9;
#if 0
					do
						scratch[length++] = '0' + (char) (ul % 10);
					while (ul /= 10);

					while (length < index)
						scratch[length++] = '0';
#else
					// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

					memset(scratch + length, '0', 9);

					do
						scratch[length++] = '0' + (char) (ul % 10);
					while (ul /= 10);

					length = index;
#endif
				}
				else if (length == 0)
					do
						scratch[length++] = '0' + (char) (ul % 10);
					while (ul /= 10);
				else
					while (ul > 0)
						scratch[length++] = '0' + (char) (ul % 10), ul /= 10;
			} while (ull);
#endif // OPTIMIZE
					// index of decimal integer (leading '0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			goto NUMBER;
		}
		else if (*format == 'o')
		{			// unsigned octal integer:
					//  [{ ' ' | '0' } ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == 'j'
			 || size == 'L')
				ull = va_arg(input, unsigned long long);
			else if (size == 'h')
				ull = va_arg(input, unsigned short);
			else if (size == 'H')
				ull = va_arg(input, unsigned char);
			else // 'l', 't', 'z', '\0'
				ull = va_arg(input, unsigned int);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = '0' + (char) (ull % 8);
			while (ull /= 8);
					// index of octal integer (leading '0' padding)
			index = length;

			if (hash)
#if 0
				if (precision >= 0)
					precision++;
#else
				precision += (precision >= 0);
#endif
			if (precision > length)
				precision -= length;
			else
				precision = 0;
					// prefix '0'
			if (hash)
				scratch[length++] = '0';

			goto NUMBER;
		}
#if 0
		else if (*format == 'x'
		      || *format == 'X')
#else
		else if ((*format | 'x' ^ 'X') == 'x')
#endif
		{			// unsigned hexadecimal integer:
					//  [ ' ' ... ] [ '0' { 'x' | 'X' }] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]

			ansi = "0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ 'X') / ('x' ^ 'X');

			if (size == 'j'
			 || size == 'L')
				ull = va_arg(input, unsigned long long);
			else if (size == 'h')
				ull = va_arg(input, unsigned short);
			else if (size == 'H')
				ull = va_arg(input, unsigned char);
			else // 'l', 't', 'z', '\0'
				ull = va_arg(input, unsigned int);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = ansi[ull % 16];
			while (ull /= 16);
					// index of hexadecimal integer (leading '0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			if (hash)
			{		// prefix "0x" or "0X"
				scratch[length++] = *format;
				scratch[length++] = '0';
			}

		NUMBER:	if (width > precision + length)
				width -= precision + length;
			else
				width = 0;

			if (left)
				while (length > index && output < end)
					*output++ = scratch[--length];
			else if (pad == '0')
			{
				while (length > index && output < end)
					*output++ = scratch[--length];
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = '0';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, '0', width);

				output += width;
#endif // OPTIMIZE
			}
			else
			{
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = ' ';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
#endif // OPTIMIZE
				while (length > index && output < end)
					*output++ = scratch[--length];
			}
#ifndef OPTIMIZE
			while (precision-- > 0 && output < end)
				*output++ = '0';
#else
			if (precision > end - output)
				precision = end - output;

			// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

			memset(output, '0', precision);

			output += precision;
#endif // OPTIMIZE
			while (length > 0 && output < end)
				*output++ = scratch[--length];
#ifndef OPTIMIZE
		//	if (left)
		RIGHT:		while (width-- > 0 && output < end)
					*output++ = ' ';
#else
		RIGHT:	if (left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
			}
#endif // OPTIMIZE
		}
#if 0
		else if (*format == 'p'
		      || *format == 'P')
#else
		else if ((*format | 'p' ^ 'P') == 'p')
#endif
		{			// unsigned hexadecimal pointer:
					//  [ ' ' ... ] [ '0' { 'x' | 'X' }] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]
			if (size != '\0'
			 || sign != '\0'
			 || pad != ' '
			 || precision != -1)
				goto INVALID;

			ansi = "0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ 'P') / ('p' ^ 'P');

			ul = va_arg(input, unsigned long);

			if (ul == 0)
				hash = 0;
			do
				scratch[length++] = ansi[ul % 16];
			while (ul /= 16);
#if 0
			while (length < 2 * sizeof(ul))
				scratch[length++] = '0';
#else
			if (length < 2 * sizeof(ul))
			{
				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(scratch + length, '0', 2 * sizeof(ul) - length);

				length = 2 * sizeof(ul);
			}
#endif
			if (hash)
			{		// prefix "0x" or "0X"
				scratch[length++] = *format ^ 'P' | 'X';
				scratch[length++] = '0';
			}

			if (!left)
				while (width-- > length && output < end)
					*output++ = ' ';

			while (length > 0 && output < end)
				*output++ = scratch[--length];

			goto RIGHT;
		}
		else if (*format == 'c')
		{			// ANSI or UNICODE character:
					//  [ ' ' ... ] [ <character> ] [ ' ' ... ]
			if (precision != -1)
				goto INVALID;

			if (size == 'l')
				goto C;	// UNICODE character?

			if (size != '\0'
			 && size != 'h')
				goto INVALID;

		c:	cb = va_arg(input, char);

		COMMON:	if (cb != '\0')
#if 0
				if (width > 0)
					width--;
#else
				width -= (width > 0);
#endif
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = ' ';
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
			}
#endif // OPTIMIZE
			if (cb != '\0' && output < end)
				*output++ = cb;

			goto RIGHT;
		}
		else if (*format == 'C')
		{			// ANSI or UNICODE character:
					//  [ ' ' ... ] [ <character> ] [ ' ' ... ]
			if (precision != -1)
				goto INVALID;

			if (size == 'h')
				goto c;	// ANSI character?

			if (size != '\0'
			 && size != 'l')
				goto INVALID;

		C:	cw = va_arg(input, wchar_t);

			if (cw == UCS2[cw & L'\xFF'])
				cb = (char) cw;
			else
				for (cb = '?', ul = 0x80; ul < 0xA0; ul++)
					if (cw == UCS2[ul])
					{
						cb = (char) ul;
						break;
					}
			goto COMMON;
		}
		else if (*format == 's')
		{			// ANSI or UNICODE character string:
					//  [ ' ' ... ] [ <character> ... ] [ ' ' ... ]
			if (size == 'l')
				goto S;	// UNICODE character string?

			if (size != '\0'
			 && size != 'h')
				goto INVALID;

		s:	ansi = va_arg(input, char *);

			if (ansi == NULL)
				ansi = "(null)";
#if 0
			// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

			length = strlen(ansi);

			if (precision >= 0 && length > precision)
				length = precision;
#else
			if (precision < 0)

				// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

				length = strlen(ansi);
			else
				for (length = 0; length < precision; length++)
					if (ansi[length] == '\0')
						break;
#endif
		STRING:	if (width > length)
				width -= length;
			else
				width = 0;
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = ' ';

			while (length-- > 0 && output < end)
				*output++ = *ansi++;
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
			}

			if (length > end - output)
				length = end - output;

			// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

			memcpy(output, ansi, length);

			output += length;
#endif // OPTIMIZE
			goto RIGHT;
		}
		else if (*format == 'S')
		{			// ANSI or UNICODE character string:
					//  [ ' ' ... ] [ <character> ... ] [ ' ' ... ]
			if (size == 'h')
				goto s;	// ANSI character string?

			if (size != '\0'
			 && size != 'l')
				goto INVALID;

		S:	unicode = va_arg(input, wchar_t *);

			if (unicode == NULL)
				unicode = L"(null)";
#if 0
			// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

			length = wcslen(unicode);

			if (precision >= 0 && length > precision)
				length = precision;
#else
			if (precision < 0)

				// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

				length = wcslen(unicode);
			else
				for (length = 0; length < precision; length++)
					if (unicode[length] == L'\0')
						break;
#endif
			if (width > length)
				width -= length;
			else
				width = 0;
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = ' ';
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
			}
#endif // OPTIMIZE
			while (length-- > 0 && output < end)
			{
				cw = *unicode++;

				if (cw == UCS2[cw & L'\xFF'])
					cb = (char) cw;
				else
					for (cb = '?', ul = 0x80; ul < 0xA0; ul++)
						if (cw == UCS2[ul])
						{
							cb = (char) ul;
							break;
						}

				*output++ = cb;
			}

			goto RIGHT;
		}
		else if (*format == 'n')
		{			// assign character count
			if (size == 'j'
			 || size == 'L')
				*va_arg(input, long long *) = (long long) (output - begin);
			else if (size == 'h')
				*va_arg(input, short *) = (short) (output - begin);
			else if (size == 'H')
				*va_arg(input, char *) = (char) (output - begin);
			else
				*va_arg(input, int *) = (int) (output - begin);
		}
#if 0
		else if (*format == 'a'
		      || *format == 'A')
#else
		else if ((*format | 'a' ^ 'A') == 'a')
#endif
		{			// hexadecimal double-precision floating-point:
					//  [ ' ' ... | '+' | '-' ] '0' { 'x' | 'X' } [ '0' ... ] <digit> [ '.' <digit> ... ] { 'p' | 'P' } { '+' | '-' } <digit> ... [ ' ' ... ]
			if (size != '\0')
				goto INVALID;

			ansi = "0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ 'A') / ('a' ^ 'A');
			ull = va_arg(input, unsigned long long);

			if ((long long) ull < 0)
			{		// strip sign bit
				ull &= ~0ULL >> 1;
				sign = '-';
			}
					// biased exponent
			exponent = (int) (ull >> 52);

			if (exponent == 2047)
			{
				if (sign != '\0')
					scratch[length++] = sign;
#ifndef OPTIMIZE
				if (ull & ~0ULL >> 12)
				{
					scratch[length++] = *format ^ 'A' | 'N';
					scratch[length++] = *format ^ 'A' | 'A';
					scratch[length++] = *format ^ 'A' | 'N';
				}
				else
				{
					scratch[length++] = *format ^ 'A' | 'I';
					scratch[length++] = *format ^ 'A' | 'N';
					scratch[length++] = *format ^ 'A' | 'F';
				}

				scratch[length] = '\0';
#else
				// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

				if (ull & ~0ULL >> 12)
					memcpy(scratch + length, *format == 'a' ? "nan" : "NAN", sizeof("NAN"));
				else
					memcpy(scratch + length, *format == 'a' ? "inf" : "INF", sizeof("INF"));

				length += 3;
#endif // OPTIMIZE
				ansi = scratch;

				goto STRING;
			}
			else if (exponent > 0)
			{		// unbiased exponent
				exponent -= 1023;
					// strip exponent, insert hidden bit
				ull &= ~0ULL >> 12;
				ull |= 1ULL << 52;
			}
			else		// 0.0 or denormal floating-point number
#ifndef OPTIMIZE
				if (ull)
				{	// unbiased exponent
					exponent = 1 - 1023;
					// normalize significand
					while (ull < 1ULL << 52)
						ull += ull, exponent--;
				}
#else
				// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

				if (_BitScanReverse(&ul, (unsigned long) (ull >> 32)))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - 32 - ul);
					// normalize significand
					ull = __ll_lshift(ull, 52 - 32 - ul);
				}
				else if (_BitScanReverse(&ul, (unsigned long) ull))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - ul);
					// normalize significand
					ull = __ll_lshift(ull, 52 - 31);
					ull = __ll_lshift(ull, 31 - ul);
				}
#endif // OPTIMIZE
					// convert significand to hexadecimal, one digit per byte,
					//  least significant digit in decimal[length - 1],
					//   most significant digit in decimal[0],
					//    hexadecimal point after most significant digit
			do
				decimal[length++] = (char) (ull >> 52);
			while (ull &= ~0ULL >> 12, ull <<= 4);

			point = 0;
			extra = 0;

			if (precision < 0)
				precision = length - 1;
			else if (precision > length - 1)
			{		// number of additional trailing zeroes
				extra = precision - length + 1;
				precision = length - 1;
				hash = 1;
			}
			else if (precision < length - 1
			      && (decimal[precision + 1] > 8U
			       || decimal[precision + 1] == 8U && decimal[precision] % 2U
			       || tiebreak(decimal, precision + 2, length)))
			{		// round to nearest, ties to even
#if 0
				index = precision + 1;
				digit = 8U;

				do
					digit += decimal[index],
					decimal[index] = digit % 16U;
				while (digit /= 16U && index-- > 0);

				if (digit)
				{	// hexadecimal overflow: increment exponent
					exponent++;
					// set leading digit
					decimal[0] = digit;
				}
#else
				index = precision;

				while (index >= 0 && decimal[index] == 15U)
					decimal[index--] = 0U;

				if (index < 0)
				{	// hexadecimal overflow: increment exponent
					exponent++;
					// set leading digit
					decimal[0] = 1U;
				}
				else
					decimal[index] += 1U;
#endif
			}
					// hexadecimal exponent
			if (exponent < 0)
			{
				exponent = -exponent;
				size = '-';
			}
			else if (exponent > 0)
				size = '+';
		//	else
		//		size = '\0';

			length = 0;

			do		// hexadecimal exponent
				scratch[length++] = '0' + (char) (exponent % 10U);
			while (exponent /= 10U);
					// exponent sign
			if (size != '\0')
				scratch[length++] = size;
					// exponent prefix 'p' or 'P'
			scratch[length++] = *format ^ 'A' | 'P';
					// index of exponent prefix (trailing '0' padding)
			exponent = length;
					// hexadecimal point
			if (precision > 0 || hash > 0)
				scratch[length++] = '.';
					// most significant digit
			scratch[length++] = '0' + decimal[0];
					// index of most significant digit (leading '0' padding)
			index = length;
					// prefix "0x" or "0X"
			scratch[length++] = *format ^ 'A' | 'X';
			scratch[length++] = '0';
					// sign
			if (sign != '\0')
				scratch[length++] = sign;

			goto DOUBLE;
		}
#if 0
		else if (*format == 'e'
		      || *format == 'E')
#else
		else if ((*format | 'e' ^ 'E') == 'e')
#endif
		{			// decimal double-precision floating-point:
					//  [ ' ' ... | '+' | '-' ] [ '0' ... ] <digit> [ '.' <digit> ... [ '0' ... ]] { 'e' | 'E' } [ '+' | '-' ] <digit> ... [ ' ' ... ]
			if (size != '\0')
				goto INVALID;

			ansi = NULL;

			ull = va_arg(input, unsigned long long);

			if ((long long) ull < 0)
			{		// strip sign bit
				ull &= ~0ULL >> 1;
				sign = '-';
			}
					// biased exponent
			exponent = (int) (ull >> 52);

			if (exponent == 2047)
			{
				if (sign != '\0')
					scratch[length++] = sign;
#ifndef OPTIMIZE
				if (ull & ~0ULL >> 12)
				{
					scratch[length++] = *format ^ 'E' | 'N';
					scratch[length++] = *format ^ 'E' | 'A';
					scratch[length++] = *format ^ 'E' | 'N';
				}
				else
				{
					scratch[length++] = *format ^ 'E' | 'I';
					scratch[length++] = *format ^ 'E' | 'N';
					scratch[length++] = *format ^ 'E' | 'F';
				}

				scratch[length] = '\0';
#else
				// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

				if (ull & ~0ULL >> 12)
					memcpy(scratch + length, *format == 'e' ? "nan" : "NAN", sizeof("NAN"));
				else
					memcpy(scratch + length, *format == 'e' ? "inf" : "INF", sizeof("INF"));

				length += 3;
#endif // OPTIMIZE
				ansi = scratch;

				goto STRING;
			}
			else if (exponent > 0)
			{		// unbiased exponent
				exponent -= 1023;
					// strip exponent, insert hidden bit
				ull &= ~0ULL >> 12;
				ull |= 1ULL << 52;
			}
			else		// 0.0 or denormal floating-point number
#ifndef OPTIMIZE
				if (ull)
				{	// unbiased exponent
					exponent = 1 - 1023;
					// normalize significand
					while (ull < 1ULL << 52)
						ull += ull, exponent--;
				}
#else
				// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

				if (_BitScanReverse(&ul, (unsigned long) (ull >> 32)))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - 32 - ul);
					// normalize significand
					ull = __ll_lshift(ull, 52 - 32 - ul);
				}
				else if (_BitScanReverse(&ul, (unsigned long) ull))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - ul);
					// normalize significand
					ull = __ll_lshift(ull, 52 - 31);
					ull = __ll_lshift(ull, 31 - ul);
				}
#endif // OPTIMIZE
					// convert significand to decimal, one digit per byte,
					//  least significant digit in decimal[length - 1],
					//   most significant digit in decimal[0],
					//    decimal point after most significant digit:
					//     0x0                 = 0[.]
					//     0x1[.]0000000000000 = 1[.]
					//     0x1[.]0000000000001 = 1[.]0000000000000002220446049250313080847263336181640625
					//     0x1[.]8000000000000 = 1[.]5
					//     0x1[.]FFFFFFFFFFFFF = 1[.]9999999999999997779553950749686919152736663818359375
			do
				decimal[length++] = (char) (ull >> 52);
			while (ull &= ~0ULL >> 12, ull *= 10);

					// decimal point past unhidden bit
					//  (also decimal exponent)
			point = 0;
					// multiply by 2**exponent
			while (exponent > 0)
			{
				index = length;
				digit = 0U;

				if (decimal[0] < 5U)
				{
					do
						digit += decimal[index - 1] * 2U,
						decimal[index - 1] = digit % 10U, digit /= 10U;
					while (--index > 0);
				}
				else	// decimal overflow: move decimal point and all digits
				{	//  one place right, insert new non-zero leading digit
					point++;
					length++;

					do
						digit += decimal[index - 1] * 2U,
						decimal[index] = digit % 10U, digit /= 10U;
					while (--index > 0);

					decimal[0] = digit;
				}

				exponent--;
			}
					// divide by 2**-exponent
			while (exponent < 0)
			{
				index = 0;

				if (decimal[length - 1] % 2U)
					decimal[length++] = 0U;

				if (decimal[0] > 1U)
				{
					digit = 0U;

					do
						digit = digit * 10U + decimal[index],
						decimal[index] = digit / 2U, digit %= 2U;
					while (++index < length);
				}
				else	// decimal underflow: discard new leading 0, move
				{	//  decimal point and all digits one place left
					point--;
					length--;
					digit = decimal[0];

					do
						digit = digit * 10U + decimal[index + 1],
						decimal[index] = digit / 2U, digit %= 2U;
					while (++index < length);
				}

				exponent++;
			}

			if (precision < 0)
				precision = 6;

			extra = 0;

			if (precision > length - 1)
			{		// number of additional trailing zeroes
				extra = precision - length + 1;
				precision = length - 1;
				hash = 1;
			}
			else if (precision < length - 1
			      && (decimal[precision + 1] > 5U
			       || decimal[precision + 1] == 5U && decimal[precision] % 2U
			       || tiebreak(decimal, precision + 2, length)))
			{		// round to nearest, ties to even
#if 0
				index = precision + 1;
				digit = 5U;

				do
					digit += decimal[index],
					decimal[index] = digit % 10U;
				while (digit /= 10U && index-- > 0);

				if (digit)
				{	// decimal overflow: move decimal point right
					point++;
					// set leading digit
					decimal[0] = digit;
				}
#else
				index = precision;

				while (index >= 0 && decimal[index] == 9U)
					decimal[index--] = 0U;

				if (index < 0)
				{	// decimal overflow: move decimal point right
					point++;
					// set leading digit
					decimal[0] = 1U;
				}
				else
					decimal[index] += 1U;
#endif
			}
					// decimal exponent
			if (point < 0)
			{
				point = -point;
				size = '-';
			}
			else if (point > 0)
				size = '+';
		//	else
		//		size = '\0';

			do		// decimal exponent
				scratch[exponent++] = '0' + (char) (point % 10U);
			while (point /= 10U);
					// exponent sign
			if (size != '\0')
				scratch[exponent++] = size;
					// exponent prefix 'e' or 'E'
			scratch[exponent++] = *format;
					// index of exponent prefix (trailing '0' padding)
			length = exponent;
					// decimal point
			if (precision > 0 || hash > 0)
				scratch[length++] = '.';
					// most significant digit
			scratch[length++] = '0' + decimal[0];
					// index of most significant digit (leading '0' padding)
			index = length;
					// sign
			if (sign != '\0')
				scratch[length++] = sign;

		DOUBLE:	if (width > length + precision + extra)
				width -= length + precision + extra;
			else
				width = 0;

			if (left)
			{		// sign, prefix, most significant digit and point
				while (length > exponent && output < end)
					*output++ = scratch[--length];

				if (ansi == NULL)
					while (point < precision && output < end)
						*output++ = '0' + decimal[++point];
				else
					while (point < precision && output < end)
						*output++ = ansi[decimal[++point]];
#ifndef OPTIMIZE
				while (extra-- > 0 && output < end)
					*output++ = '0';
#else
				if (extra > end - output)
					extra = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, '0', extra);

				output += extra;
#endif // OPTIMIZE
				while (length > 0 && output < end)
					*output++ = scratch[--length];

				goto RIGHT;
			}

			if (pad == '0')
			{		// sign and prefix
				while (length > index && output < end)
					*output++ = scratch[--length];
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = '0';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, '0', width);

				output += width;
#endif // OPTIMIZE
			}
			else
			{
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = ' ';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
#endif // OPTIMIZE
				while (length > index && output < end)
					*output++ = scratch[--length];
			}

			while (length > exponent && output < end)
				*output++ = scratch[--length];

			if (ansi == NULL)
				while (point < precision && output < end)
					*output++ = '0' + decimal[++point];
			else
				while (point < precision && output < end)
					*output++ = ansi[decimal[++point]];
#ifndef OPTIMIZE
			while (extra-- > 0 && output < end)
				*output++ = '0';
#else
			if (extra > end - output)
				extra = end - output;

			// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

			memset(output, '0', extra);

			output += extra;
#endif // OPTIMIZE
			while (length > 0 && output < end)
				*output++ = scratch[--length];
		}
		else			// return negative index of invalid character
		INVALID:		//  in format string
			return (int) (start - format);

		format++;
	}

	if (output != NULL)
		*output = '\0';

	return (int) (output - begin);	// return length of output string
}

// https://msdn.microsoft.com/en-us/library/2ts7cx93.aspx

int	__cdecl	snprintf(char *output, size_t count, const char *format, ...)
{
	int	result;

	// https://msdn.microsoft.com/en-us/library/kb57fad8.aspx

	va_list	input;
	va_start(input, format);

	// https://msdn.microsoft.com/en-us/library/1kt27hek.aspx

	result = vsnprintf(output, count, format, input);

	va_end(input);

	return result;
}

// https://msdn.microsoft.com/en-us/library/28d5ce15.aspx

__declspec(safebuffers)
int	__cdecl	vswprintf(wchar_t *output, size_t count, wchar_t const *format, va_list input)
{
	wchar_t	const * const start = format;
	wchar_t	const * const begin = output;
	wchar_t	const * const end = (count < 2) ? output : output + count - 2;
	wchar_t	const *unicode;
	char	const *ansi;
	char	decimal[768];		// digits from floating-point conversion:
					//  2.0**-x has x fractional decimal digits;
					//  0x1p-1074 = 4.9406564584124654...e-324
					//   has 323 significant zeroes after the decimal point,
					//    followed by 751 digits;
					//  0x0.FFFFFFFFFFFFFp-1022 = 2.2250738585072008...e-308
					//   has 307 significant zeroes after the decimal point,
					//    followed by 767 digits;
					//  0x1p+1024 = 1.7976931348623159...e+308 has 309 digits
	char	digit;
	char	cb;
	wchar_t	cw;
	wchar_t	pad;
	wchar_t	sign;
	wchar_t	size;
	wchar_t	scratch[24];		// characters from integer conversion:
					//  ~0ULL has 22 octal characters
	int	length;			// number of characters from conversion
	int	index;			// index of first digit from conversion
	int	left;
	int	hash;
	int	blank;
//	int	group;
	int	width;
	int	precision;
	int	exponent;
	int	point;
	int	extra;

	unsigned long		ul;
	unsigned long long	ull;

	if (UCS2['\t'] == L'\0')	// ANSI to UTF-16LE conversion table not initialized?

		// https://msdn.microsoft.com/en-us/library/ms776413.aspx
		// https://msdn.microsoft.com/en-us/library/dd319072.aspx

		if (MultiByteToWideChar(CP_ACP,
		                        MB_ERR_INVALID_CHARS | MB_PRECOMPOSED,
		                        ANSI, 256,
		                        UCS2, 256) != 256)
			return ~2147483647;

	for (;;)
	{
		while (*format != L'%'	// copy format string up to next conversion specifier
		    && *format != L'\0'	//  or its end to output string
		    && output < end)
			*output++ = *format++;

		if (*format == L'\0')	// format string processed completely?
			break;

		if (*++format == L'%')	// double L'%' in format string?
		{
			if (output < end)
				*output++ = *format++;
			continue;
		}
					// conversion specifier in format string
		pad = L' ';
		sign = L'\0';
		size = L'\0';
		left = 0;
		hash = 0;
		blank = 0;
	//	group = 0;
		width = 0;
		length = 0;
		precision = -1;

		do			// flag(s) in format string?
			if (*format == L'-')
				left = 1;
			else if (*format == L'+')
				sign = L'+';
			else if (*format == L'0')
				pad = L'0';
			else if (*format == L' ')
				blank = 1;
			else if (*format == L'#')
				hash = 1;
		//	else if (*format == L'\'')
		//		group = 1;
			else
				break;
		while (*++format != L'\0');

		if (blank && sign == L'\0')
			sign = L' ';

		if (*format >= L'0'	// field width in format string?
		 && *format <= L'9')
		{
			for (width = *format++ - L'0';
			     *format >= L'0' && *format <= L'9';
			     width *= 10, width += *format++ - L'0')
				continue;

			if (width < 0)
				width = 2147483647;
		}
		else if (*format == L'*')
		{
			format++;

			width = va_arg(input, int);

			if (width < 0)
			{
				if (width == ~2147483647)
					width = 0;
				else
					width = -width;

				left = 1;
			}
		}

	//	if (left)
	//		pad = L' ';

		if (*format == L'.')	// precision in format string?
		{
			pad = L' ';

			if (*++format == L'*')
			{
				format++;

				precision = va_arg(input, int);

			//	if (precision < 0)
			//		precision = -1;
			}
			else if (*format >= L'0' && *format <= L'9')
			{
				for (precision = *format++ - L'0';
				     *format >= L'0' && *format <= L'9';
				     precision *= 10, precision += *format++ - L'0')
					continue;

				if (precision < 0)
					precision = 2147483647;
			}
			else
				precision = 0;
		}

		if (*format == L'j'	// size modifier in format string?
		 || *format == L't'
		 || *format == L'z')
			size = *format++;
		else if (*format == L'h'
		      || *format == L'l')
		{
			size = *format++;

			if (size == *format)
			{
				format++;
					// encode L"hh" as L'H' and L"ll" as L'L'
				size ^= L'a' ^ L'A';
			}
		}

		if (*format == L'd'	// conversion type in format string?
		 || *format == L'i')
		{			// signed decimal integer:
					//  [ ' ' ... | '+' | '-' ] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == L'j'
			 || size == L'L')
				(long long) ull = va_arg(input, long long);
			else if (size == L'h')
				(long long) ull = va_arg(input, short);
			else if (size == L'H')
				(long long) ull = va_arg(input, char);
			else // L'l', L't', L'z', L'\0'
				(long long) ull = va_arg(input, int);

			if ((long long) ull < 0)
			{
				ull = 0 - ull;
				sign = L'-';
			}
					// convert to decimal characters,
					//  least significant digit in scratch[0],
					//   most significant digit in scratch[length - 1]
#ifndef OPTIMIZE
			do
			{
				ull = __ull_divrem(ull, 10, &ul);
				scratch[length++] = L'0' + (wchar_t) ul;
			} while (ull);
#else
			do		// least number of 6432-bit divisions
			{
				ull = __ull_divrem(ull, 1000000000, &ul);

				if (ull > 0)
				{
					index = length + 9;
#if 0
					do
						scratch[length++] = L'0' + (wchar_t) (ul % 10);
					while (ul /= 10);

					while (length < index)
						scratch[length++] = L'0';
#else
					// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

					wmemset(scratch + length, L'0', 9);

					do
						scratch[length++] = L'0' + (wchar_t) (ul % 10);
					while (ul /= 10);

					length = index;
#endif
				}
				else if (length == 0)
					do
						scratch[length++] = L'0' + (wchar_t) (ul % 10);
					while (ul /= 10);
				else
					while (ul > 0)
						scratch[length++] = L'0' + (wchar_t) (ul % 10), ul /= 10;
			} while (ull);
#endif // OPTIMIZE
					// index of decimal integer (leading L'0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			if (sign != L'\0')
				scratch[length++] = sign;

			goto NUMBER;
		}
		else if (*format == L'u')
		{			// unsigned decimal integer:
					//  [{ ' ' | '0' } ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == L'j'
			 || size == L'L')
				ull = va_arg(input, unsigned long long);
			else if (size == L'h')
				ull = va_arg(input, unsigned short);
			else if (size == L'H')
				ull = va_arg(input, unsigned char);
			else // L'l', L't', L'z', L'\0'
				ull = va_arg(input, unsigned int);

					// convert to decimal characters,
					//  least significant digit in scratch[0],
					//   most significant digit in scratch[length - 1]
#ifndef OPTIMIZE
			do
			{
				ull = __ull_divrem(ull, 10, &ul);
				scratch[length++] = L'0' + (wchar_t) ul;
			} while (ull);
#else
			do		// least number of 6432-bit divisions
			{
				ull = __ull_divrem(ull, 1000000000, &ul);

				if (ull > 0)
				{
					index = length + 9;
#if 0
					do
						scratch[length++] = L'0' + (wchar_t) (ul % 10);
					while (ul /= 10);

					while (length < index)
						scratch[length++] = L'0';
#else
					// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

					wmemset(scratch + length, L'0', 9);

					do
						scratch[length++] = L'0' + (wchar_t) (ul % 10);
					while (ul /= 10);

					length = index;
#endif
				}
				else if (length == 0)
					do
						scratch[length++] = L'0' + (wchar_t) (ul % 10);
					while (ul /= 10);
				else
					while (ul > 0)
						scratch[length++] = L'0' + (wchar_t) (ul % 10), ul /= 10;
			} while (ull);
#endif // OPTIMIZE
					// index of decimal integer (leading L'0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			goto NUMBER;
		}
		else if (*format == L'o')
		{			// unsigned octal integer:
					//  [{ ' ' | '0' } ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == L'j'
			 || size == L'L')
				ull = va_arg(input, unsigned long long);
			else if (size == L'h')
				ull = va_arg(input, unsigned short);
			else if (size == L'H')
				ull = va_arg(input, unsigned char);
			else // L'l', L't', L'z', L'\0'
				ull = va_arg(input, unsigned int);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = L'0' + (wchar_t) (ull % 8);
			while (ull /= 8);
					// index of octal integer (leading L'0' padding)
			index = length;

			if (hash)
#if 0
				if (precision >= 0)
					precision++;
#else
				precision += (precision >= 0);
#endif
			if (precision > length)
				precision -= length;
			else
				precision = 0;
					// prefix L'0'
			if (hash)
				scratch[length++] = L'0';

			goto NUMBER;
		}
#if 0
		else if (*format == L'x'
		      || *format == L'X')
#else
		else if ((*format | L'x' ^ L'X') == L'x')
#endif
		{			// unsigned hexadecimal integer:
					//  [ ' ' ... ] [ '0' { 'x' | 'X' }] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]

			unicode = L"0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ L'X') / (L'x' ^ L'X');

			if (size == L'j'
			 || size == L'L')
				ull = va_arg(input, unsigned long long);
			else if (size == L'h')
				ull = va_arg(input, unsigned short);
			else if (size == L'H')
				ull = va_arg(input, unsigned char);
			else // L'l', L't', L'z', L'\0'
				ull = va_arg(input, unsigned int);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = unicode[ull % 16];
			while (ull /= 16);
					// index of hexadecimal integer (leading L'0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			if (hash)
			{		// prefix L"0x" or L"0X"
				scratch[length++] = *format;
				scratch[length++] = L'0';
			}

		NUMBER:	if (width > precision + length)
				width -= precision + length;
			else
				width = 0;

			if (left)
				while (length > index && output < end)
					*output++ = scratch[--length];
			else if (pad == L'0')
			{
				while (length > index && output < end)
					*output++ = scratch[--length];
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = L'0';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L'0', width);

				output += width;
#endif // OPTIMIZE
			}
			else
			{
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = L' ';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
#endif // OPTIMIZE
				while (length > index && output < end)
					*output++ = scratch[--length];
			}
#ifndef OPTIMIZE
			while (precision-- > 0 && output < end)
				*output++ = L'0';
#else
			if (precision > end - output)
				precision = end - output;

			// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

			wmemset(output, L'0', precision);

			output += precision;
#endif // OPTIMIZE
			while (length > 0 && output < end)
				*output++ = scratch[--length];
#ifndef OPTIMIZE
		//	if (left)
		RIGHT:		while (width-- > 0 && output < end)
					*output++ = L' ';
#else
		RIGHT:	if (left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
			}
#endif // OPTIMIZE
		}
#if 0
		else if (*format == L'p'
		      || *format == L'P')
#else
		else if ((*format | L'p' ^ L'P') == L'p')
#endif
		{			// unsigned hexadecimal pointer:
					//  [ ' ' ... ] [ '0' { 'x' | 'X' }] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]
			if (size != L'\0'
			 || sign != L'\0'
			 || pad != L' '
			 || precision != -1)
				goto INVALID;

			unicode = L"0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ L'P') / (L'p' ^ L'P');

			ul = va_arg(input, unsigned long);

			if (ul == 0)
				hash = 0;
			do
				scratch[length++] = unicode[ul % 16];
			while (ul /= 16);

			while (length < 2 * sizeof(ul))
				scratch[length++] = L'0';

			if (hash)
			{		// prefix L"0x" or L"0X"
				scratch[length++] = *format ^ L'P' | L'X';
				scratch[length++] = L'0';
			}

			if (!left)
				while (width-- > length && output < end)
					*output++ = L' ';

			while (length > 0 && output < end)
				*output++ = scratch[--length];

			goto RIGHT;
		}
		else if (*format == L'c')
		{			// ANSI or UNICODE character:
					//  [ ' ' ... ] [ <character> ] [ ' ' ... ]
			if (precision != -1)
				goto INVALID;

			if (size == L'h')
				goto C;	// ANSI character?

			if (size != L'\0'
			 && size != L'l')
				goto INVALID;

		c:	cw = va_arg(input, wchar_t);

		COMMON:	if (cw != L'\0')
#if 0
				if (width > 0)
					width--;
#else
				width -= (width > 0);
#endif
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = L' ';
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
			}
#endif // OPTIMIZE
			if (cw != L'\0' && output < end)
				*output++ = cw;

			goto RIGHT;
		}
		else if (*format == L'C')
		{			// ANSI or UNICODE character:
					//  [ ' ' ... ] [ <character> ] [ ' ' ... ]
			if (precision != -1)
				goto INVALID;

			if (size == L'l')
				goto c;	// UNICODE character?

			if (size != L'\0'
			 && size != L'h')
				goto INVALID;

		C:	cb = va_arg(input, char);
			cw = UCS2[cb];

			goto COMMON;
		}
		else if (*format == L's')
		{			// ANSI or UNICODE character string:
					//  [ ' ' ... ] [ <character> ... ] [ ' ' ... ]
			if (size == L'h')
				goto S;	// ANSI character string?

			if (size != L'\0'
			 && size != L'l')
				goto INVALID;

		s:	unicode = va_arg(input, wchar_t *);

			if (unicode == NULL)
				unicode = L"(null)";
#if 0
			// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

			length = wcslen(unicode);

			if (precision >= 0 && length > precision)
				length = precision;
#else
			if (precision < 0)

				// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

				length = wcslen(unicode);
			else
				for (length = 0; length < precision; length++)
					if (unicode[length] == L'\0')
						break;
#endif
		STRING:	if (width > length)
				width -= length;
			else
				width = 0;
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = L' ';

			while (length-- > 0 && output < end)
				*output++ = *unicode++;
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
			}

			if (length > end - output)
				length = end - output;

			// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

			wmemcpy(output, unicode, length);

			output += length;
#endif // OPTIMIZE
			goto RIGHT;
		}
		else if (*format == L'S')
		{			// ANSI or UNICODE character string:
					//  [ ' ' ... ] [ <character> ... ] [ ' ' ... ]
			if (size == L'l')
				goto s;	// UNICODE character string?

			if (size != L'\0'
			 && size != L'h')
				goto INVALID;

		S:	ansi = va_arg(input, char *);

			if (ansi == NULL)
				ansi = "(null)";
#if 0
			// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

			length = strlen(ansi);

			if (precision >= 0 && length > precision)
				length = precision;
#else
			if (precision < 0)

				// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

				length = strlen(ansi);
			else
				for (length = 0; length < precision; length++)
					if (ansi[length] == '\0')
						break;
#endif
			if (width > length)
				width -= length;
			else
				width = 0;
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = L' ';
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
			}
#endif // OPTIMIZE
			while (length-- > 0 && output < end)
				*output++ = UCS2[*ansi++];

			goto RIGHT;
		}
		else if (*format == L'n')
		{			// assign character count
			if (size == L'j'
			 || size == L'L')
				*va_arg(input, long long *) = (long long) (output - begin);
			else if (size == L'h')
				*va_arg(input, short *) = (short) (output - begin);
			else if (size == L'H')
				*va_arg(input, char *) = (char) (output - begin);
			else
				*va_arg(input, int *) = (int) (output - begin);
		}
#if 0
		else if (*format == L'a'
		      || *format == L'A')
#else
		else if ((*format | L'a' ^ L'A') == L'a')
#endif
		{			// hexadecimal double-precision floating-point:
					//  [ ' ' ... | '+' | '-' ] '0' { 'x' | 'X' } [ '0' ... ] <digit> [ '.' <digit> ... ] { 'p' | 'P' } { '+' | '-' } <digit> ... [ ' ' ... ]
			if (size != L'\0')
				goto INVALID;

			unicode = L"0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ L'A') / (L'a' ^ L'A');
			ull = va_arg(input, unsigned long long);

			if ((long long) ull < 0)
			{		// strip sign bit
				ull &= ~0ULL >> 1;
				sign = L'-';
			}
					// biased exponent
			exponent = (int) (ull >> 52);

			if (exponent == 2047)
			{
				if (sign != L'\0')
					scratch[length++] = sign;
#ifndef OPTIMIZE
				if (ull & ~0ULL >> 12)
				{
					scratch[length++] = *format ^ L'A' | L'N';
					scratch[length++] = *format ^ L'A' | L'A';
					scratch[length++] = *format ^ L'A' | L'N';
				}
				else
				{
					scratch[length++] = *format ^ L'A' | L'I';
					scratch[length++] = *format ^ L'A' | L'N';
					scratch[length++] = *format ^ L'A' | L'F';
				}

				scratch[length] = L'\0';
#else
				// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

				if (ull & ~0ULL >> 12)
					wmemcpy(scratch + length, *format == L'a' ? L"nan" : L"NAN", sizeof("NAN"));
				else
					wmemcpy(scratch + length, *format == L'a' ? L"inf" : L"INF", sizeof("INF"));

				length += 3;
#endif // OPTIMIZE
				unicode = scratch;

				goto STRING;
			}
			else if (exponent > 0)
			{		// unbiased exponent
				exponent -= 1023;
					// strip exponent, insert hidden bit
				ull &= ~0ULL >> 12;
				ull |= 1ULL << 52;
			}
			else		// 0.0 or denormal floating-point number
#ifndef OPTIMIZE
				if (ull)
				{	// unbiased exponent
					exponent = 1 - 1023;
					// normalize significand
					while (ull < 1ULL << 52)
						ull += ull, exponent--;
				}
#else
				// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

				if (_BitScanReverse(&ul, (unsigned long) (ull >> 32)))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - 32 - ul);
					// normalize significand
					ull = __ll_lshift(ull, 52 - 32 - ul);
				}
				else if (_BitScanReverse(&ul, (unsigned long) ull))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - ul);
					// normalize significand
					ull = __ll_lshift(ull, 52 - 31);
					ull = __ll_lshift(ull, 31 - ul);
				}
#endif // OPTIMIZE
					// convert significand to hexadecimal, one digit per byte,
					//  least significant digit in decimal[length - 1],
					//   most significant digit in decimal[0],
					//    hexadecimal point after most significant digit
			do
				decimal[length++] = (char) (ull >> 52);
			while (ull &= ~0ULL >> 12, ull <<= 4);

			point = 0;
			extra = 0;

			if (precision < 0)
				precision = length - 1;
			else if (precision > length - 1)
			{		// number of additional trailing zeroes
				extra = precision - length + 1;
				precision = length - 1;
				hash = 1;
			}
			else if (precision < length - 1
			      && (decimal[precision + 1] > 8U
			       || decimal[precision + 1] == 8U && decimal[precision] % 2U
			       || tiebreak(decimal, precision + 2, length)))
			{		// round to nearest, ties to even
#if 0
				index = precision + 1;
				digit = 8U;

				do
					digit += decimal[index],
					decimal[index] = digit % 16U;
				while (digit /= 16U && index-- > 0);

				if (digit)
				{	// hexadecimal overflow: increment exponent
					exponent++;
					// set leading digit
					decimal[0] = digit;
				}
#else
				index = precision;

				while (index >= 0 && decimal[index] == 15U)
					decimal[index--] = 0U;

				if (index < 0)
				{	// hexadecimal overflow: increment exponent
					exponent++;
					// set leading digit
					decimal[0] = 1U;
				}
				else
					decimal[index] += 1U;
#endif
			}
					// hexadecimal exponent
			if (exponent < 0)
			{
				exponent = -exponent;
				size = L'-';
			}
			else if (exponent > 0)
				size = L'+';
		//	else
		//		size = L'\0';

			length = 0;

			do		// hexadecimal exponent
				scratch[length++] = L'0' + (wchar_t) (exponent % 10U);
			while (exponent /= 10U);
					// exponent sign
			if (size != L'\0')
				scratch[length++] = size;
					// exponent prefix L'p' or L'P'
			scratch[length++] = *format ^ L'A' | L'P';
					// index of exponent prefix (trailing L'0' padding)
			exponent = length;
					// hexadecimal point
			if (precision > 0 || hash > 0)
				scratch[length++] = L'.';
					// most significant digit
			scratch[length++] = L'0' + decimal[0];
					// index of most significant digit (leading L'0' padding)
			index = length;
					// prefix L"0x" or L"0X"
			scratch[length++] = *format ^ L'A' | L'X';
			scratch[length++] = L'0';
					// sign
			if (sign != L'\0')
				scratch[length++] = sign;

			goto DOUBLE;
		}
#if 0
		else if (*format == L'e'
		      || *format == L'E')
#else
		else if ((*format | L'e' ^ L'E') == L'e')
#endif
		{			// decimal double-precision floating-point:
					//  [ ' ' ... | '+' | '-' ] [ '0' ... ] <digit> [ '.' <digit> ... [ '0' ... ]] { 'e' | 'E' } [ '+' | '-' ] <digit> ... [ ' ' ... ]
			if (size != L'\0')
				goto INVALID;

			unicode = NULL;

			ull = va_arg(input, unsigned long long);

			if ((long long) ull < 0)
			{		// strip sign bit
				ull &= ~0ULL >> 1;
				sign = L'-';
			}
					// biased exponent
			exponent = (int) (ull >> 52);

			if (exponent == 2047)
			{
				if (sign != L'\0')
					scratch[length++] = sign;
#ifndef OPTIMIZE
				if (ull & ~0ULL >> 12)
				{
					scratch[length++] = *format ^ L'E' | L'N';
					scratch[length++] = *format ^ L'E' | L'A';
					scratch[length++] = *format ^ L'E' | L'N';
				}
				else
				{
					scratch[length++] = *format ^ L'E' | L'I';
					scratch[length++] = *format ^ L'E' | L'N';
					scratch[length++] = *format ^ L'E' | L'F';
				}

				scratch[length] = L'\0';
#else
				// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

				if (ull & ~0ULL >> 12)
					wmemcpy(scratch + length, *format == L'e' ? L"nan" : L"NAN", sizeof("NAN"));
				else
					wmemcpy(scratch + length, *format == L'e' ? L"inf" : L"INF", sizeof("INF"));

				length += 3;
#endif // OPTIMIZE
				unicode = scratch;

				goto STRING;
			}
			else if (exponent > 0)
			{		// unbiased exponent
				exponent -= 1023;
					// strip exponent, insert hidden bit
				ull &= ~0ULL >> 12;
				ull |= 1ULL << 52;
			}
			else		// 0.0 or denormal floating-point number
#ifndef OPTIMIZE
				if (ull)
				{	// unbiased exponent
					exponent = 1 - 1023;
					// normalize significand
					while (ull < 1ULL << 52)
						ull += ull, exponent--;
				}
#else
				// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

				if (_BitScanReverse(&ul, (unsigned long) (ull >> 32)))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - 32 - ul);
					// normalize significand
					ull = __ll_lshift(ull, 52 - 32 - ul);
				}
				else if (_BitScanReverse(&ul, (unsigned long) ull))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - ul);
					// normalize significand
					ull = __ll_lshift(ull, 52 - 31);
					ull = __ll_lshift(ull, 31 - ul);
				}
#endif // OPTIMIZE
					// convert significand to decimal, one digit per byte,
					//  least significant digit in decimal[length - 1],
					//   most significant digit in decimal[0],
					//    decimal point after most significant digit:
					//     0x0                 = 0[.]
					//     0x1[.]0000000000000 = 1[.]
					//     0x1[.]0000000000001 = 1[.]0000000000000002220446049250313080847263336181640625
					//     0x1[.]8000000000000 = 1[.]5
					//     0x1[.]FFFFFFFFFFFFF = 1[.]9999999999999997779553950749686919152736663818359375
			do
				decimal[length++] = (char) (ull >> 52);
			while (ull &= ~0ULL >> 12, ull *= 10);

					// decimal point past unhidden bit
					//  (also decimal exponent)
			point = 0;
					// multiply by 2**exponent
			while (exponent > 0)
			{
				index = length;
				digit = 0U;

				if (decimal[0] < 5U)
				{
					do
						digit += decimal[index - 1] * 2U,
						decimal[index - 1] = digit % 10U, digit /= 10U;
					while (--index > 0);
				}
				else	// decimal overflow: move decimal point and all digits
				{	//  one place right, insert new non-zero leading digit
					point++;
					length++;

					do
						digit += decimal[index - 1] * 2U,
						decimal[index] = digit % 10U, digit /= 10U;
					while (--index > 0);

					decimal[0] = digit;
				}

				exponent--;
			}
					// divide by 2**-exponent
			while (exponent < 0)
			{
				index = 0;

				if (decimal[length - 1] % 2U)
					decimal[length++] = 0U;

				if (decimal[0] > 1U)
				{
					digit = 0U;

					do
						digit = digit * 10U + decimal[index],
						decimal[index] = digit / 2U, digit %= 2U;
					while (++index < length);
				}
				else	// decimal underflow: discard new leading 0, move
				{	//  decimal point and all digits one place left
					point--;
					length--;
					digit = decimal[0];

					do
						digit = digit * 10U + decimal[index + 1],
						decimal[index] = digit / 2U, digit %= 2U;
					while (++index < length);
				}

				exponent++;
			}

			if (precision < 0)
				precision = 6;

			extra = 0;

			if (precision > length - 1)
			{		// number of additional trailing zeroes
				extra = precision - length + 1;
				precision = length - 1;
				hash = 1;
			}
			else if (precision < length - 1
			      && (decimal[precision + 1] > 5U
			       || decimal[precision + 1] == 5U && decimal[precision] % 2U
			       || tiebreak(decimal, precision + 2, length)))
			{		// round to nearest, ties to even
#if 0
				index = precision + 1;
				digit = 5U;

				do
					digit += decimal[index],
					decimal[index] = digit % 10U;
				while (digit /= 10U && index-- > 0);

				if (digit)
				{	// decimal overflow: move decimal point right
					point++;
					// set leading digit
					decimal[0] = digit;
				}
#else
				index = precision;

				while (index >= 0 && decimal[index] == 9U)
					decimal[index--] = 0U;

				if (index < 0)
				{	// decimal overflow: move decimal point right
					point++;
					// set leading digit
					decimal[0] = 1U;
				}
				else
					decimal[index] += 1U;
#endif
			}
					// decimal exponent
			if (point < 0)
			{
				point = -point;
				size = L'-';
			}
			else if (point > 0)
				size = L'+';
		//	else
		//		size = L'\0';

			do		// decimal exponent
				scratch[exponent++] = L'0' + (wchar_t) (point % 10U);
			while (point /= 10U);
					// exponent sign
			if (size != L'\0')
				scratch[exponent++] = size;
					// exponent prefix L'e' or L'E'
			scratch[exponent++] = *format;
					// index of exponent prefix (trailing L'0' padding)
			length = exponent;
					// decimal point
			if (precision > 0 || hash > 0)
				scratch[length++] = L'.';
					// most significant digit
			scratch[length++] = L'0' + decimal[0];
					// index of most significant digit (leading L'0' padding)
			index = length;
					// sign
			if (sign != L'\0')
				scratch[length++] = sign;

		DOUBLE:	if (width > length + precision + extra)
				width -= length + precision + extra;
			else
				width = 0;

			if (left)
			{		// sign, prefix, most significant digit and point
				while (length > exponent && output < end)
					*output++ = scratch[--length];

				if (unicode == NULL)
					while (point < precision && output < end)
						*output++ = L'0' + decimal[++point];
				else
					while (point < precision && output < end)
						*output++ = unicode[decimal[++point]];
#ifndef OPTIMIZE
				while (extra-- > 0 && output < end)
					*output++ = L'0';
#else
				if (extra > end - output)
					extra = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L'0', extra);

				output += extra;
#endif // OPTIMIZE
				while (length > 0 && output < end)
					*output++ = scratch[--length];

				goto RIGHT;
			}

			if (pad == L'0')
			{		// sign and prefix
				while (length > index && output < end)
					*output++ = scratch[--length];
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = L'0';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L'0', width);

				output += width;
#endif // OPTIMIZE
			}
			else
			{
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = L' ';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
#endif // OPTIMIZE
				while (length > index && output < end)
					*output++ = scratch[--length];
			}

			while (length > exponent && output < end)
				*output++ = scratch[--length];

			if (unicode == NULL)
				while (point < precision && output < end)
					*output++ = L'0' + decimal[++point];
			else
				while (point < precision && output < end)
					*output++ = unicode[decimal[++point]];
#ifndef OPTIMIZE
			while (extra-- > 0 && output < end)
				*output++ = L'0';
#else
			if (extra > end - output)
				extra = end - output;

			// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

			wmemset(output, L'0', extra);

			output += extra;
#endif // OPTIMIZE
			while (length > 0 && output < end)
				*output++ = scratch[--length];
		}
		else			// return negative index of invalid character
		INVALID:		//  in format string
			return (int) (start - format);

		format++;
	}

	if (output != NULL)
		*output = L'\0';

	return (int) (output - begin);	// return length of output string
}

// https://msdn.microsoft.com/en-us/library/ybk95axf.aspx

int	__cdecl	swprintf(wchar_t *output, size_t count, const wchar_t *format, ...)
{
	int	result;

	// https://msdn.microsoft.com/en-us/library/kb57fad8.aspx

	va_list	input;
	va_start(input, format);

	// https://msdn.microsoft.com/en-us/library/28d5ce15.aspx

	result = vswprintf(output, count, format, input);

	va_end(input);

	return result;
}
<<NOKEEP

I386-DLG.OBJ: I386-MSC.MAK
	ML.EXE /Brepro /c /Fo$@ /nologo /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; https://msdn.microsoft.com/en-us/library/ss9fh0d6.aspx

	.model	flat, C

; https://msdn.microsoft.com/en-us/library/7cycf5bc.aspx

	extern	__pfnDliDefaultHook2 :ptr proc
	extern	__pfnDliFailureHook2 (__pfnDliDefaultHook2) :ptr proc
	extern	__pfnDliNotifyHook2 (__pfnDliDefaultHook2) :ptr proc
	end
<<NOKEEP

I386-DLH.OBJ: I386-MSC.MAK
	CL.EXE /Brepro /c /Fo$@ /GAFwyz /nologo /O2txy /Tc<< /W4 /wd4201 /wd4204 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(lib, "kernel32.lib")
#if 0
#pragma comment(linker, "/ALTERNATENAME:___pfnDliFailureHook2=___pfnDliDefaultHook2")
#pragma comment(linker, "/ALTERNATENAME:___pfnDliNotifyHook2=___pfnDliDefaultHook2")
#endif
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

#define STRICT
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

#ifndef EXCEPTION_DELAY_LOAD_INVALID_PARAMETER
#define EXCEPTION_DELAY_LOAD_INVALID_PARAMETER	0xC06D0057
#endif

// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

#ifndef EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND
#define EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND	0xC06D007E
#endif

#ifndef EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND
#define EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND	0xC06D007F
#endif

extern	const	IMAGE_DOS_HEADER	__ImageBase;

typedef	DWORD	RVA;

// https://msdn.microsoft.com/en-us/library/f0fkfy9y.aspx
// https://msdn.microsoft.com/en-us/library/16b2dyk5.aspx

typedef	enum	dliNotify
{
	dliStartProcessing,
	dliNotePreLoadLibrary,
	dliNotePreGetProcAddress,
	dliFailLoadLib,
	dliFailGetProc,
	dliNoteEndProcessing
} dliNotify;

typedef	struct	DelayLoadDescr
{
	DWORD	dwAttributes;		// 1UL = all members are RVAs
	DWORD	dwDllName;
	DWORD	dwHMODULE;		// RVA of module handle
	DWORD	dwIAT;			// RVA of import address table
	DWORD	dwINT;			// RVA of import name table
	DWORD	dwBoundIAT;		// RVA of optional bound import address table
	DWORD	dwUnloadIAT;		// RVA of optional copy of original import address table
	DWORD	dwTimeStamp;
} DelayLoadDescr;

typedef	struct	DelayLoadProc
{
	BOOL	fImportByName;
	union
	{
		LPCSTR	szProcName;
		DWORD	dwOrdinal;
	};
} DelayLoadProc;

typedef	struct	DelayLoadInfo
{
	DWORD		cb;		// size of structure
	DelayLoadDescr	*pidd;		// raw form of data (everything is there)
	FARPROC		*ppfn;		// points to address of function to load
	LPCSTR		szDll;		// name of DLL
	DelayLoadProc	dlp;		// name or ordinal of function to load
	HMODULE		hmodCur;	// handle of DLL
	FARPROC		pfnCur;		// actual function that will be called
	DWORD		dwLastError;	// error received (if an error notification)
} DelayLoadInfo;

typedef	FARPROC	(WINAPI *PfnDliHook) (dliNotify, DelayLoadInfo *);

extern	PfnDliHook	__pfnDliNotifyHook2;
extern	PfnDliHook	__pfnDliFailureHook2;
const	PfnDliHook	__pfnDliDefaultHook2 = NULL;

// https://msdn.microsoft.com/en-us/library/2b054ds4.aspx

BOOL	WINAPI	__FUnloadDelayLoadedDLL2(LPCSTR szDll);

// https://msdn.microsoft.com/en-us/library/09t6x5ds.aspx

FARPROC	WINAPI	__delayLoadHelper2(DelayLoadDescr *lpDLD, FARPROC *lpfnIATEntry)
{
	HMODULE	hModule;

	// https://msdn.microsoft.com/en-us/library/ff1shykw.aspx

	HMODULE			*lpHMODULE = (HMODULE *) ((LPBYTE) &__ImageBase + lpDLD->dwHMODULE);
	IMAGE_THUNK_DATA	*lpINT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwINT);
	IMAGE_THUNK_DATA	*lpIAT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwIAT);
	IMAGE_THUNK_DATA	*lpBoundIAT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwBoundIAT);
	DWORD			dwEntry = (IMAGE_THUNK_DATA *) lpfnIATEntry - lpIAT;

	// NOTE: *lpfnIATEntry == lpIAT[dwEntry].u1.Function

	DelayLoadInfo	dli = {sizeof(DelayLoadInfo),
			       lpDLD,
			       lpfnIATEntry,
			       (LPCSTR) &__ImageBase + lpDLD->dwDllName,
			       {!IMAGE_SNAP_BY_ORDINAL(lpINT[dwEntry].u1.Ordinal),
			         IMAGE_SNAP_BY_ORDINAL(lpINT[dwEntry].u1.Ordinal)
			       ? IMAGE_ORDINAL(lpINT[dwEntry].u1.Ordinal)
			       : ((IMAGE_IMPORT_BY_NAME *) ((LPBYTE) &__ImageBase + lpINT[dwEntry].u1.AddressOfData))->Name},
			       *lpHMODULE,
			       (FARPROC) NULL,
			       ERROR_SUCCESS};

	if (__pfnDliNotifyHook2 != NULL)
	{
		dli.pfnCur = (*__pfnDliNotifyHook2)(dliStartProcessing, &dli);

		if (dli.pfnCur != NULL)
			goto SUCCESS;
	}

	if (lpDLD->dwAttributes != 0UL)
	{
		dli.dwLastError = ERROR_INVALID_PARAMETER;

		// https://msdn.microsoft.com/en-us/library/ms680552.aspx
		// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

		RaiseException(EXCEPTION_DELAY_LOAD_INVALID_PARAMETER,
		               EXCEPTION_NONCONTINUABLE,
		               1UL,
		               (DWORD_PTR *) &dli);

		goto FAILURE;
	}

	if (dli.hmodCur != NULL)	// module already loaded?
		goto ADDRESS;

	if (__pfnDliNotifyHook2 != NULL)
		dli.hmodCur = (HMODULE) (*__pfnDliNotifyHook2)(dliNotePreLoadLibrary, &dli);

	if (dli.hmodCur != NULL)	// module handle resolved by notification routine?
		goto ADDRESS;
#ifndef LOAD_LIBRARY_SEARCH_SYSTEM32
	// https://msdn.microsoft.com/en-us/library/ms684175.aspx

	dli.hmodCur = LoadLibraryA(dli.szDll);
#else
	// https://msdn.microsoft.com/en-us/library/ms684179.aspx

	dli.hmodCur = LoadLibraryExA(dli.szDll, NULL, LOAD_LIBRARY_SEARCH_SYSTEM32);
#endif
	if (dli.hmodCur == NULL)	// module not loaded?
	{
		// https://msdn.microsoft.com/en-us/library/ms679360.aspx

		dli.dwLastError = GetLastError();

		if (__pfnDliFailureHook2 != NULL)
			dli.hmodCur = (HMODULE) (*__pfnDliFailureHook2)(dliFailLoadLib, &dli);

		if (dli.hmodCur == NULL)
		{
			// https://msdn.microsoft.com/en-us/library/ms680552.aspx
			// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

			RaiseException(EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND,
			               EXCEPTION_NONCONTINUABLE,
			               1UL,
			               (DWORD_PTR *) &dli);

			goto FAILURE;
		}

		// https://msdn.microsoft.com/en-us/library/ms683609.aspx

		hModule = (HMODULE) InterlockedExchangePointer((LPVOID *) lpHMODULE, dli.hmodCur);

		if (hModule == dli.hmodCur)

			// https://msdn.microsoft.com/en-us/library/ms683152.aspx

			FreeLibrary(dli.hmodCur);
		else
			if (lpDLD->dwUnloadIAT != 0UL)
			{
				//...
			}
	}
ADDRESS:
	if (__pfnDliNotifyHook2 != NULL)
		dli.pfnCur = (*__pfnDliNotifyHook2)(dliNotePreGetProcAddress, &dli);

	if (dli.pfnCur != NULL)		// function address resolved by notification routine?
		goto SUCCESS;

	if (lpBoundIAT != NULL && lpDLD->dwTimeStamp != 0UL)
	{
		// https://msdn.microsoft.com/en-us/library/ms680336.aspx

		IMAGE_NT_HEADERS	*lpModule = (IMAGE_NT_HEADERS *) ((LPBYTE) dli.hmodCur + ((IMAGE_DOS_HEADER *) dli.hmodCur)->e_lfanew);

		if (lpModule->Signature == IMAGE_NT_SIGNATURE
		 && lpModule->FileHeader.TimeDateStamp == lpDLD->dwTimeStamp
		 && lpModule->OptionalHeader.ImageBase == dli.hmodCur)
		{
			dli.pfnCur = (FARPROC) lpBoundIAT[dwEntry].u1.Function;

			if (dli.pfnCur != NULL)
				goto SUCCESS;
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms683212.aspx

	dli.pfnCur = GetProcAddress(dli.hmodCur, dli.dlp.szProcName);

	if (dli.pfnCur != NULL)		// function address found?
		goto SUCCESS;

	// https://msdn.microsoft.com/en-us/library/ms679360.aspx

	dli.dwLastError = GetLastError();

	if (__pfnDliFailureHook2 != NULL)
		dli.pfnCur = (*__pfnDliFailureHook2)(dliFailGetProc, &dli);

	if (dli.pfnCur != NULL)		// function address resolved by failure routine?
		goto SUCCESS;

	// https://msdn.microsoft.com/en-us/library/ms680552.aspx
	// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

	RaiseException(EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND,
	               EXCEPTION_NONCONTINUABLE,
	               1UL,
	               (DWORD_PTR *) &dli);

	goto FAILURE;

SUCCESS:
	*lpfnIATEntry = dli.pfnCur;
FAILURE:
	if (__pfnDliNotifyHook2 != NULL)
		(*__pfnDliNotifyHook2)(dliNoteEndProcessing, &dli);

	return dli.pfnCur;
}
<<NOKEEP

I386-DLI.OBJ: I386-MSC.MAK
	CL.EXE /Brepro /c /Fo$@ /GAFwyz /nologo /O2txy /Tc<< /W4 /wd4201 /wd4204 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(lib, "kernel32.lib")
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

#define STRICT
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

#ifndef EXCEPTION_DELAY_LOAD_INVALID_PARAMETER
#define EXCEPTION_DELAY_LOAD_INVALID_PARAMETER	0xC06D0057
#endif

// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

#ifndef EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND
#define EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND	0xC06D007E
#endif

#ifndef EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND
#define EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND	0xC06D007F
#endif

extern	const	IMAGE_DOS_HEADER	__ImageBase;

typedef	DWORD	RVA;

// https://msdn.microsoft.com/en-us/library/f0fkfy9y.aspx
// https://msdn.microsoft.com/en-us/library/16b2dyk5.aspx

typedef	enum	dliNotify
{
	dliStartProcessing,
	dliNotePreLoadLibrary,
	dliNotePreGetProcAddress,
	dliFailLoadLib,
	dliFailGetProc,
	dliNoteEndProcessing
} dliNotify;

typedef	struct	DelayLoadDescr
{
	DWORD	dwAttributes;		// 1UL = all members are RVAs
	DWORD	dwDllName;
	DWORD	dwHMODULE;		// RVA of module handle
	DWORD	dwIAT;			// RVA of import address table
	DWORD	dwINT;			// RVA of import name table
	DWORD	dwBoundIAT;		// RVA of optional bound import address table
	DWORD	dwUnloadIAT;		// RVA of optional copy of original import address table
	DWORD	dwTimeStamp;
} DelayLoadDescr;

typedef	struct	DelayLoadProc
{
	BOOL	fImportByName;
	union
	{
		LPCSTR	szProcName;
		DWORD	dwOrdinal;
	};
} DelayLoadProc;

typedef	struct	DelayLoadInfo
{
	DWORD		cb;		// size of structure
	DelayLoadDescr	*pidd;		// raw form of data (everything is there)
	FARPROC		*ppfn;		// points to address of function to load
	LPCSTR		szDll;		// name of DLL
	DelayLoadProc	dlp;		// name or ordinal of function to load
	HMODULE		hmodCur;	// handle of DLL
	FARPROC		pfnCur;		// actual function that will be called
	DWORD		dwLastError;	// error received (if an error notification)
} DelayLoadInfo;

typedef	FARPROC	(WINAPI *PfnDliHook) (dliNotify, DelayLoadInfo *);

// https://msdn.microsoft.com/en-us/library/2b054ds4.aspx

BOOL	WINAPI	__FUnloadDelayLoadedDLL2(LPCSTR szDll);

// https://msdn.microsoft.com/en-us/library/09t6x5ds.aspx

FARPROC	WINAPI	__delayLoadHelper2(DelayLoadDescr *lpDLD, FARPROC *lpfnIATEntry)
{
	HMODULE	hModule;

	// https://msdn.microsoft.com/en-us/library/ff1shykw.aspx

	HMODULE			*lpHMODULE = (HMODULE *) ((LPBYTE) &__ImageBase + lpDLD->dwHMODULE);
	IMAGE_THUNK_DATA	*lpINT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwINT);
	IMAGE_THUNK_DATA	*lpIAT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwIAT);
	IMAGE_THUNK_DATA	*lpBoundIAT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwBoundIAT);
	DWORD			dwEntry = (IMAGE_THUNK_DATA *) lpfnIATEntry - lpIAT;

	// NOTE: *lpfnIATEntry == lpIAT[dwEntry].u1.Function

	DelayLoadInfo	dli = {sizeof(DelayLoadInfo),
			       lpDLD,
			       lpfnIATEntry,
			       (LPCSTR) &__ImageBase + lpDLD->dwDllName,
			       {!IMAGE_SNAP_BY_ORDINAL(lpINT[dwEntry].u1.Ordinal),
			         IMAGE_SNAP_BY_ORDINAL(lpINT[dwEntry].u1.Ordinal)
			       ? IMAGE_ORDINAL(lpINT[dwEntry].u1.Ordinal)
			       : ((IMAGE_IMPORT_BY_NAME *) ((LPBYTE) &__ImageBase + lpINT[dwEntry].u1.AddressOfData))->Name},
			       *lpHMODULE,
			       (FARPROC) NULL,
			       ERROR_SUCCESS};

	if (lpDLD->dwAttributes != 0UL)
	{
		dli.dwLastError = ERROR_INVALID_PARAMETER;

		// https://msdn.microsoft.com/en-us/library/ms680552.aspx
		// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

		RaiseException(EXCEPTION_DELAY_LOAD_INVALID_PARAMETER,
		               EXCEPTION_NONCONTINUABLE,
		               1UL,
		               (DWORD_PTR *) &dli);

		return (FARPROC) NULL;
	}

	if (dli.hmodCur == NULL)	// module not yet loaded?
	{
#ifndef LOAD_LIBRARY_SEARCH_SYSTEM32
		// https://msdn.microsoft.com/en-us/library/ms684175.aspx

		dli.hmodCur = LoadLibraryA(dli.szDll);
#else
		// https://msdn.microsoft.com/en-us/library/ms684179.aspx

		dli.hmodCur = LoadLibraryExA(dli.szDll, NULL, LOAD_LIBRARY_SEARCH_SYSTEM32);
#endif
		if (dli.hmodCur == NULL)
		{
			// https://msdn.microsoft.com/en-us/library/ms679360.aspx

			dli.dwLastError = GetLastError();

			// https://msdn.microsoft.com/en-us/library/ms680552.aspx
			// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

			RaiseException(EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND,
			               EXCEPTION_NONCONTINUABLE,
			               1UL,
			               (DWORD_PTR *) &dli);

			return (FARPROC) NULL;
		}

		// https://msdn.microsoft.com/en-us/library/ms683609.aspx

		hModule = (HMODULE) InterlockedExchangePointer((LPVOID *) lpHMODULE, dli.hmodCur);

		if (hModule == dli.hmodCur)

			// https://msdn.microsoft.com/en-us/library/ms683152.aspx

			FreeLibrary(dli.hmodCur);
		else
			if (lpDLD->dwUnloadIAT != 0UL)
			{
				//...
			}
	}

	if (lpBoundIAT != NULL && lpDLD->dwTimeStamp != 0UL)
	{
		// https://msdn.microsoft.com/en-us/library/ms680336.aspx

		IMAGE_NT_HEADERS	*lpModule = (IMAGE_NT_HEADERS *) ((LPBYTE) dli.hmodCur + ((IMAGE_DOS_HEADER *) dli.hmodCur)->e_lfanew);

		if (lpModule->Signature == IMAGE_NT_SIGNATURE
		 && lpModule->FileHeader.TimeDateStamp == lpDLD->dwTimeStamp
		 && lpModule->OptionalHeader.ImageBase == dli.hmodCur)
		{
			dli.pfnCur = (FARPROC) lpBoundIAT[dwEntry].u1.Function;

			if (dli.pfnCur != NULL)
				return *lpfnIATEntry = dli.pfnCur;
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms683212.aspx

	dli.pfnCur = GetProcAddress(dli.hmodCur, dli.dlp.szProcName);

	if (dli.pfnCur != NULL)		// function address resolved?
		return *lpfnIATEntry = dli.pfnCur;

	// https://msdn.microsoft.com/en-us/library/ms679360.aspx

	dli.dwLastError = GetLastError();

	// https://msdn.microsoft.com/en-us/library/ms680552.aspx
	// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

	RaiseException(EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND,
	               EXCEPTION_NONCONTINUABLE,
	               1UL,
	               (DWORD_PTR *) &dli);

	return (FARPROC) NULL;
}
<<NOKEEP

I386-FPU.OBJ: I386-MSC.MAK
	ML.EXE /Brepro /Cp /Cx /c /coff /Fo$@ /Gy /nologo /safeseh /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

	.686
	.model	flat, C

single	record	sign:1, exponent:8, mantissa:23

bias	equ	1 shl (width exponent - 1) - 1

	.const

	public	_fltused
_fltused dword	9876h

	.code

; MSC internal intrinsic _CIacos():
; receives argument in FPU st(0), returns result in FPU st(0)

; NOTE: _CIacos() returns correct result for 0.0 and 1.0

_CIacos	proc	public

	fld	st(0)		; st(0) = st(1) = argument
	fmul	st(0), st(0)	; st(0) = argument**2,
				; st(1) = argument
	fld1			; st(0) = 1.0,
				; st(1) = argument**2,
				; st(2) = argument
	fsubrp	st(1), st(0)	; st(0) = 1.0 - argument**2,
				; st(1) = argument
	fsqrt			; st(0) = square root of (1.0 - argument**2),
				; st(1) = argument
	fxch	st(1)		; st(0) = argument,
				; st(1) = square root of (1.0 - argument**2)
	fpatan			; st(0) = inverse circular cosine of argument
	ret

_CIacos	endp

; MSC internal intrinsic _CIasin():
; receives argument in FPU st(0), returns result in FPU st(0)

; NOTE: _CIasin() returns correct result for 0.0 and 1.0

_CIasin	proc	public

	fld	st(0)		; st(0) = st(1) = argument
	fmul	st(0), st(0)	; st(0) = argument**2,
				; st(1) = argument
	fld1			; st(0) = 1.0,
				; st(1) = argument**2,
				; st(2) = argument
	fsubrp	st(1), st(0)	; st(0) = 1.0 - argument**2,
				; st(1) = argument
	fsqrt			; st(0) = square root of (1.0 - argument**2),
				; st(1) = argument
	fpatan			; st(0) = inverse circular sine of argument
	ret

_CIasin	endp

; MSC internal intrinsic _CIatan():
; receives argument in FPU st(0), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770582.aspx

; NOTE: _CIatan() returns correct result for 0.0 and INFINITY

_CIatan	proc	public

	fld1			; st(0) = 1.0,
				; st(1) = argument
	fpatan			; st(0) = inverse circular tangent of (argument / 1.0)
	ret

_CIatan	endp

; MSC internal intrinsic _CIatan2():
; receives arguments in FPU st(0) and st(1), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770581.aspx

; NOTE: _CIatan2() returns correct result for 0.0 and INFINITY

_CIatan2 proc	public

	fxch	st(1)		; st(0) = denominator,
				; st(1) = numerator
	fpatan			; st(0) = inverse circular tangent of (numerator / denominator)
	ret

_CIatan2 endp

; MSC internal intrinsic _CIcos():
; receives argument in FPU st(0), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770589.aspx

_CIcos	proc	public

	fcos			; st(0) = cosine of argument
	fstsw	ax		; ax = FPU status word,
				; ah = B:C3:T:O:P:C2:C1:C0
	sahf			; SF:ZF:0:AF:0:PF:1:CF = ah
	jnp	short return	; |argument| < 2**63?

	fldpi			; st(0) = pi,
				; st(1) = argument
	fadd	st(0), st(0)	; st(0) = 2.0 * pi,
				; st(1) = argument
	fxch	st(1)		; st(0) = argument,
				; st(1) = 2.0 * pi
reduce:
	fprem1			; st(0) = argument modulo (2.0 * pi),
				; st(1) = 2.0 * pi
	fstsw	ax		; ax = FPU status word,
				; ah = B:C3:T:O:P:C2:C1:C0
	sahf			; SF:ZF:0:AF:0:PF:1:CF = ah
	jp	short reduce

	fstp	st(1)		; st(0) = argument modulo (2.0 * pi)
	fcos			; st(0) = cosine of argument modulo (2.0 * pi)
return:
	ret

_CIcos	endp

; MSC internal intrinsic _CIcosh():
; receives argument in FPU st(0), returns result in FPU st(0)

_CIcosh	proc	public

	call	_CIexp		; st(0) = e**argument
	fld1			; st(0) = 1.0,
				; st(1) = e**argument
	fdiv	st(0), st(1)	; st(0) = 1.0 / e**argument = e**-argument,
				; st(1) = e**argument
	faddp	st(1), st(0)	; st(0) = e**argument + e**-argument
	push	(bias - 1) shl width mantissa
				; [esp] = 0x3F000000
				;       = 0.5F
	fmul	real4 ptr [esp]	; st(0) = hyperbolic cosine of argument
	pop	eax
	ret

_CIcosh	endp

; MSC internal intrinsic _CIexp():
; receives argument in FPU st(0), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770600.aspx

; NOTE: _CIexp() returns correct result for INFINITY

_CIexp	proc	public

	fldl2e			; st(0) = log2(e),
				; st(1) = exponent
	fmulp	st(1), st(0)	; st(0) = exponent * log2(e)
if 0
	fld1			; st(0) = 1.0,
				; st(1) = exponent * log2(e)
	fld	st(1)		; st(0) = exponent * log2(e),
				; st(1) = 1.0,
				; st(2) = exponent * log2(e)
	fprem			; st(0) = (exponent * log2(e)) modulo 1.0,
				; st(1) = 1.0,
				; st(2) = exponent * log2(e)
	f2xm1			; st(0) = 2.0**((exponent * log2(e)) modulo 1.0) - 1.0,
				; st(1) = 1.0,
				; st(2) = exponent * log2(e)
	faddp	st(1), st(0)	; st(0) = 2.0**((exponent * log2(e)) modulo 1.0),
				; st(1) = exponent * log2(e)
	fscale			; st(0) = e**exponent,
				; st(1) = exponent * log2(e)
else
	fld	st(0)		; st(0) = st(1) = exponent * log2(e)
	frndint			; st(0) = integer(exponent * log2(e)),
				; st(1) = exponent * log2(e)
	fsub	st(1), st(0)	; st(0) = integer(exponent * log2(e)),
				; st(1) = fraction(exponent * log2(e))
	fxch	st(1)		; st(0) = fraction(exponent * log2(e)),
				; st(1) = integer(exponent * log2(e))
	f2xm1			; st(0) = 2.0**fraction(exponent * log2(e)) - 1.0,
				; st(1) = integer(exponent * log2(e))
	fld1			; st(0) = 1.0,
				; st(1) = 2.0**fraction(exponent * log2(e)) - 1.0,
				; st(2) = integer(exponent * log2(e))
	faddp	st(1), st(0)	; st(0) = 2.0**fraction(exponent * log2(e)),
				; st(1) = integer(exponent * log2(e))
	fscale			; st(0) = e**exponent,
				; st(1) = integer(exponent * log2(e))
endif
	fstp	st(1)		; st(0) = e**exponent
	ret

_CIexp	endp

; MSC internal intrinsic _CIfmod():
; receives arguments in FPU st(0) and st(1), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770591.aspx

_CIfmod	proc	public

reduce:
	fprem			; st(0) = remainder,
				; st(1) = divisor
	fstsw	ax		; ax = FPU status word,
				; ah = B:C3:T:O:P:C2:C1:C0
	sahf			; SF:ZF:0:AF:0:PF:1:CF = ah
	jp	short reduce

	fstp	st(1)		; st(0) = remainder
	ret

_CIfmod	endp

; MSC internal intrinsic _CIlog():
; receives argument in FPU st(0), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770580.aspx

_CIlog	proc	public

	fldln2			; st(0) = ln(2.0),
				; st(1) = argument
	fxch	st(1)		; st(0) = argument,
				; st(1) = ln(2.0)
	fyl2x			; st(0) = natural logarithm of argument
	ret

_CIlog	endp

; MSC internal intrinsic _CIlog10():
; receives argument in FPU st(0), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770578.aspx

_CIlog10 proc	public

	fldlg2			; st(0) = log10(2.0),
				; st(1) = argument
	fxch	st(1)		; st(0) = argument,
				; st(1) = log10(2.0)
	fyl2x			; st(0) = logarithm to base 10 of argument
	ret

_CIlog10 endp

; MSC internal intrinsic _CIpow():
; receives arguments in FPU st(0) and st(1), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770583.aspx

_CIpow	proc	public

	fxch	st(1)		; st(0) = base,
				; st(1) = exponent
	fyl2x			; st(0) = exponent * log2(base)
	fld	st(0)		; st(0) = st(1) = exponent * log2(base)
	frndint			; st(0) = integer(exponent * log2(base)),
				; st(1) = exponent * log2(base)
	fsub	st(1), st(0)	; st(0) = integer(exponent * log2(base)),
				; st(1) = fraction(exponent * log2(base))
	fxch	st(1)		; st(0) = fraction(exponent * log2(base)),
				; st(1) = integer(exponent * log2(base))
	f2xm1			; st(0) = 2.0**fraction(exponent * log2(base)) - 1.0,
				; st(1) = integer(exponent * log2(base))
	fld1			; st(0) = 1.0,
				; st(1) = 2.0**fraction(exponent * log2(base)) - 1.0,
				; st(2) = integer(exponent * log2(base))
	faddp	st(1), st(0)	; st(0) = 2.0**fraction(exponent * log2(base)),
				; st(1) = integer(exponent * log2(base))
	fscale			; st(0) = base**exponent,
				; st(1) = integer(exponent * log2(base))
	fstp	st(1)		; st(0) = base**exponent
	ret

_CIpow	endp

; MSC internal intrinsic _CIsin():
; receives argument in FPU st(0), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770597.aspx

_CIsin	proc	public

	fsin			; st(0) = sine of argument
	fstsw	ax		; ax = FPU status word,
				; ah = B:C3:T:O:P:C2:C1:C0
	sahf			; SF:ZF:0:AF:0:PF:1:CF = ah
	jnp	short return	; |argument| < 2**63?

	fldpi			; st(0) = pi,
				; st(1) = argument
	fadd	st(0), st(0)	; st(0) = 2.0 * pi,
				; st(1) = argument
	fxch	st(1)		; st(0) = argument,
				; st(1) = 2.0 * pi
reduce:
	fprem1			; st(0) = argument modulo (2.0 * pi),
				; st(1) = 2.0 * pi
	fstsw	ax		; ax = FPU status word,
				; ah = B:C3:T:O:P:C2:C1:C0
	sahf			; SF:ZF:0:AF:0:PF:1:CF = ah
	jp	short reduce

	fstp	st(1)		; st(0) = argument modulo (2.0 * pi)
	fsin			; st(0) = sine of argument modulo (2.0 * pi)
return:
	ret

_CIsin	endp

; MSC internal intrinsic _CIsinh():
; receives argument in FPU st(0), returns result in FPU st(0)

_CIsinh	proc	public

	call	_CIexp		; st(0) = e**argument
	fld1			; st(0) = 1.0,
				; st(1) = e**argument
	fdiv	st(0), st(1)	; st(0) = 1.0 / e**argument = e**-argument,
				; st(1) = e**argument
	fsubp	st(1), st(0)	; st(0) = e**argument - e**-argument
	push	(bias - 1) shl width mantissa
				; [esp] = 0x3F000000
				;       = 0.5F
	fmul	real4 ptr [esp]	; st(0) = hyperbolic sine of argument
	pop	eax
	ret

_CIsinh	endp

; MSC internal intrinsic _CIsqrt():
; receives argument in FPU st(0), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770588.aspx

_CIsqrt	proc	public

	fsqrt			; st(0) = square root of radicand
	ret

_CIsqrt	endp

; MSC internal intrinsic _CItan():
; receives argument in FPU st(0), returns result in FPU st(0)

; https://msdn.microsoft.com/en-us/library/ff770595.aspx

_CItan	proc	public

	fptan			; st(0) = 1.0,
				; st(1) = tangent of argument
	fstsw	ax		; ax = FPU status word,
				; ah = B:C3:T:O:P:C2:C1:C0
	sahf			; SF:ZF:0:AF:0:PF:1:CF = ah
	jnp	short return	; |argument| < 2**63?

	fldpi			; st(0) = pi,
				; st(1) = argument
	fadd	st(0), st(0)	; st(0) = 2.0 * pi,
				; st(1) = argument
	fxch	st(1)		; st(0) = argument,
				; st(1) = 2.0 * pi
reduce:
	fprem1			; st(0) = argument modulo (2.0 * pi),
				; st(1) = 2.0 * pi
	fstsw	ax		; ax = FPU status word,
				; ah = B:C3:T:O:P:C2:C1:C0
	sahf			; SF:ZF:0:AF:0:PF:1:CF = ah
	jp	short reduce

	fstp	st(1)		; st(0) = argument modulo (2.0 * pi)
	fptan			; st(0) = 1.0,
				; st(1) = tangent of argument modulo (2.0 * pi)
return:
	fstp	st(0)		; st(0) = tangent of argument
	ret

_CItan	endp

; MSC internal intrinsic _CItanh():
; receives argument in FPU st(0), returns result in FPU st(0)

_CItanh	proc	public

	call	_CIexp		; st(0) = e**argument
	fmul	st(0), st(0)	; st(0) = e**argument * e**argument
				;       = e**(argument + argument)
	fld1			; st(0) = 1.0,
				; st(1) = e**(argument + argument)
	fadd	st(1), st(0)	; st(0) = 1.0,
				; st(1) = e**(argument + argument) + 1.0
	fadd	st(0), st(0)	; st(0) = 2.0,
				; st(1) = e**(argument + argument) + 1.0
	fdivrp	st(1), st(0)	; st(0) = 2.0 / (e**(argument + argument) + 1.0)
	fld1			; st(0) = 1.0,
				; st(1) = 2.0 / (e**(argument + argument) + 1.0)
	fsubrp	st(1), st(0)	; st(0) = 1.0 - 2.0 / (e**(argument + argument) + 1.0)
				;       = hyperbolic tangent of argument
	ret

_CItanh	endp

; MSC internal intrinsic _ftol():
; receives argument in FPU st(0), returns result in eax

; https://msdn.microsoft.com/en-us/library/6d9xx1d2.aspx

; NOTE: fistp rounds to nearest (even) integer!

_ftol	proc	public

	push	eax
	fistp	dword ptr [esp]	; [esp] = integer(argument)
	pop	eax		; eax = integer(argument)
	ret

_ftol	endp

; MSC internal intrinsic _ftol2():
; receives argument in FPU st(0), returns result in edx:eax

; NOTE: fistp rounds to nearest (even) integer!

_ftol2	proc	public

	push	edx
	push	eax
	fistp	qword ptr [esp]	; [esp] = integer(argument)
	pop	eax
	pop	edx		; edx:eax = integer(argument)
	ret

_ftol2	endp

; MSC internal intrinsic _ftol2_sse():
; receives argument in FPU st(0), returns result in edx:eax

; NOTE: fisttp truncates, i.e. rounds towards 0!

_ftol2_sse proc	public

	push	edx
	push	eax
	fisttp	qword ptr [esp]	; [esp] = integer(argument)
	pop	eax
	pop	edx		; edx:eax = integer(argument)
	ret

_ftol2_sse endp
	end
<<NOKEEP

I386-GS.OBJ: I386-MSC.MAK
	ML.EXE /Brepro /Cp /Cx /c /coff /Fo$@ /Gy /nologo /safeseh /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

	.586
	.model	flat; C

; https://msdn.microsoft.com/en-us/library/9a89h429.aspx

	extern	___safe_se_handler_count :abs
	extern	___safe_se_handler_table :dword

; https://msdn.microsoft.com/en-us/library/ms680328.aspx

_lcu_32	struct	4		; IMAGE_LOAD_CONFIG_DIRECTORY32
	dword	sizeof _lcu_32
	dword	'VOID'		; 2006-04-21 21:32:06 UTC
	word	0, 815
	dword	10 dup (0)
	word	0, 2048		; LOAD_LIBRARY_SEARCH_SYSTEM32
	dword	0
	dword	offset ___security_cookie
	dword	offset ___safe_se_handler_table
	dword	offset ___safe_se_handler_count
	dword	5 dup (0)
_lcu_32	ends

	.const

; https://msdn.microsoft.com/en-us/library/gg463119.aspx

	public	__load_config_used
__load_config_used \
	_lcu_32	<>

	.data

	public	___security_cookie
___security_cookie \
	dword	3141592654

	.code

; https://msdn.microsoft.com/en-us/library/8dbf701c.aspx
; https://msdn.microsoft.com/en-us/library/aa290051.aspx

@__security_check_cookie@4 \
	proc	public		; void __fastcall __security_check_cookie(dword cookie)

	cmp	ecx, ___security_cookie
	jne	short fastfail
	ret

; https://msdn.microsoft.com/en-us/library/dn774154.aspx

fastfail:
	xor	ecx, ecx
	int	41

@__security_check_cookie@4 \
	endp

; https://msdn.microsoft.com/en-us/library/ms235362.aspx

___security_init_cookie \
	proc	public		; void __cdecl __security_init_cookie(void)

	mov	eax, ___security_cookie
	cmp	eax, 3141592654
	je	short init

	test	eax, eax
	jne	short exit
init:
	rdtsc			; eax = low dword of time stamp counter,
				; edx = high dword of time stamp counter
	xor	eax, edx	; eax = pseudo random number
	mov	___security_cookie, eax
exit:
	ret

___security_init_cookie	\
	endp
	end
<<NOKEEP

I386-I64.OBJ: I386-MSC.MAK
	ML.EXE /Brepro /Cp /Cx /c /coff /DJCCLESS /Fo$@ /Gy /nologo /safeseh /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

	.386
	.model	flat, C
	.code

; MSC internal _alldiv():
; receives arguments on stack, returns quotient in edx:eax

; https://msdn.microsoft.com/en-us/library/ms648425.aspx

; NOTE: _alldiv() can raise 'division by zero' exception; it does
;       not raise 'integer overflow' exception on quotient overflow,
;       but returns 2**63 for -2**63 / -1!

_alldiv	proc	public		; sqword _alldiv(sqword dividend, sqword divisor)

	; determine sign of dividend and compute |dividend|

	mov	eax, [esp+8]
	mov	ecx, [esp+4]	; eax:ecx = dividend

	cdq			; edx = (dividend < 0) ? -1 : 0
	xor	ecx, edx
	xor	eax, edx	; eax:ecx = (dividend < 0) ? ~dividend : dividend
	sub	ecx, edx
	sbb	eax, edx	; eax:ecx = (dividend < 0) ? -dividend : dividend
				;         = |dividend|

	mov	[esp+4], ecx	; write |dividend| back on stack
	mov	[esp+8], eax

	push	edx		; [esp] = (dividend < 0) ? -1 : 0

	; determine sign of divisor and compute |divisor|

	mov	edx, [esp+20]
	mov	eax, [esp+16]	; edx:eax = divisor

	mov	ecx, edx
	sar	ecx, 31		; ecx = (divisor < 0) ? -1 : 0
	xor	eax, ecx
	xor	edx, ecx	; edx:eax = (divisor < 0) ? ~divisor : divisor
	sub	eax, ecx
	sbb	edx, ecx	; edx:eax = (divisor < 0) ? -divisor : divisor
				;         = |divisor|

	mov	[esp+16], eax	; write |divisor| back on stack
	mov	[esp+20], edx

	xor	[esp], ecx	; [esp] = (dividend < 0) ^ (divisor < 0) ? -1 : 0
				;       = (quotient < 0) ? -1 : 0

	mov	ecx, [esp+12]	; ecx = high dword of dividend
	cmp	[esp+8], eax
	sbb	ecx, edx
	jb	short trivial	; dividend < divisor?

	bsr	ecx, edx	; ecx = index of most significant '1' bit
				;        in high dword of divisor
	jnz	short extended	; high dword of divisor <> 0?

	; remainder < divisor < 2**32

	mov	ecx, eax	; ecx = (low dword of) divisor
	mov	eax, [esp+12]	; eax = high dword of dividend
	cmp	eax, ecx
	jae	short long	; high dword of dividend >= divisor?

	; perform normal division
normal:
	mov	edx, eax	; edx = high dword of dividend
	mov	eax, [esp+8]	; edx:eax = dividend
	div	ecx		; eax = (low dword of) quotient,
				; edx = (low dword of) remainder
;;	xor	edx, edx	; edx:eax = |quotient|

	jmp	short quotient

	; perform "long" alias "schoolbook" division
long:
;;	xor	edx, edx	; edx:eax = high dword of dividend
	div	ecx		; eax = high dword of quotient,
				; edx = high dword of remainder'
	push	eax		; [esp] = high dword of quotient
	mov	eax, [esp+12]	; eax = low dword of dividend
	div	ecx		; eax = low dword of quotient,
				; edx = (low dword of) remainder
	pop	edx		; edx:eax = |quotient|

	pop	ecx		; ecx = (quotient < 0) ? -1 : 0
	xor	eax, ecx
	xor	edx, ecx
	sub	eax, ecx
	sbb	edx, ecx	; edx:eax = quotient

	ret	16		; callee restores stack

	; dividend < divisor: quotient = 0
trivial:
	pop	ecx		; ecx = (quotient < 0) ? -1 : 0
	xor	eax, eax
	xor	edx, edx	; edx:eax = quotient = 0

	ret	16		; callee restores stack

	; 2**63 >= dividend >= divisor >= 2**32: quotient < 2**32
extended:
	xor	ecx, 31		; ecx = number of leading '0' bits
				;        in (high dword of) divisor
	jz	short special	; divisor = 2**63?

	; perform "extended & adjusted" division

	shld	edx, eax, cl	; edx = divisor / 2**(index + 1)
				;     = divisor'
;;	shl	eax, cl
	push	ebx
	mov	ebx, edx	; ebx = divisor'
ifndef JCCLESS
	xor	eax, eax	; eax = high dword of quotient' = 0
	mov	edx, [esp+16]	; edx = high dword of dividend
	cmp	edx, ebx
	jb	short @f	; high dword of dividend < divisor'?

	; high dword of dividend >= divisor':
	; subtract divisor' from high dword of dividend to prevent possible
	; division overflow and set most significant bit of quotient"

	sub	edx, ebx	; edx = high dword of dividend - divisor'
				;     = high dword of dividend'
	inc	eax		; eax = high dword of quotient' = 1
@@:
	push	eax		; [esp] = high dword of quotient'
else ; JCCLESS
	mov	edx, [esp+16]	; edx = high dword of dividend
	cmp	edx, ebx	; CF = (high dword of dividend < divisor')
	sbb	eax, eax	; eax = (high dword of dividend < divisor') ? -1 : 0
	inc	eax		; eax = (high dword of dividend < divisor') ? 0 : 1
				;     = high dword of quotient'
	push	eax		; [esp] = high dword of quotient'
if 0
	neg	eax		; eax = (high dword of dividend < divisor') ? 0 : -1
	and	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
else
	imul	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
endif
	sub	edx, eax	; edx = high dword of dividend
				;     - (high dword of dividend < divisor') ? 0 : divisor'
				;     = high dword of dividend'
endif ; JCCLESS
	mov	eax, [esp+16]	; eax = low dword of dividend
				;     = low dword of dividend'
	div	ebx		; eax = dividend' / divisor'
				;     = low dword of quotient',
				; edx = remainder'
	pop	ebx		; ebx = high dword of quotient'
	shld	ebx, eax, cl	; ebx = quotient' / 2**(index + 1)
				;     = dividend / divisor'
				;     = quotient"
;;	shl	eax, cl
	mov	eax, [esp+20]	; eax = low dword of divisor
	mul	ebx		; edx:eax = low dword of divisor * quotient"
	mov	ecx, [esp+24]	; ecx = high dword of divisor
	imul	ecx, ebx	; ecx = high dword of divisor * quotient"
	add	edx, ecx	; edx:eax = divisor * quotient"
	mov	ecx, [esp+16]	; ecx = high dword of dividend
	cmp	[esp+12], eax
	sbb	ecx, edx	; CF = (dividend < divisor * quotient")
				;    = (remainder" < 0)
	sbb	eax, eax	; eax = (remainder" < 0) ? -1 : 0
	add	eax, ebx	; eax = quotient" - (remainder" < 0)
				;     = (low dword of) |quotient|
;;	xor	edx, edx	; edx:eax = |quotient|
	pop	ebx
quotient:
	pop	edx		; edx = (quotient < 0) ? -1 : 0
	xor	eax, edx
	sub	eax, edx
	sbb	edx, edx	; edx:eax = quotient

	ret	16		; callee restores stack

	; dividend = divisor = -2**63: quotient = 1
special:
	pop	eax		; eax = sign of quotient = 0
	inc	eax		; eax = (low dword of) quotient = 1
	xor	edx, edx	; edx:eax = quotient = 1

	ret	16		; callee restores stack

_alldiv	endp

; MSC internal _alldvrm():
; receives arguments on stack, returns quotient in edx:eax and remainder in ebx:ecx

; NOTE: _alldvrm() can raise 'division by zero' exception; it does
;       not raise 'integer overflow' exception on quotient overflow,
;       but returns 2**63 for -2**63 / -1 and 0 for -2**63 % -1!

_alldvrm proc	public		; sqword _alldvrm(sqword dividend, sqword divisor)

	; determine sign of dividend and compute |dividend|

	mov	edx, [esp+8]
	mov	eax, [esp+4]	; edx:eax = dividend

	mov	ebx, edx
	sar	ebx, 31		; ebx = (dividend < 0) ? -1 : 0
				;     = (remainder < 0) ? -1 : 0
	xor	eax, ebx
	xor	edx, ebx	; edx:eax = (dividend < 0) ? ~dividend : dividend
	sub	eax, ebx
	sbb	edx, ebx	; edx:eax = (dividend < 0) ? -dividend : dividend
				;         = |dividend|

	mov	[esp+4], eax	; write |dividend| back on stack
	mov	[esp+8], edx

	; determine sign of divisor and compute |divisor|

	mov	edx, [esp+16]
	mov	eax, [esp+12]	; edx:eax = divisor

	mov	ecx, edx
	sar	ecx, 31		; ecx = (divisor < 0) ? -1 : 0
	xor	eax, ecx
	xor	edx, ecx	; edx:eax = (divisor < 0) ? ~divisor : divisor
	sub	eax, ecx
	sbb	edx, ecx	; edx:eax = (divisor < 0) ? -divisor : divisor
				;         = |divisor|

	mov	[esp+12], eax	; write |divisor| back on stack
	mov	[esp+16], edx

	xor	ecx, ebx	; ecx = (divisor < 0) ^ (dividend < 0) ? -1 : 0
				;     = (quotient < 0) ? -1 : 0
	push	ecx		; [esp] = (quotient < 0) ? -1 : 0
	push	ebx		; [esp] = (remainder < 0) ? -1 : 0

	mov	ecx, [esp+16]	; ecx = high dword of dividend
	cmp	[esp+12], eax
	sbb	ecx, edx
	jb	short trivial	; dividend < divisor?

	bsr	ecx, edx	; ecx = index of most significant '1' bit
				;        in high dword of divisor
	jnz	short extended	; high dword of divisor <> 0?

	; remainder < divisor < 2**32

	mov	ecx, eax	; ecx = (low dword of) divisor
	mov	eax, [esp+16]	; eax = high dword of dividend
	cmp	eax, ecx
	jae	short long	; high dword of dividend >= divisor?

	; perform normal division
normal:
	mov	edx, eax	; edx = high dword of dividend
	xor	ebx, ebx	; ebx = high dword of quotient = 0
	jmp	short next

	; perform "long" alias "schoolbook" division
long:
;;	xor	edx, edx	; edx:eax = high dword of dividend
	div	ecx		; eax = high dword of quotient,
				; edx = high dword of remainder'
	mov	ebx, eax	; ebx = high dword of quotient
next:
	mov	eax, [esp+12]	; eax = low dword of dividend
	div	ecx		; eax = low dword of quotient,
				; edx = (low dword of) remainder

	mov	ecx, edx	; ecx = (low dword of) |remainder|
	mov	edx, ebx	; edx:eax = |quotient|
;;	xor	ebx, ebx	; ebx:ecx = |remainder|
if 0
	mov	ebx, [esp+4]	; ebx = (quotient < 0) ? -1 : 0
	xor	eax, ebx
	xor	edx, ebx
	sub	eax, ebx
	sbb	edx, ebx	; edx:eax = quotient

	pop	ebx		; ebx = (remainder < 0) ? -1 : 0
	xor	ecx, ebx
	sub	ecx, ebx
	sbb	ebx, ebx	; ebx:ecx = remainder
else
	pop	ebx		; ebx = (remainder < 0) ? -1 : 0
	xor	ecx, ebx
	sub	ecx, ebx
	sbb	ebx, ebx	; ebx:ecx = remainder

	xor	eax, [esp]
	xor	edx, [esp]
	sub	eax, [esp]
	sbb	edx, [esp]	; edx:eax = quotient
endif
	add	esp, 4
	ret	16		; callee restores stack

	; dividend < divisor: quotient = 0, remainder = dividend
trivial:
	pop	eax		; eax = (remainder < 0) ? -1 : 0
	mov	ecx, [esp+8]
	mov	ebx, [esp+12]	; ebx:ecx = |remainder| = |dividend|
	xor	ecx, eax
	xor	ebx, eax
	sub	ecx, eax
	sbb	ebx, eax	; ebx:ecx = remainder

	pop	edx		; edx = (quotient < 0) ? -1 : 0
	xor	eax, eax
	xor	edx, edx	; edx:eax = quotient = 0

	ret	16		; callee restores stack

	; 2**63 >= dividend >= divisor >= 2**32: quotient < 2**32
extended:
	xor	ecx, 31		; ecx = number of leading '0' bits
				;        in (high dword of) divisor
	jz	short special	; divisor = 2**63?

	; perform "extended & adjusted" division

	shld	edx, eax, cl	; edx = divisor / 2**(index + 1)
				;     = divisor'
;;	shl	eax, cl
	mov	ebx, edx	; ebx = divisor'
ifndef JCCLESS
	xor	eax, eax	; eax = high dword of quotient' = 0
	mov	edx, [esp+16]	; edx = high dword of dividend
	cmp	edx, ebx
	jb	short @f	; high dword of dividend < divisor'?

	; high dword of dividend >= divisor':
	; subtract divisor' from high dword of dividend to prevent possible
	; division overflow and set most significant bit of quotient"

	sub	edx, ebx	; edx = high dword of dividend - divisor'
				;     = high dword of dividend'
	inc	eax		; eax = high dword of quotient' = 1
@@:
	push	eax		; [esp] = high dword of quotient'
else ; JCCLESS
	mov	edx, [esp+16]	; edx = high dword of dividend
	cmp	edx, ebx	; CF = (high dword of dividend < divisor')
	sbb	eax, eax	; eax = (high dword of dividend < divisor') ? -1 : 0
	inc	eax		; eax = (high dword of dividend < divisor') ? 0 : 1
				;     = high dword of quotient'
	push	eax		; [esp] = high dword of quotient'
if 0
	neg	eax		; eax = (high dword of dividend < divisor') ? 0 : -1
	and	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
else
	imul	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
endif
	sub	edx, eax	; edx = high dword of dividend
				;     - (high dword of dividend < divisor') ? 0 : divisor'
				;     = high dword of dividend'
endif ; JCCLESS
	mov	eax, [esp+16]	; eax = low dword of dividend
				;     = low dword of dividend'
	div	ebx		; eax = dividend' / divisor'
				;     = low dword of quotient',
				; edx = remainder'
	pop	ebx		; ebx = high dword of quotient'
	shld	ebx, eax, cl	; ebx = quotient' / 2**(index + 1)
				;     = dividend / divisor'
				;     = quotient"
;;	shl	eax, cl
	push	ebx		; [esp] = quotient"
	mov	eax, [esp+24]	; eax = low dword of divisor
	mul	ebx		; edx:eax = low dword of divisor * quotient"
	imul	ebx, [esp+28]	; ebx = high dword of divisor * quotient"
	add	edx, ebx	; edx:eax = divisor * quotient"
	mov	ecx, [esp+16]
	mov	ebx, [esp+20]	; ebx:ecx = dividend
	sub	ecx, eax
	sbb	ebx, edx	; ebx:ecx = dividend - divisor * quotient"
				;         = remainder"
ifndef JCCLESS
	pop	eax		; eax = quotient"
	jnb	short @f	; remainder" >= 0?
				;  (with borrow, it is off by divisor,
				;   and quotient" is off by 1)
	add	ecx, [esp+20]
	adc	ebx, [esp+24]	; ebx:ecx = remainder" + divisor
				;         = |remainder|
	dec	eax		; eax = quotient" - 1
				;     = low dword of |quotient|
@@:
else ; JCCLESS
	sbb	eax, eax	; eax = (remainder" < 0) ? -1 : 0
	cdq			; edx = (remainder" < 0) ? -1 : 0
	add	[esp], eax	; [esp] = quotient" - (remainder" < 0)
				;       = (low dword of) |quotient|
	and	eax, [esp+24]
	and	edx, [esp+28]	; edx:eax = (remainder" < 0) ? divisor : 0
	add	ecx, eax
	adc	ebx, edx	; ebx:ecx = remainder" + divisor
				;         = |remainder|
	pop	eax		; eax = (low dword of) |quotient|
endif ; JCCLESS
;;	xor	edx, edx	; edx:eax = |quotient|

	pop	edx		; edx = (remainder < 0) ? -1 : 0
	xor	ecx, edx
	xor	ebx, edx
	sub	ecx, edx
	sbb	ebx, edx	; ebx:ecx = remainder

	pop	edx		; edx = (quotient < 0) ? -1 : 0
	xor	eax, edx
	sub	eax, edx
	sbb	edx, edx	; edx:eax = quotient

	ret	16		; callee restores stack

	; dividend = divisor = -2**63: quotient = 1, remainder = 0
special:
	pop	ebx		; ebx = sign of remainder = -1
	inc	ebx
;;	xor	ecx, ecx	; ebx:ecx = remainder = 0

	pop	eax		; eax = sign of quotient = 0
	inc	eax		; eax = (low dword of) quotient = 1
	xor	edx, edx	; edx:eax = quotient = 1

	ret	16		; callee restores stack

_alldvrm endp

; MSC internal _allmul():
; receives arguments on stack, returns product modulo 2**64 in edx:eax

; https://docs.microsoft.com/en-us/windows/win32/devnotes/-win32-allmul

_allmul	proc	public		; sqword _allmul(sqword multiplicand, sqword multiplier)

	mov	eax, [esp+4]	; eax = low dword of multiplicand
	mov	edx, [esp+8]	; edx = high dword of multiplicand
	imul	edx, [esp+12]	; edx = high dword of multiplicand
				;     * low dword of multiplier
	mov	ecx, [esp+16]	; ecx = high dword of multiplier
	imul	ecx, eax	; ecx = high dword of multiplier
				;     * low dword of multiplicand
	add	ecx, edx	; ecx = high dword of multiplier
				;     * low dword of multiplicand
				;     + high dword of multiplicand
				;     * low dword of multiplier
	mul	dword ptr [esp+12]
				; edx:eax = low dword of multiplicand
				;         * low dword of multiplier
	add	edx, ecx	; edx:eax = product % 2**64

	ret	16		; callee restores stack

_allmul	endp

; MSC internal _allrem():
; receives arguments on stack, returns remainder in edx:eax

; NOTE: _allrem() can raise 'division by zero' exception; it does
;       not raise 'integer overflow' exception on quotient overflow,
;       but returns 0 for -2**63 % -1!

_allrem	proc	public		; sqword _allrem(sqword dividend, sqword divisor)

	; determine sign of dividend and compute |dividend|

	mov	eax, [esp+8]
	mov	ecx, [esp+4]	; eax:ecx = dividend

	cdq			; edx = (dividend < 0) ? -1 : 0
	xor	ecx, edx
	xor	eax, edx	; ecx:eax = (dividend < 0) ? ~dividend : dividend
	sub	ecx, edx
	sbb	eax, edx	; ecx:eax = (dividend < 0) ? -dividend : dividend
				;         = |dividend|

	mov	[esp+4], ecx	; write |dividend| back on stack
	mov	[esp+8], eax

	push	edx		; [esp] = (dividend < 0) ? -1 : 0

	; determine sign of divisor and compute |divisor|

	mov	edx, [esp+20]
	mov	eax, [esp+16]	; edx:eax = divisor

	mov	ecx, edx
	sar	ecx, 31		; ecx = (divisor < 0) ? -1 : 0
	xor	eax, ecx
	xor	edx, ecx	; edx:eax = (divisor < 0) ? ~divisor : divisor
	sub	eax, ecx
	sbb	edx, ecx	; edx:eax = (divisor < 0) ? -divisor : divisor
				;         = |divisor|

	mov	[esp+16], eax	; write |divisor| back on stack
	mov	[esp+20], edx

	mov	ecx, [esp+12]	; ecx = high dword of dividend
	cmp	[esp+8], eax
	sbb	ecx, edx
	jb	short trivial	; dividend < divisor?

	bsr	ecx, edx	; ecx = index of most significant '1' bit
				;        in high dword of divisor
	jnz	short extended	; high dword of divisor <> 0?

	; remainder < divisor < 2**32

	mov	ecx, eax	; ecx = (low dword of) divisor
	mov	eax, [esp+12]	; eax = high dword of dividend
	cmp	eax, ecx
	jae	short long	; high dword of dividend >= divisor?

	; perform normal division
normal:
	mov	edx, eax	; edx = high dword of dividend
	jmp	short next

	; perform "long" alias "schoolbook" division
long:
;;	xor	edx, edx	; edx:eax = high dword of dividend
	div	ecx		; eax = high dword of quotient,
				; edx = high dword of remainder'
next:
	mov	eax, [esp+8]	; eax = low dword of dividend
	div	ecx		; eax = low dword of quotient,
				; edx = (low dword of) remainder
	mov	eax, edx	; eax = (low dword of) |remainder|
;;	xor	edx, edx	; edx:eax = |remainder|

	pop	edx		; edx = (remainder < 0) ? -1 : 0
	xor	eax, edx
	sub	eax, edx
	sbb	edx, edx	; edx:eax = remainder

	ret	16		; callee restores stack

	; dividend < divisor: remainder = dividend
trivial:
	mov	eax, [esp+8]
	mov	edx, [esp+12]	; edx:eax = |remainder| = |dividend|

	jmp	short remainder

	; 2**63 >= dividend >= divisor >= 2**32: quotient < 2**32
extended:
	xor	ecx, 31		; ecx = number of leading '0' bits
				;        in (high dword of) divisor
	jz	short special	; divisor = 2**63?

	; perform "extended & adjusted" division

	shld	edx, eax, cl	; edx = divisor / 2**(index + 1)
				;     = divisor'
;;	shl	eax, cl
	push	ebx
	mov	ebx, edx	; ebx = divisor'
ifndef JCCLESS
	xor	eax, eax	; eax = high dword of quotient' = 0
	mov	edx, [esp+16]	; edx = high dword of dividend
	cmp	edx, ebx
	jb	short @f	; high dword of dividend < divisor'?

	; high dword of dividend >= divisor':
	; subtract divisor' from high dword of dividend to prevent possible
	; division overflow and set most significant bit of quotient"

	sub	edx, ebx	; edx = high dword of dividend - divisor'
				;     = high dword of dividend'
	inc	eax		; eax = high dword of quotient' = 1
@@:
	push	eax		; [esp] = high dword of quotient'
else ; JCCLESS
	mov	edx, [esp+16]	; edx = high dword of dividend
	cmp	edx, ebx	; CF = (high dword of dividend < divisor')
	sbb	eax, eax	; eax = (high dword of dividend < divisor') ? -1 : 0
	inc	eax		; eax = (high dword of dividend < divisor') ? 0 : 1
				;     = high dword of quotient'
	push	eax		; [esp] = high dword of quotient'
if 0
	neg	eax		; eax = (high dword of dividend < divisor') ? 0 : -1
	and	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
else
	imul	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
endif
	sub	edx, eax	; edx = high dword of dividend
				;     - (high dword of dividend < divisor') ? 0 : divisor'
				;     = high dword of dividend'
endif ; JCCLESS
	mov	eax, [esp+16]	; eax = low dword of dividend
				;     = low dword of dividend'
	div	ebx		; eax = dividend' / divisor'
				;     = low dword of quotient',
				; edx = remainder'
	pop	ebx		; ebx = high dword of quotient'
	shld	ebx, eax, cl	; ebx = quotient' / 2**(index + 1)
				;     = dividend / divisor'
				;     = quotient"
;;	shl	eax, cl
	mov	eax, [esp+20]	; eax = low dword of divisor
	mul	ebx		; edx:eax = low dword of divisor * quotient"
	imul	ebx, [esp+24]	; ebx = high dword of divisor * quotient"
	add	edx, ebx	; edx:eax = divisor * quotient"
	mov	ecx, [esp+12]
	mov	ebx, [esp+16]	; ebx:ecx = dividend
	sub	ecx, eax
	sbb	ebx, edx	; ebx:ecx = dividend - divisor * quotient"
				;         = remainder"
ifndef JCCLESS
	jnb	short @f	; remainder" >= 0?
				;  (with borrow, it is off by divisor,
				;   and quotient" is off by 1)
	add	ecx, [esp+20]
	adc	ebx, [esp+24]	; ebx:ecx = remainder" + divisor
				;         = remainder
@@:
	mov	eax, ecx
	mov	edx, ebx	; edx:eax = |remainder|
else ; JCCLESS
	sbb	eax, eax	; eax = (remainder" < 0) ? -1 : 0
	cdq			; edx = (remainder" < 0) ? -1 : 0
	and	eax, [esp+20]
	and	edx, [esp+24]	; edx:eax = (remainder" < 0) ? divisor : 0
	add	eax, ecx
	adc	edx, ebx	; edx:eax = remainder" + divisor
				;         = |remainder|
endif ; JCCLESS
	pop	ebx
remainder:
	pop	ecx		; ecx = (remainder < 0) ? -1 : 0
	xor	eax, ecx
	xor	edx, ecx
	sub	eax, ecx
	sbb	edx, ecx	; edx:eax = remainder

	ret	16		; callee restores stack

	; dividend = divisor = -2**63: remainder = 0
special:
	pop	eax		; eax = sign of remainder = -1
	inc	eax
	xor	edx, edx	; edx:eax = remainder = 0

	ret	16		; callee restores stack

_allrem	endp

; MSC internal _allshl():
; receives arguments in edx:eax and cl, returns result in edx:eax

; NOTE: _allshl() applies shift count modulo 64

_allshl	proc	public		; sqword _allshl(sqword value, byte count)

	test	cl, 32
	jnz	short @f	; count > 31?

	shld	edx, eax, cl
	shl	eax, cl
	ret
@@:
	mov	edx, eax
	shl	edx, cl
	xor	eax, eax
	ret

_allshl	endp

; MSC internal _allshr():
; receives arguments in edx:eax and cl, returns result in edx:eax

; NOTE: _allshr() applies shift count modulo 64

_allshr	proc	public		; sqword _allshr(sqword value, byte count)

	test	cl, 32
	jnz	short @f	; count > 31?

	shrd	eax, edx, cl
	sar	edx, cl
	ret
@@:
	mov	eax, edx
	sar	eax, cl
	sar	edx, 31
	ret

_allshr	endp

; MSC internal _aulldiv():
; receives arguments on stack, returns quotient in edx:eax

; https://msdn.microsoft.com/en-us/library/mt703294.aspx

; NOTE: _aulldiv() can raise 'division by zero' exception!

_aulldiv proc	public		; qword _aulldiv(qword dividend, qword divisor)

	mov	ecx, [esp+8]	; ecx = high dword of dividend
	mov	eax, [esp+12]
	mov	edx, [esp+16]	; edx:eax = divisor
	cmp	[esp+4], eax
	sbb	ecx, edx
	jb	short trivial	; dividend < divisor?

	bsr	ecx, edx	; ecx = index of most significant '1' bit
				;        in high dword of divisor
	jnz	short extended	; high dword of divisor <> 0?

	; remainder < divisor < 2**32

	mov	ecx, eax	; ecx = (low dword of) divisor
	mov	eax, [esp+8]	; eax = high dword of dividend
	cmp	eax, ecx
	jae	short long	; high dword of dividend >= divisor?

	; perform normal division
normal:
	mov	edx, eax	; edx = high dword of dividend
	mov	eax, [esp+4]	; edx:eax = dividend
	div	ecx		; eax = (low dword of) quotient,
				; edx = (low dword of) remainder
	xor	edx, edx	; edx:eax = quotient

	ret	16		; callee restores stack

	; perform "long" alias "schoolbook" division
long:
;;	xor	edx, edx	; edx:eax = high dword of dividend
	div	ecx		; eax = high dword of quotient,
				; edx = high dword of remainder'
	push	eax		; [esp] = high dword of quotient
	mov	eax, [esp+8]	; eax = low dword of dividend
	div	ecx		; eax = low dword of quotient,
				; edx = (low dword of) remainder
	pop	edx		; edx:eax = quotient

	ret	16		; callee restores stack

	; dividend < divisor: quotient = 0
trivial:
	xor	eax, eax
	xor	edx, edx	; edx:eax = quotient = 0

	ret	16		; callee restores stack

	; dividend >= divisor >= 2**32: quotient < 2**32
extended:
	xor	ecx, 31		; ecx = number of leading '0' bits
				;        in (high dword of) divisor
	jz	short special	; divisor >= 2**63?

	; perform "extended & adjusted" division

	shld	edx, eax, cl	; edx = divisor / 2**(index + 1)
				;     = divisor'
;;	shl	eax, cl
	push	ebx
	mov	ebx, edx	; ebx = divisor'
ifndef JCCLESS
	xor	eax, eax	; eax = high dword of quotient' = 0
	mov	edx, [esp+12]	; edx = high dword of dividend
	cmp	edx, ebx
	jb	short @f	; high dword of dividend < divisor'?

	; high dword of dividend >= divisor':
	; subtract divisor' from high dword of dividend to prevent possible
	; division overflow and set most significant bit of quotient"

	sub	edx, ebx	; edx = high dword of dividend - divisor'
				;     = high dword of dividend'
	inc	eax		; eax = high dword of quotient' = 1
@@:
	push	eax		; [esp] = high dword of quotient'
else ; JCCLESS
	mov	edx, [esp+12]	; edx = high dword of dividend
	cmp	edx, ebx	; CF = (high dword of dividend < divisor')
	sbb	eax, eax	; eax = (high dword of dividend < divisor') ? -1 : 0
	inc	eax		; eax = (high dword of dividend < divisor') ? 0 : 1
				;     = high dword of quotient'
	push	eax		; [esp] = high dword of quotient'
if 0
	neg	eax		; eax = (high dword of dividend < divisor') ? 0 : -1
	and	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
else
	imul	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
endif
	sub	edx, eax	; edx = high dword of dividend
				;     - (high dword of dividend < divisor') ? 0 : divisor'
				;     = high dword of dividend'
endif ; JCCLESS
	mov	eax, [esp+12]	; eax = low dword of dividend
				;     = low dword of dividend'
	div	ebx		; eax = dividend' / divisor'
				;     = low dword of quotient',
				; edx = remainder'
	pop	ebx		; ebx = high dword of quotient'
	shld	ebx, eax, cl	; ebx = quotient' / 2**(index + 1)
				;     = dividend / divisor'
				;     = quotient"
;;	shl	eax, cl
	mov	eax, [esp+16]	; eax = low dword of divisor
	mul	ebx		; edx:eax = low dword of divisor * quotient"
ifndef JCCLESS
	mov	ecx, [esp+20]	; ecx = high dword of divisor
	imul	ecx, ebx	; ecx = high dword of divisor * quotient"
	add	edx, ecx	; edx:eax = divisor * quotient"
	jc	short @f	; divisor * quotient" >= 2**64?

	mov	ecx, [esp+12]	; ecx = high dword of dividend
	cmp	[esp+8], eax
	sbb	ecx, edx	; CF = (dividend < divisor * quotient")
				;    = (remainder" < 0)
@@:
	sbb	eax, eax	; eax = (remainder" < 0) ? -1 : 0
	add	eax, ebx	; eax = quotient" - (remainder" < 0)
				;     = (low dword of) quotient
	xor	edx, edx	; edx:eax = quotient
else ; JCCLESS
	mov	ecx, [esp+12]	; ecx = high dword of dividend
	cmp	[esp+8], eax
	sbb	ecx, edx	; ecx:... = dividend
				;         - low dword of divisor * quotient"
	mov	eax, [esp+20]	; eax = high dword of divisor
	imul	eax, ebx	; eax = high dword of divisor * quotient"
if 0
	sub	ecx, eax	; ecx:... = dividend - divisor * quotient"
				;         = remainder"
	sbb	eax, eax	; eax = (remainder" < 0) ? -1 : 0
	add	eax, ebx	; eax = quotient" - (remainder" < 0)
				;     = (low dword of) quotient
	xor	edx, edx	; edx:eax = quotient
else
	xor	edx, edx	; edx = high dword of quotient = 0
	sub	ecx, eax	; ecx:... = dividend - divisor * quotient"
				;         = remainder"
	mov	eax, ebx	; eax = quotient"
	sbb	eax, edx	; eax = quotient" - (remainder" < 0)
				;     = (low dword of) quotient
endif
endif ; JCCLESS
	pop	ebx
	ret	16		; callee restores stack

	; dividend >= divisor >= 2**63: quotient = 1
special:
	xor	eax, eax
	xor	edx, edx
	inc	eax		; edx:eax = quotient = 1

	ret	16		; callee restores stack

_aulldiv endp

; MSC internal _aulldvrm():
; receives arguments on stack, returns quotient in edx:eax and remainder in ebx:ecx

; NOTE: _aulldvrm() can raise 'division by zero' exception!

_aulldvrm proc	public		; qword _aulldvrm(qword dividend, qword divisor)

	mov	ecx, [esp+8]	; ecx = high dword of dividend
	mov	eax, [esp+12]
	mov	edx, [esp+16]	; edx:eax = divisor
	cmp	[esp+4], eax
	sbb	ecx, edx
	jb	short trivial	; dividend < divisor?

	bsr	ecx, edx	; ecx = index of most significant '1' bit
				;        in high dword of divisor
	jnz	short extended	; high dword of divisor <> 0?

	; remainder < divisor < 2**32

	mov	ecx, eax	; ecx = (low dword of) divisor
	mov	eax, [esp+8]	; eax = high dword of dividend
	cmp	eax, ecx
	jae	short long	; high dword of dividend >= divisor?

	; perform normal division
normal:
	mov	edx, eax	; edx = high dword of dividend
	mov	eax, [esp+4]	; edx:eax = dividend
	div	ecx		; eax = (low dword of) quotient,
				; edx = (low dword of) remainder
	mov	ecx, edx	; ecx = (low dword of) remainder
	xor	ebx, ebx	; ebx:ecx = remainder
	xor	edx, edx	; edx:eax = quotient

	ret	16		; callee restores stack

	; perform "long" alias "schoolbook" division
long:
;;	xor	edx, edx	; edx:eax = high dword of dividend
	div	ecx		; eax = high dword of quotient,
				; edx = high dword of remainder'
	mov	ebx, eax	; ebx = high dword of quotient
	mov	eax, [esp+4]	; eax = low dword of dividend
	div	ecx		; eax = low dword of quotient,
				; edx = (low dword of) remainder
if 1
	mov	ecx, edx	; ecx = (low dword of) remainder
	mov	edx, ebx	; edx:eax = quotient
	xor	ebx, ebx	; ebx:ecx = remainder
else
	xor	ecx, ecx
	xchg	edx, ebx	; edx:eax = quotient
	xchg	ebx, ecx	; ebx:ecx = remainder
endif
	ret	16		; callee restores stack

	; dividend < divisor: quotient = 0, remainder = dividend
trivial:
	mov	ecx, [esp+4]
	mov	ebx, [esp+8]	; ebx:ecx = remainder = dividend
	xor	eax, eax
	xor	edx, edx	; edx:eax = quotient = 0

	ret	16		; callee restores stack

	; dividend >= divisor >= 2**32: quotient < 2**32
extended:
	xor	ecx, 31		; ecx = number of leading '0' bits
				;        in (high dword of) divisor
	jz	short special	; divisor >= 2**63?

	; perform "extended & adjusted" division

	shld	edx, eax, cl	; edx = divisor / 2**(index + 1)
				;     = divisor'
;;	shl	eax, cl
	mov	ebx, edx	; ebx = divisor'
ifndef JCCLESS
	xor	eax, eax	; eax = high dword of quotient' = 0
	mov	edx, [esp+8]	; edx = high dword of dividend
	cmp	edx, ebx
	jb	short @f	; high dword of dividend < divisor'?

	; high dword of dividend >= divisor':
	; subtract divisor' from high dword of dividend to prevent possible
	; division overflow and set most significant bit of quotient"

	sub	edx, ebx	; edx = high dword of dividend - divisor'
				;     = high dword of dividend'
	inc	eax		; eax = high dword of quotient' = 1
@@:
	push	eax		; [esp] = high dword of quotient'
else ; JCCLESS
	mov	edx, [esp+8]	; edx = high dword of dividend
	cmp	edx, ebx	; CF = (high dword of dividend < divisor')
	sbb	eax, eax	; eax = (high dword of dividend < divisor') ? -1 : 0
	inc	eax		; eax = (high dword of dividend < divisor') ? 0 : 1
				;     = high dword of quotient'
	push	eax		; [esp] = high dword of quotient'
if 0
	neg	eax		; eax = (high dword of dividend < divisor') ? 0 : -1
	and	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
else
	imul	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
endif
	sub	edx, eax	; edx = high dword of dividend
				;     - (high dword of dividend < divisor') ? 0 : divisor'
				;     = high dword of dividend'
endif ; JCCLESS
	mov	eax, [esp+8]	; eax = low dword of dividend
				;     = low dword of dividend'
	div	ebx		; eax = dividend' / divisor'
				;     = low dword of quotient',
				; edx = remainder'
	pop	ebx		; ebx = high dword of quotient'
	shld	ebx, eax, cl	; ebx = quotient' / 2**(index + 1)
				;     = dividend / divisor'
				;     = quotient"
;;	shl	eax, cl
	mov	eax, [esp+12]	; eax = low dword of divisor
	mul	ebx		; edx:eax = low dword of divisor * quotient"
	mov	ecx, [esp+16]	; ecx = high dword of divisor
	imul	ecx, ebx	; ecx = high dword of divisor * quotient"
	push	ebx		; [esp] = quotient"
	mov	ebx, [esp+12]	; ebx = high dword of dividend
	sub	ebx, ecx	; ebx = high dword of dividend
				;     - high dword of divisor * quotient"
	mov	ecx, [esp+8]	; ecx = low dword of dividend
	sub	ecx, eax
	sbb	ebx, edx	; ebx:ecx = dividend - divisor * quotient"
				;         = remainder"
ifndef JCCLESS
	pop	eax		; eax = quotient"
	jnb	short @f	; remainder" >= 0?
				;  (with borrow, it is off by divisor,
				;   and quotient" is off by 1)
	add	ecx, [esp+12]
	adc	ebx, [esp+16]	; ebx:ecx = remainder" + divisor
				;         = remainder
	dec	eax		; eax = quotient" - 1
				;     = (low dword of) quotient
@@:
else ; JCCLESS
	sbb	eax, eax	; eax = (remainder" < 0) ? -1 : 0
	cdq			; edx = (remainder" < 0) ? -1 : 0
	add	[esp], eax	; [esp] = quotient" - (remainder" < 0)
				;       = (low dword of) quotient
	and	eax, [esp+16]
	and	edx, [esp+20]	; edx:eax = (remainder" < 0) ? divisor : 0
	add	ecx, eax
	adc	ebx, edx	; ebx:ecx = remainder" + divisor
				;         = remainder
	pop	eax		; eax = (low dword of) quotient
endif ; JCCLESS
	xor	edx, edx	; edx:eax = quotient

	ret	16		; callee restores stack

	; dividend >= divisor >= 2**63:
	; quotient = 1, remainder = dividend - divisor
special:
	mov	ecx, [esp+4]
	mov	ebx, [esp+8]	; ebx:ecx = dividend
	sub	ecx, eax
	sbb	ebx, edx	; ebx:ecx = dividend - divisor
				;         = remainder
	xor	eax, eax
	xor	edx, edx
	inc	eax		; edx:eax = quotient = 1

	ret	16		; callee restores stack

_aulldvrm endp

; MSC internal _aullrem():
; receives arguments on stack, returns remainder in edx:eax

; NOTE: _aullrem() can raise 'division by zero' exception!

_aullrem proc	public		; qword _aullrem(qword dividend, qword divisor)

	mov	ecx, [esp+8]	; ecx = high dword of dividend
	mov	eax, [esp+12]
	mov	edx, [esp+16]	; edx:eax = divisor
	cmp	[esp+4], eax
	sbb	ecx, edx
	jb	short trivial	; dividend < divisor?

	bsr	ecx, edx	; ecx = index of most significant '1' bit
				;        in high dword of divisor
	jnz	short extended	; high dword of divisor <> 0?

	; remainder < divisor < 2**32

	mov	ecx, eax	; ecx = (low dword of) divisor
	mov	eax, [esp+8]	; eax = high dword of dividend
	cmp	eax, ecx
	jae	short long	; high dword of dividend >= divisor?

	; perform normal division
normal:
	mov	edx, eax	; edx = high dword of dividend
	mov	eax, [esp+4]	; edx:eax = dividend
	div	ecx		; eax = (low dword of) quotient,
				; edx = (low dword of) remainder
	mov	eax, edx	; eax = (low dword of) remainder
	xor	edx, edx	; edx:eax = remainder

	ret	16		; callee restores stack

	; perform "long" alias "schoolbook" division
long:
;;	xor	edx, edx	; edx:eax = high dword of dividend
	div	ecx		; eax = high dword of quotient,
				; edx = high dword of remainder'
	mov	eax, [esp+4]	; eax = low dword of dividend
	div	ecx		; eax = low dword of quotient,
				; edx = (low dword of) remainder
	mov	eax, edx	; eax = (low dword of) remainder
	xor	edx, edx	; edx:eax = remainder

	ret	16		; callee restores stack

	; dividend < divisor: remainder = dividend
trivial:
	mov	eax, [esp+4]
	mov	edx, [esp+8]	; edx:eax = remainder = dividend

	ret	16		; callee restores stack

	; dividend >= divisor >= 2**32: quotient < 2**32
extended:
	xor	ecx, 31		; ecx = number of leading '0' bits
				;        in (high dword of) divisor
	jz	short special	; divisor >= 2**63?

	; perform "extended & adjusted" division

	shld	edx, eax, cl	; edx = divisor / 2**(index + 1)
				;     = divisor'
;;	shl	eax, cl
	push	ebx
	mov	ebx, edx	; ebx = divisor'
ifndef JCCLESS
	xor	eax, eax	; eax = high dword of quotient' = 0
	mov	edx, [esp+12]	; edx = high dword of dividend
	cmp	edx, ebx
	jb	short @f	; high dword of dividend < divisor'?

	; high dword of dividend >= divisor':
	; subtract divisor' from high dword of dividend to prevent possible
	; division overflow and set most significant bit of quotient"

	sub	edx, ebx	; edx = high dword of dividend - divisor'
				;     = high dword of dividend'
	inc	eax		; eax = high dword of quotient' = 1
@@:
	push	eax		; [esp] = high dword of quotient'
else ; JCCLESS
	mov	edx, [esp+12]	; edx = high dword of dividend
	cmp	edx, ebx	; CF = (high dword of dividend < divisor')
	sbb	eax, eax	; eax = (high dword of dividend < divisor') ? -1 : 0
	inc	eax		; eax = (high dword of dividend < divisor') ? 0 : 1
				;     = high dword of quotient'
	push	eax		; [esp] = high dword of quotient'
if 0
	neg	eax		; eax = (high dword of dividend < divisor') ? 0 : -1
	and	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
else
	imul	eax, ebx	; eax = (high dword of dividend < divisor') ? 0 : divisor'
endif
	sub	edx, eax	; edx = high dword of dividend
				;     - (high dword of dividend < divisor') ? 0 : divisor'
				;     = high dword of dividend'
endif ; JCCLESS
	mov	eax, [esp+12]	; eax = low dword of dividend
				;     = low dword of dividend'
	div	ebx		; eax = dividend' / divisor'
				;     = low dword of quotient',
				; edx = remainder'
	pop	ebx		; ebx = high dword of quotient'
	shld	ebx, eax, cl	; ebx = quotient' / 2**(index + 1)
				;     = dividend / divisor'
				;     = quotient"
;;	shl	eax, cl
	mov	eax, [esp+16]	; eax = low dword of divisor
	mul	ebx		; edx:eax = low dword of divisor * quotient"
	imul	ebx, [esp+20]	; ebx = high dword of divisor * quotient"
	mov	ecx, [esp+12]	; ecx = high dword of dividend
	sub	ecx, ebx	; ecx = high dword of dividend
				;     - high dword of divisor * quotient"
	mov	ebx, [esp+8]	; ebx = low dword of dividend
	sub	ebx, eax
	sbb	ecx, edx	; ecx:ebx = dividend - divisor * quotient"
				;         = remainder"
ifndef JCCLESS
	jnb	short @f	; remainder" >= 0?
				;  (with borrow, it is off by divisor,
				;   and quotient" is off by 1)
	add	ebx, [esp+16]
	adc	ecx, [esp+20]	; ecx:ebx = remainder" + divisor
				;         = remainder
@@:
	mov	eax, ebx
	mov	edx, ecx	; edx:eax = remainder
else ; JCCLESS
	sbb	eax, eax	; eax = (remainder" < 0) ? -1 : 0
	cdq			; edx = (remainder" < 0) ? -1 : 0
	and	eax, [esp+16]
	and	edx, [esp+20]	; edx:eax = (remainder" < 0) ? divisor : 0
	add	eax, ebx
	adc	edx, ecx	; edx:eax = remainder" + divisor
				;         = remainder
endif ; JCCLESS
	pop	ebx
	ret	16		; callee restores stack

	; dividend >= divisor >= 2**63: remainder = dividend - divisor
special:
if 0
	mov	eax, [esp+4]
	mov	edx, [esp+8]	; edx:eax = dividend
	sub	eax, [esp+12]
	sbb	edx, [esp+16]	; edx:eax = dividend - divisor
				;         = remainder
else
	neg	edx
	neg	eax
	sbb	edx, ecx	; edx:eax = -divisor
	add	eax, [esp+4]
	adc	edx, [esp+8]	; edx:eax = dividend - divisor
				;         = remainder
endif
	ret	16		; callee restores stack

_aullrem endp

; MSC internal _aullshr():
; receives arguments in edx:eax and cl, returns result in edx:eax

; NOTE: _aullshr() applies shift count modulo 64

_aullshr proc	public		; qword _aullshr(qword value, byte count)

	test	cl, 32
	jnz	short @f	; count > 31?

	shrd	eax, edx, cl
	shr	edx, cl
	ret
@@:
	mov	eax, edx
	shr	eax, cl
	xor	edx, edx
	ret

_aullshr endp
	end
<<NOKEEP

I386-MEM.OBJ: I386-MSC.MAK
	ML.EXE /Brepro /Cp /Cx /c /coff /Fo$@ /Gy /nologo /safeseh /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

	.386
	.model	flat, C
	.code

; https://msdn.microsoft.com/en-us/library/d7zdhf37.aspx

memchr	proc	public		; void *memchr(void const *destination, int character, size_t count)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of destination
	mov	eax, [esp+8]	; eax = character
	mov	ecx, [esp+12]	; ecx = count
	repne	scasb
	dec	edi		; edi = address of character
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = character)
	sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
	and	eax, edi	; eax = (ecx = 0) ? 0 : address of character
	mov	edi, edx
	ret

memchr	endp

; https://msdn.microsoft.com/en-us/library/zyaebf12.aspx

memcmp	proc	public		; int memcmp(void *source, void *destination, size_t count)

	mov	eax, [esp+4]	; eax = address of source
	mov	edx, [esp+8]	; edx = address of destination
	cmp	edx, eax
	je	short equal	; address of destination = address of source?

	mov	ecx, [esp+12]	; ecx = count
if 0
	jecxz	short equal	; count = 0?
else
	cmp	ebx, ebx	; CF = 0, ZF = 1 (required when count is 0)
endif
	xchg	esi, eax	; esi = address of source
	xchg	edi, edx	; edi = address of destination
	repe	cmpsb
	mov	edi, edx
	mov	esi, eax
	seta	al
	movzx	eax, al		; eax = (*source > *destination)
	sbb	eax, 0		; eax = (*source > *destination)
				;     - (*source < *destination)
				;     = {1, 0, -1}
	ret
equal:
	xor	eax, eax
	ret

memcmp	endp

; https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

memcpy	proc	public		; void *memcpy(void *destination, void const *source, size_t count)

	mov	eax, [esp+4]	; eax = address of destination
	mov	edx, [esp+8]	; edx = address of source
	cmp	edx, eax
	je	short return	; address of source = address of destination?

	mov	ecx, [esp+12]	; ecx = count
;;	jecxz	short return	; count = 0?

	xchg	esi, edx	; esi = address of source
	xchg	edi, eax	; edi = address of destination
	rep	movsb
	mov	esi, edx
	mov	edi, eax
	mov	eax, [esp+4]	; eax = address of destination
return:
	ret

memcpy	endp

memmem	proc	public		; void *memmem(void const *haystack, size_t count,
				;              void const *needle, size_t length)

	xor	eax, eax	; eax = address of needle in haystack = 0
	mov	ecx, [esp+8]	; ecx = length of haystack
	test	ecx, ecx
	jz	short empty	; length of haystack = 0?

	mov	edx, [esp+16]	; edx = length of needle
	cmp	edx, ecx
	ja	short empty	; length of needle > length of haystack?

	mov	eax, [esp+4]	; eax = address of haystack
	test	edx, edx
	jz	short empty	; length of needle = 0?

	push	ebx
	push	edi
	mov	edi, eax	; edi = address of haystack
	push	esi
search:
	mov	esi, [esp+24]	; esi = address of needle
	mov	al, [esi]	; al = first character of needle
				; edi = current address in haystack
	repne	scasb		; edi = next address in haystack,
				; ecx = current length of haystack
	jne	short break	; (first character of) needle not in haystack?

	dec	ecx		; ecx = next length of haystack

	mov	al, [esi+edx-1]	; al = last character of needle
	cmp	al, [edi+edx-2]
	jne	short continue	; last character of needle not in haystack?
compare:
	mov	eax, edi	; eax = next address in haystack
	mov	ebx, ecx	; ebx = next length of haystack
if 0
	dec	edi		; edi = current address in haystack
				;     = address of matching character
				; esi = address of needle
	mov	ecx, edx	; ecx = length of needle
else
				; edi = next address in haystack
	inc	esi		; esi = address of needle + 1
	mov	ecx, edx
	dec	ecx		; ecx = length of needle - 1,
				; ZF = (ecx = 0)
;;	jz	short match	; needle in haystack?
endif
	repe	cmpsb
	je	short match	; needle in haystack?

	mov	edi, eax	; edi = current address in haystack
	mov	ecx, ebx	; ecx = current length of haystack
continue:
	cmp	ecx, edx
	jae	short search	; length of haystack >= length of needle?
break:
	xor	eax, eax
	pop	esi
	pop	edi
	pop	ebx
empty:
	ret
match:
	dec	eax		; eax = address of needle in haystack
	pop	esi
	pop	edi
	pop	ebx
	ret

memmem	endp

; https://msdn.microsoft.com/en-us/library/8k35d1fx.aspx

memmove	proc	public		; void *memmove(void *destination, void const *source, size_t count)

	mov	eax, [esp+4]	; eax = address of destination
	mov	edx, [esp+8]	; edx = address of source
	cmp	edx, eax
	je	short return	; address of source = address of destination?

	mov	ecx, [esp+12]	; ecx = count
;;	jecxz	short return	; count = 0?

	xchg	esi, edx	; esi = address of source
	xchg	edi, eax	; edi = address of destination

	ja	short default	; address of source > address of destination?
overlap:
	lea	edi, [edi+ecx-1]
	lea	esi, [esi+ecx-1]
	std
default:
	rep	movsb
	cld
	mov	esi, edx
	mov	edi, eax
	mov	eax, [esp+4]	; eax = address of destination
return:
	ret

memmove	endp

memrchr	proc	public		; void *memrchr(void const *destination, int character, size_t count)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of destination
	mov	eax, [esp+8]	; eax = character
	mov	ecx, [esp+12]	; ecx = count
	lea	edi, [edi+ecx-1]
	std
	repne	scasb
	cld
	inc	edi		; edi = address of character
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = character)
	sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
	and	eax, edi	; eax = (ecx = 0) ? 0 : address of character
	mov	edi, edx
	ret

memrchr	endp

; https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

memset	proc	public		; void *memset(void *destination, int character, size_t count)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of destination
	mov	eax, [esp+8]	; eax = character
	mov	ecx, [esp+12]	; ecx = count
;;	jecxz	short @f	; count = 0?

	rep	stosb
@@:
	mov	eax, [esp+4]	; eax = address of destination
	mov	edi, edx
	ret

memset	endp

; https://msdn.microsoft.com/en-us/library/d7zdhf37.aspx

wmemchr	proc	public		; wchar_t *wmemchr(wchar_t *destination, wchar_t character, size_t count)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of destination
	mov	eax, [esp+8]	; eax = wide character
	mov	ecx, [esp+12]	; ecx = count
	repne	scasw
	dec	edi
	dec	edi		; edi = address of wide character
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = wide character)
	sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
	and	eax, edi	; eax = (ecx = 0) ? 0 : address of wide character
	mov	edi, edx
	ret

wmemchr	endp

; https://msdn.microsoft.com/en-us/library/zyaebf12.aspx

wmemcmp	proc	public		; int wmemcmp(wchar_t *source, wchar_t *destination, size_t count)

	mov	eax, [esp+4]	; eax = address of source
	mov	edx, [esp+8]	; edx = address of destination
	cmp	edx, eax
	je	short equal	; address of destination = address of source?

	mov	ecx, [esp+12]	; ecx = count
if 0
	jecxz	short equal	; count = 0?
else
	cmp	ebx, ebx	; CF = 0, ZF = 1 (required when count is 0)
endif
	xchg	esi, eax	; esi = address of source
	xchg	edi, edx	; edi = address of destination
	repe	cmpsw
	mov	edi, edx
	mov	esi, eax
	seta	al
	movzx	eax, al		; eax = (*source > *destination)
	sbb	eax, 0		; eax = (*source > *destination)
				;     - (*source < *destination)
				;     = {1, 0, -1}
	ret
equal:
	xor	eax, eax
	ret

wmemcmp	endp

; https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

wmemcpy	proc	public		; wchar_t *wmemcpy(wchar_t *destination, wchar_t const *source, size_t count)

	mov	eax, [esp+4]	; eax = address of destination
	mov	edx, [esp+8]	; edx = address of source
	cmp	edx, eax
	je	short return	; address of source = address of destination?

	mov	ecx, [esp+12]	; ecx = count
;;	jecxz	short return	; count = 0?

	xchg	esi, edx	; esi = address of source
	xchg	edi, eax	; edi = address of destination
	rep	movsw
	mov	esi, edx
	mov	edi, eax
	mov	eax, [esp+4]	; eax = address of destination
return:
	ret

wmemcpy	endp

wmemmem	proc	public		; wchar_t *wmemmem(wchar_t const *haystack, size_t count,
				;                  wchar_t const *needle, size_t length)

	xor	eax, eax	; eax = address of needle in haystack = 0
	mov	ecx, [esp+8]	; ecx = length of haystack
	test	ecx, ecx
	jz	short empty	; length of haystack = 0?

	mov	edx, [esp+16]	; edx = length of needle
	cmp	edx, ecx
	jb	short empty	; length of needle > length of haystack?

	mov	eax, [esp+4]	; eax = address of haystack
	test	edx, edx
	jz	short empty	; length of needle = 0?

	push	ebx
	push	edi
	mov	edi, eax	; edi = address of haystack
	push	esi
search:
	mov	esi, [esp+24]	; esi = address of needle
	mov	ax, [esi]	; ax = first wide character of needle
				; edi = current address in haystack
	repne	scasw		; edi = next address in haystack,
				; ecx = current length of haystack
	jne	short break	; (first wide character of) needle not in haystack?

	dec	ecx		; ecx = next length of haystack

	mov	ax, [esi+edx*2-2]
				; ax = last wide character of needle
	cmp	ax, [edi+edx*2-4]
	jne	short continue	; last wide character of needle not in haystack?
compare:
	mov	eax, edi	; eax = next address in haystack
	mov	ebx, ecx	; ebx = next length of haystack
if 0
	dec	edi
	dec	edi		; edi = current address in haystack
				;     = address of matching wide character
				; esi = address of needle
	mov	ecx, edx	; ecx = length of needle
else
				; edi = next address in haystack
	inc	esi
	inc	esi		; esi = address of needle + 2
	mov	ecx, edx
	dec	ecx		; ecx = length of needle - 1,
				; ZF = (ecx = 0)
;;	jz	short match	; needle in haystack?
endif
	repe	cmpsw
	je	short match	; needle in haystack?

	mov	edi, eax	; edi = current address in haystack
	mov	ecx, ebx	; ecx = current length of haystack
continue:
	cmp	ecx, edx
	jae	short search	; length of haystack >= length of needle?
break:
	xor	eax, eax
	pop	esi
	pop	edi
	pop	ebx
empty:
	ret
match:
	dec	eax
	dec	eax		; eax = address of needle in haystack
	pop	esi
	pop	edi
	pop	ebx
	ret

wmemmem	endp

; https://msdn.microsoft.com/en-us/library/8k35d1fx.aspx

wmemmove proc	public		; wchar_t *wmemmove(wchar_t *destination, wchar_t const *source, size_t count)

	mov	eax, [esp+4]	; eax = address of destination
	mov	edx, [esp+8]	; edx = address of source
	cmp	edx, eax
	je	short return	; address of source = address of destination?

	mov	ecx, [esp+12]	; ecx = count
;;	jecxz	short return	; count = 0?

	xchg	esi, edx	; esi = address of source
	xchg	edi, eax	; edi = address of destination

	ja	short default	; address of source > address of destination?
overlap:
	lea	edi, [edi+ecx*2-2]
	lea	esi, [esi+ecx*2-2]
	std
default:
	rep	movsw
	cld
	mov	esi, edx
	mov	edi, eax
	mov	eax, [esp+4]	; eax = address of destination
return:
	ret

wmemmove endp

wmemrchr proc	public		; wchar_t *wmemrchr(wchar_t *destination, wchar_t character, size_t count)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of destination
	mov	eax, [esp+8]	; eax = wide character
	mov	ecx, [esp+12]	; ecx = count
	lea	edi, [edi+ecx*2-2]
	std
	repne	scasw
	cld
	inc	edi
	inc	edi		; edi = addrsss of wide character
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = wide character)
	sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
	and	eax, edi	; eax = (ecx = 0) ? 0 : address of wide character
	mov	edi, edx
	ret

wmemrchr endp

; https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

wmemset	proc	public		; wchar_t *wmemset(wchar_t *destination, wchar_t character, size_t count)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of destination
	mov	eax, [esp+8]	; eax = wide character
	mov	ecx, [esp+12]	; ecx = count
;;	jecxz	short @f	; count = 0?

	rep	stosw
@@:
	mov	eax, [esp+4]	; eax = address of destination
	mov	edi, edx
	ret

wmemset	endp
	end
<<NOKEEP

I386-RTC.OBJ: I386-MSC.MAK

I386-STK.OBJ: I386-MSC.MAK
	ML.EXE /Brepro /Cp /Cx /c /coff /Fo$@ /Gy /nologo /safeseh /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

	.386
	.model	flat, C

; https://msdn.microsoft.com/en-us/library/dn449700.aspx
; https://msdn.microsoft.com/en-us/library/ms253960.aspx

_nt_tib	struct	4		; thread information block
chain	dword	?		; address of first exception registration record
base	dword	?		; stack base
limit	dword	?		; stack limit
	dword	?		; address of subsystem thread information block
fiber	dword	?		; fiber data
pointer	dword	?		; arbitrary user pointer
self	dword	?		; address of _nt_tib
_nt_tib	ends

	.code

; MSC internal intrinsic _alloca() alias _chkstk():
; receives argument in eax, returns result in esp

; https://msdn.microsoft.com/en-us/library/wb1s57t5.aspx
; https://msdn.microsoft.com/en-us/library/ms648426.aspx

; NOTE: _alloca() must preserve all registers except eax;
;       it can raise 'stack overflow' exception!

;;	alias	<_chkstk> = <_alloca_probe>

_alloca_probe proc	public	; void _alloca_probe(size_t)
_chkstk	proc	public		; void _chkstk(size_t)

	push	ebx		; decrement stack pointer, save register
	lea	ebx, [esp+8]	; ebx = stack pointer of caller
	sub	ebx, eax	; ebx = new (unaligned) stack pointer
if 0
	sbb	eax, eax	; eax = (ebx < 0) ? -1 : 0
	not	eax		; eax = (ebx < 0) ? 0 : -1
else
	cmc			; CF = ~(ebx < 0)
	sbb	eax, eax	; eax = (ebx < 0) ? 0 : -1
endif
	and	eax, ebx	; eax = (ebx < 0) ? 0 : new (aligned) stack pointer

	assume	fs :flat
	mov	ebx, fs:[_nt_tib.limit]
				; ebx = (current) stack limit
	cmp	ebx, eax
	jna	short ready	; stack limit not above new stack pointer?
probe:
	sub	ebx, 4096	; ebx = next stack page
	test	ebx, [ebx]	; probe next stack page, eventually raise
				; 'guard page' or 'stack overflow' exception
	cmp	ebx, eax
	ja	short probe	; stack limit above new stack pointer?
ready:
	pop	ebx		; restore register
	xchg	eax, esp	; esp = new stack pointer,
				; eax = old stack pointer
				;     = address of return address
	push	[eax]		; decrement stack pointer, write return address
	ret

_chkstk	endp
_alloca_probe endp

_alloca_probe_8 proc	public	; void _alloca_probe_8(size_t)

	push	ebx		; decrement stack pointer, save register
	lea	ebx, [esp+8]	; ebx = stack pointer of caller
	sub	ebx, eax	; ebx = new (unaligned) stack pointer
if 0
	sbb	eax, eax	; eax = (ebx < 0) ? -1 : 0
	not	eax		; eax = (ebx < 0) ? 0 : -1
else
	cmc			; CF = ~(ebx < 0)
	sbb	eax, eax	; eax = (ebx < 0) ? 0 : -1
endif
	shl	eax, 3		; eax = (ebx < 0) ? 0 : -8
	and	eax, ebx	; eax = (ebx < 0) ? 0 : new (aligned) stack pointer

	assume	fs :flat
	mov	ebx, fs:[_nt_tib.limit]
				; ebx = (current) stack limit
	cmp	ebx, eax
	jna	short ready	; stack limit not above new stack pointer?
probe:
	sub	ebx, 4096	; ebx = next stack page
	test	ebx, [ebx]	; probe next stack page, eventually raise
				; 'guard page' or 'stack overflow' exception
	cmp	ebx, eax
	ja	short probe	; stack limit above new stack pointer?
ready:
	pop	ebx		; restore register
	xchg	eax, esp	; esp = new stack pointer,
				; eax = old stack pointer
				;     = address of return address
	push	[eax]		; decrement stack pointer, write return address
	ret

_alloca_probe_8 endp

_alloca_probe_16 proc	public	; void _alloca_probe_16(size_t)

	push	ebx		; decrement stack pointer, save register
	lea	ebx, [esp+8]	; ebx = stack pointer of caller
	sub	ebx, eax	; ebx = new (unaligned) stack pointer
if 0
	sbb	eax, eax	; eax = (ebx < 0) ? -1 : 0
	not	eax		; eax = (ebx < 0) ? 0 : -1
else
	cmc			; CF = ~(ebx < 0)
	sbb	eax, eax	; eax = (ebx < 0) ? 0 : -1
endif
	shl	eax, 4		; eax = (ebx < 0) ? 0 : -16
	and	eax, ebx	; eax = (ebx < 0) ? 0 : new (aligned) stack pointer

	assume	fs :flat
	mov	ebx, fs:[_nt_tib.limit]
				; ebx = (current) stack limit
	cmp	ebx, eax
	jna	short ready	; stack limit not above new stack pointer?
probe:
	sub	ebx, 4096	; ebx = next stack page
	test	ebx, [ebx]	; probe next stack page, eventually raise
				; 'guard page' or 'stack overflow' exception
	cmp	ebx, eax
	ja	short probe	; stack limit above new stack pointer?
ready:
	pop	ebx		; restore register
	xchg	eax, esp	; esp = new stack pointer,
				; eax = old stack pointer
				;     = address of return address
	push	[eax]		; decrement stack pointer, write return address
	ret

_alloca_probe_16 endp
	end
<<NOKEEP

I386-STR.OBJ: I386-MSC.MAK
	ML.EXE /Brepro /Cp /Cx /c /coff /Fo$@ /Gy /nologo /safeseh /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; NOTE: characters are unsigned!

	.386
	.model	flat, C
	.code

; https://msdn.microsoft.com/en-us/library/h1x0y282.aspx

strcat	proc	public		; char *strcat(char *destination, char const *source)

	mov	edx, edi
	mov	edi, [esp+8]	; edi = address of source string
	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	not	ecx		; ecx = length of source string (including '\0')
	push	ecx
	mov	edi, [esp+8]	; edi = address of destination string
;;	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	dec	edi		; edi = address of '\0'
				;     = end of destination string
	mov	eax, esi
	mov	esi, [esp+12]	; esi = address of source string
	pop	ecx		; ecx = length of source string (including '\0')
	rep	movsb
	mov	edi, edx
	mov	esi, eax
	mov	eax, [esp+4]	; eax = address of destination string
	ret

strcat	endp

; https://msdn.microsoft.com/en-us/library/b34ccac3.aspx

strchr	proc	public		; char *strchr(char const *string, int character)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	not	ecx		; ecx = length of string (including '\0')
	sub	edi, ecx	; edi = address of string
	mov	eax, [esp+8]	; eax = character
	repne	scasb
	dec	edi		; edi = address of character
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = character)
	sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
	and	eax, edi	; eax = (ecx = 0) ? 0 : address of character
	mov	edi, edx
	ret

strchr	endp

; https://msdn.microsoft.com/en-us/library/e0z9k731.aspx

strcmp	proc	public		; int strcmp(char const *source, char const *destination)

	push	esi
	push	edi
	xor	eax, eax	; eax = 0
	mov	esi, [esp+12]	; esi = address of source string
	mov	edi, [esp+16]	; edi = address of destination string
	cmp	edi, esi
	je	short equal	; address of destination string = address of source string?

;;	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	not	ecx		; ecx = length of destination string (including '\0')
	sub	edi, ecx	; edi = address of destination string
;;	xor	eax, eax	; eax = 0
	repe	cmpsb
	seta	al		; eax = (*source > *destination)
	sbb	eax, 0		; eax = (*source > *destination)
				;     - (*source < *destination)
				;     = {1, 0, -1}
equal:
	pop	edi
	pop	esi
	ret

strcmp	endp

; https://msdn.microsoft.com/en-us/library/a7cwbx4t.aspx

; NOTE: strcoll() is another implementation of strcmp()!

strcoll	proc	public		; int strcoll(char const *source, char const *destination)

	mov	ecx, [esp+4]	; ecx = address of source string
	mov	edx, [esp+8]	; edx = address of destination string
	sub	edx, ecx
	jz	short equal	; address of destination string = address of source string?
compare:
	mov	al, [ecx]
	cmp	al, [ecx+edx]
	jne	short different

	inc	ecx
	test	al, al
	jnz	short compare	; *source <> '\0'?
equal:
	xor	eax, eax	; eax = 0
	ret
different:
	sbb	eax, eax	; eax = (*source < *destination) ? -1 : 0
	or	eax, 1		; eax = (*source < *destination)
				;     - (*source > *destination)
				;     = {-1, 0, 1}
	ret

strcoll	endp

; https://msdn.microsoft.com/en-us/library/kk6xf663.aspx

strcpy	proc	public		; char *strcpy(char *destination, char const *source)

	mov	edx, edi
	mov	edi, [esp+8]	; edi = address of source string
	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	not	ecx		; ecx = length of source string (including '\0')
	sub	edi, ecx	; edi = address of source string
	mov	eax, esi
	mov	esi, edi	; esi = address of source string
	mov	edi, [esp+4]	; edi = address of destination string
	rep	movsb
	mov	edi, edx
	mov	esi, eax
	mov	eax, [esp+4]	; eax = address of destination string
	ret

strcpy	endp

; https://msdn.microsoft.com/en-us/library/xe8sk0x7.aspx

strcspn	proc	public		; size_t strcspn(char const *string, char const *delimiter)

	mov	eax, [esp+4]	; eax = address of string
	mov	edx, [esp+8]	; edx = address of delimiter
	xor	ecx, ecx	; ecx = 0
	cmp	cl, [edx]
	je	short empty	; delimiter[0] = '\0'?

	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx		; bitmap[0..255] = 0,
				; esp = address of bitmap
setup:
	bts	[esp], ecx	; bitmap[ecx] = 1
	mov	cl, [edx]	; ecx = *delimiter
	inc	edx		; edx = ++delimiter
	cmp	cl, ch
	jne	short setup	; ecx <> '\0'?

	mov	edx, eax	; edx = address of string
skip:
	mov	cl, [eax]	; ecx = *string
	inc	eax		; eax = ++string
	bt	[esp], ecx
	jnc	short skip	; bitmap[ecx] = 0 (no match)?
stop:
	sbb	eax, edx	; eax = number of non-matching characters
	add	esp, 32
	ret
empty:
	mov	edx, eax	; edx = address of string
count:
	inc	eax		; eax = ++string
	cmp	cl, [eax-1]
	jne	short count
if 1
	dec	eax		; eax = --string
	sub	eax, edx
else
	stc
	sbb	eax, edx	; eax = number of characters
endif
	ret

strcspn	endp

; https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

strlen	proc	public		; size_t strlen(char const *string)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb		; ecx = -1 - (address of '\0' + 1 - address of string)
				;     = -1 - (length of string + 1)
				;     = -2 - length of string
if 0
	mov	eax, -2
	sub	eax, ecx	; eax = -2 + 2 + length of string
				;     = length of string
else
	mov	eax, ecx	; eax = -1 - (length of string + 1)
	not	eax		; eax = length of string + 1
	dec	eax		; eax = length of string
endif
	mov	edi, edx
	ret

strlen	endp

; https://msdn.microsoft.com/en-us/library/tbyd7s1y.aspx

strncat	proc	public		; char *strncat(char *destination, char const *source, size_t count)

	push	esi
	push	edi
	mov	esi, [esp+16]	; esi = address of source string
	mov	edx, [esp+20]	; edx = count
	mov	edi, esi	; edi = address of source string
	mov	ecx, edx	; ecx = count
	xor	eax, eax	; eax = '\0'
	repne	scasb
	sub	edx, ecx	; edx = length of source string (including '\0')
	mov	edi, [esp+12]	; edi = address of destination string
;;	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	dec	edi		; edi = address of '\0'
				;     = end of destination string
	mov	ecx, edx	; ecx = length of source string (including '\0')
	rep	movsb
;;	xor	eax, eax	; eax = '\0'
	stosb
	mov	eax, [esp+12]	; eax = address of destination string
	pop	edi
	pop	esi
	ret

strncat	endp

; https://msdn.microsoft.com/en-us/library/tbyd7s1y.aspx

strncmp	proc	public		; int strncmp(char const *source, char const *destination, size_t count)

	push	esi
	push	edi
	xor	eax, eax	; eax = 0
	mov	esi, [esp+12]	; esi = address of source string
	mov	edi, [esp+16]	; edi = address of destination string
	cmp	edi, esi
	je	short equal	; address of destination string = address of source string?

	mov	ecx, [esp+20]	; ecx = count
	test	ecx, ecx
	jz	short equal	; count = 0?

;;	xor	eax, eax	; eax = 0,
;;				; CF = 0, ZF = 1 (required when count is 0)
	repe	cmpsb
	seta	al		; eax = (*source > *destination)
	sbb	eax, 0		; eax = (*source > *destination)
				;     - (*source < *destination)
				;     = {1, 0, -1}
equal:
	pop	edi
	pop	esi
	ret

strncmp	endp

; https://msdn.microsoft.com/en-us/library/xdsywd25.aspx

strncpy	proc	public		; char *strncpy(char *destination, char const *source, size_t count)

	push	esi
	push	edi
	mov	esi, [esp+16]	; esi = address of source string
	mov	edx, [esp+20]	; edx = count
	mov	edi, esi	; edi = address of source string
	mov	ecx, edx	; ecx = count
	xor	eax, eax	; eax = '\0'
	repne	scasb
	sub	ecx, edx
	neg	ecx		; ecx = length of source string (including '\0')
	sub	edx, ecx	; edx = count - length of source string (including '\0')
	mov	edi, [esp+12]	; edi = address of destination string
	rep	movsb
	mov	ecx, edx	; ecx = count - length of source string (including '\0')
;;	xor	eax, eax	; eax = '\0'
	rep	stosb
	mov	eax, [esp+12]	; eax = address of destination string
	pop	edi
	pop	esi
	ret

strncpy	endp

; https://msdn.microsoft.com/en-us/library/z50ty2zh.aspx

strnlen	proc	public		; size_t strnlen(char const *string, size_t count)

	mov	ecx, [esp+8]	; ecx = count
	test	ecx, ecx
	jz	short empty	; count = 0?

	xor	eax, eax	; eax = '\0'
	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
;;	test	edi, edi	; ZF = 0 (required when count is 0)
	repne	scasb		; ecx = (length of string < count)
				;     ? count - (length of string + 1) : 0
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = '\0')
				;    = (length of string < count),
				; ecx = (length of string < count)
				;     ? length of string + 1 - count : 0
	sbb	ecx, eax	; ecx = (length of string < count)
				;     ? length of string - count : 0
	add	ecx, [esp+8]	; ecx = (length of string < count)
				;     ? length of string : count
	mov	edi, edx
empty:
	mov	eax, ecx	; eax = (length of string < count)
				;     ? length of string : count
	ret

strnlen	endp

; https://msdn.microsoft.com/en-us/library/ms235495.aspx

strnset	proc	public		; char *strnset(char *string, int character, size_t count)

	mov	edx, [esp+4]	; edx = address of string
	mov	ecx, [esp+12]	; ecx = count
	test	ecx, ecx
	jz	short zero	; count = 0?

	xor	eax, eax	; eax = '\0'
	push	edi
	mov	edi, edx	; edi = address of string
;;	test	edi, edi	; ZF = 0 (required when count is 0)
	repne	scasb		; ecx = (length of string < count)
				;     ? count - (length of string + 1) : 0
	mov	edi, edx	; edi = address of string
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = '\0')
				;    = (length of string < count),
				; ecx = (length of string < count)
				;     ? length of string + 1 - count : 0
	sbb	ecx, eax	; ecx = (length of string < count)
				;     ? length of string - count : 0
	add	ecx, [esp+16]	; ecx = (length of string < count)
				;     ? length of string : count
	mov	eax, [esp+12]	; eax = character
	rep	stosb
	pop	edi
zero:
	mov	eax, edx	; eax = address of string
	ret

strnset	endp

; https://msdn.microsoft.com/en-us/library/217yyhy9.aspx

strpbrk	proc	public		; char *strpbrk(char const *string, char const *delimiter)

	mov	eax, [esp+4]	; eax = address of string
	mov	edx, [esp+8]	; edx = address of delimiter
	xor	ecx, ecx	; ecx = 0
	cmp	cl, [edx]
	je	short empty	; delimiter[0] = '\0'?

	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx		; bitmap[0..255] = 0,
				; esp = address of bitmap
setup:
	bts	[esp], ecx	; bitmap[ecx] = 1
	mov	cl, [edx]	; ecx = *delimiter
	inc	edx		; edx = ++delimiter
	cmp	cl, ch
	jne	short setup	; ecx <> '\0'?
skip:
	mov	cl, [eax]	; ecx = *string
	inc	eax		; eax = ++string
	bt	[esp], ecx
	jnc	short skip	; bitmap[ecx] = 0 (no match)?
stop:
	dec	eax		; eax = --string
	neg	ecx
	sbb	ecx, ecx	; ecx = (*string = '\0') ? 0 : -1
	and	eax, ecx	; eax = (*string = '\0') ? 0 : address of string
	add	esp, 32
	ret
empty:
	xor	eax, eax
	ret

strpbrk	endp

; https://msdn.microsoft.com/en-us/library/ftw0heb9.aspx

strrchr	proc	public		; char *strrchr(char const *string, int character)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	not	ecx		; ecx = length of string (including '\0')
	dec	edi		; edi = address of '\0'
				;     = end of string
	mov	eax, [esp+8]	; eax = character
	std
	repne	scasb
	cld
	inc	edi		; edi = address of character
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = character)
	sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
	and	eax, edi	; eax = (ecx = 0) ? 0 : address of character
	mov	edi, edx
	ret

strrchr	endp

; https://msdn.microsoft.com/en-us/library/ms235404.aspx

strrev	proc	public		; char *strrev(char *string)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	add	ecx, edi	; ecx = address of string - 1
	dec	edi		; edi = address of '\0'
				;     = end of string
	jmp	short continue
reverse:
	mov	al, [ecx]
	mov	ah, [edi]
	mov	[ecx], ah
	mov	[edi], al
continue:
	inc	ecx
	dec	edi
	cmp	edi, ecx
	ja	short reverse

	mov	edi, edx
	mov	eax, [esp+4]	; eax = address of string
	ret

strrev	endp

; https://msdn.microsoft.com/en-us/library/ms235359.aspx

strset	proc	public		; char *strset(char *string, int character)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	not	ecx		; ecx = length of string (including '\0')
	sub	edi, ecx	; edi = address of string
	dec	ecx		; ecx = length of string
	mov	eax, [esp+8]	; eax = character
	rep	stosb
	mov	edi, edx
	mov	eax, [esp+4]	; eax = address of string
	ret

strset	endp

; https://msdn.microsoft.com/en-us/library/kthzzay9.aspx

strspn	proc	public		; size_t strspn(char const *string, char const *delimiter)

	mov	eax, [esp+4]	; eax = address of string
	mov	edx, [esp+8]	; edx = address of delimiter
	xor	ecx, ecx	; ecx = 0
	cmp	cl, [edx]
	je	short empty	; delimiter[0] = '\0'?

	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx		; bitmap[0..255] = 0,
				; esp = address of bitmap
	mov	cl, [edx]	; ecx = *delimiter
	inc	edx		; edx = ++delimiter
setup:
	bts	[esp], ecx	; bitmap[ecx] = 1
	mov	cl, [edx]	; ecx = *delimiter
	inc	edx		; edx = ++delimiter
	cmp	cl, ch
	jne	short setup	; ecx <> '\0'?

	mov	edx, eax	; edx = address of string
skip:
	mov	cl, [eax]	; ecx = *string
	inc	eax		; eax = ++string
	bt	[esp], ecx
	jc	short skip	; bitmap[ecx] = 1 (match)?
if 1
	dec	eax		; eax = --string
	sub	eax, edx
else
	stc
	sbb	eax, edx	; eax = number of matching characters
endif
	add	esp, 32
	ret
empty:
	xor	eax, eax
	ret

strspn	endp

; https://msdn.microsoft.com/en-us/library/z9da80kz.aspx

strstr	proc	public		; char *strstr(char const *haystack, char const *needle)

	push	edi
	mov	edi, [esp+12]	; edi = address of needle
	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	not	ecx		; ecx = length of needle (including '\0')
	dec	ecx		; ecx = length of needle
	mov	eax, [esp+8]	; eax = address of haystack
	jz	short empty	; length of needle = 0?

	mov	edx, ecx	; edx = length of needle
ifdef SIMPLE
	push	esi
compare:
	mov	esi, eax	; esi = current address in haystack
	mov	edi, [esp+16]	; edi = address of needle
	mov	ecx, edx	; ecx = length of needle
	repe	cmpsb
	je	short match	; needle in haystack?

	inc	eax		; eax = next address in haystack
	cmp	byte ptr [esi-1], 0
	jne	short compare	; non-matching character in haystack <> '\0'?

	xor	eax, eax
match:
else ; SIMPLE
	mov	edi, eax	; edi = address of haystack
	xor	eax, eax	; eax = '\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasb
	not	ecx		; ecx = length of haystack (including '\0')
	sub	edi, ecx	; edi = address of haystack
	dec	ecx		; ecx = length of haystack
	jz	short empty	; length of haystack = 0?

	cmp	ecx, edx
	jb	short empty	; length of haystack < length of needle?

	push	esi
	push	ebx
search:
	mov	esi, [esp+20]	; esi = address of needle
	mov	al, [esi]	; al = first character of needle
				; edi = current address in haystack
	repne	scasb		; edi = next address in haystack,
				; ecx = current length of haystack
	jne	short break	; (first character of) needle not in haystack?

	dec	ecx		; ecx = next length of haystack

	mov	al, [esi+edx-1]	; al = last character of needle
	cmp	al, [edi+edx-2]
	jne	short continue	; last character of needle not in haystack?
compare:
	mov	eax, edi	; eax = next address in haystack
	mov	ebx, ecx	; ebx = next length of haystack
if 0
	dec	edi		; edi = current address in haystack
				;     = address of matching character
				; esi = address of needle
	mov	ecx, edx	; ecx = length of needle
else
				; edi = next address in haystack
	inc	esi		; esi = address of needle + 1
	mov	ecx, edx
	dec	ecx		; ecx = length of needle - 1,
				; ZF = (ecx = 0)
;;	jz	short match	; needle in haystack?
endif
	repe	cmpsb
	je	short match	; needle in haystack?

	mov	edi, eax	; edi = current address in haystack
	mov	ecx, ebx	; ecx = current length of haystack
continue:
	cmp	ecx, edx
	jae	short search	; length of haystack >= length of needle?
break:
	xor	eax, eax
	pop	ebx
	pop	esi
	pop	edi
	ret
match:
	dec	eax		; eax = address of needle in haystack
	pop	ebx
endif ; SIMPLE
	pop	esi
empty:
	pop	edi
	ret

strstr	endp

strtok_r proc	public		; char *strtok_r(char *string, char const *delimiter, char **next)

	mov	ecx, [esp+4]	; ecx = address of string
	mov	eax, [esp+8]	; eax = address of delimiter
	mov	edx, [esp+12]	; edx = address of address of next
	test	ecx, ecx
	jnz	short start	; address of string <> 0?

	or	ecx, [edx]	; ecx = address of next
	jz	short null	; address of next = 0 = address of string?
start:
	cmp	byte ptr [ecx], 0
	je	short null	; string[0] = '\0'?

	cmp	byte ptr [eax], 0
	je	short empty	; delimiter[0] = '\0'?

	push	ebx
	xor	ebx, ebx	; ebx = 0
	push	ebx
	push	ebx
	push	ebx
	push	ebx
	push	ebx
	push	ebx
	push	ebx
	push	ebx		; bitmap[0..255] = 0,
				; esp = address of bitmap
	mov	bl, [eax]	; ebx = *delimiter
	inc	eax		; eax = ++delimiter
setup:
	bts	[esp], ebx	; bitmap[ebx] = 1
	mov	bl, [eax]	; ebx = *delimiter
	inc	eax		; eax = ++delimiter
	cmp	bl, bh
	jne	short setup	; ebx <> '\0'?
skip:
	mov	bl, [ecx]	; ebx = *string
	inc	ecx		; ecx = ++string
	bt	[esp], ebx
	jc	short skip	; bitmap[ebx] = 1 (ebx is a delimiter)?

	cmp	bl, bh
	je	short none	; ebx = '\0'?

	mov	bl, bh		; ebx = 0
	bts	[esp], ebx	; bitmap['\0'] = 1
	mov	eax, ecx
	dec	eax		; eax = address of token
token:
	mov	bl, [ecx]	; ebx = *string
	inc	ecx		; ecx = ++string
	bt	[esp], ebx
	jnc	short token	; bitmap[ebx] = 0 (ebx is not a delimiter)?

	cmp	bl, bh
	je	short last	; ebx = '\0'?

	mov	[ecx-1], bh	; string[-1] = '\0' (terminate token)
	mov	[edx], ecx	; *next = address of string
	add	esp, 32
	pop	ebx
	ret
none:
	mov	eax, ebx	; eax = 0
last:
	mov	[edx], ebx	; *next = 0
	add	esp, 32
	pop	ebx
	ret
null:
	xor	eax, eax	; eax = 0
	mov	[edx], eax	; *next = 0
	ret
empty:
	mov	eax, ecx	; eax = address of string
	xor	ecx, ecx
	mov	[edx], ecx	; *next = 0
	ret

strtok_r endp
	end
<<NOKEEP

I386-TLS.OBJ: I386-MSC.MAK
	ML.EXE /Brepro /Cp /Cx /c /coff /Fo$@ /Gy /nologo /safeseh /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; https://msdn.microsoft.com/en-us/library/ms680547.aspx

	.model	flat, C

	public	_tls_array
_tls_array equ	44		; offset of 'ThreadLocalStoragePointer' member in TEB;
				;  symbol referenced in code generated by the compiler!
_tls_32	struct	4
	dword	offset _tls_begin
	dword	offset _tls_end
	dword	offset _tls_index
	dword	offset _tls_start
	dword	'VOID'		; BUG: the module loader does NOT support the 'SizeOfZeroFill' member!
	dword	0
_tls_32	ends

_tls_bss segment alias(".bss$$T") dword read write 'BSS'
	public	_tls_index
_tls_index dword ?		; assigned by the module loader!
_tls_bss ends

_tls_note segment alias(".comment") discard info read 'INFO'
	byte	"(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
_tls_note ends

; https://msdn.microsoft.com/en-us/library/wxz26dz2.aspx
; https://msdn.microsoft.com/en-us/library/sf9b18xk.aspx

_tls_info segment alias(".drectve") discard info read 'INFO'
	byte	"/MERGE:.ptr=.rdata /SECTION:.tls,!w"
_tls_info ends

_tls_start segment alias(".ptr") dword read 'CONST'
_tls_start ends

_tls_stop segment alias(".ptr$$~~~") dword read 'CONST'
	dword	0		; callback function array terminator
_tls_stop ends

_tls	segment alias(".rdata$$T") dword read 'CONST'
	public	_tls_used
_tls_used _tls_32 <>		; IMAGE_TLS_DIRECTORY32
_tls	ends

_tls_begin segment alias(".tls") para read write 'DATA'
_tls_begin ends

_tls_end segment alias(".tls$$~~~") byte read write 'DATA'
_tls_end ends
	end
<<NOKEEP

I386-WCS.OBJ: I386-MSC.MAK
	ML.EXE /Brepro /Cp /Cx /c /coff /Fo$@ /Gy /nologo /safeseh /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; NOTE: counts and lengths are numbers of wide characters, not bytes!

	.686
	.model	flat, C
	.code

; https://msdn.microsoft.com/en-us/library/h1x0y282.aspx

wcscat	proc	public		; wchar_t *wcscat(wchar_t *destination, wchar_t const *source)

	mov	edx, edi
	mov	edi, [esp+8]	; edi = address of source string
	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	not	ecx		; ecx = length of source string (including L'\0')
	push	ecx
	mov	edi, [esp+8]	; edi = address of destination string
;;	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	dec	edi
	dec	edi		; edi = address of L'\0'
				;     = end of destination string
	mov	eax, esi
	mov	esi, [esp+12]	; esi = address of source string
	pop	ecx		; ecx = length of source string (including L'\0')
	rep	movsw
	mov	edi, edx
	mov	esi, eax
	mov	eax, [esp+4]	; eax = address of destination string
	ret

wcscat	endp

; https://msdn.microsoft.com/en-us/library/b34ccac3.aspx

wcschr	proc	public		; wchar_t *wcschr(wchar_t const *string, wchar_t character)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	not	ecx		; ecx = length of string (including L'\0')
	sub	edi, ecx
	sub	edi, ecx	; edi = address of string
	mov	eax, [esp+8]	; eax = wide character
	repne	scasw
	dec	edi
	dec	edi		; edi = address of wide character
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = wide character)
	sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
	and	eax, edi	; eax = (ecx = 0) ? 0 : address of wide character
	mov	edi, edx
	ret

wcschr	endp

; https://msdn.microsoft.com/en-us/library/e0z9k731.aspx

wcscmp	proc	public		; int wcscmp(wchar_t const *source, wchar_t const *destination)

	push	esi
	push	edi
	xor	eax, eax	; eax = 0
	mov	esi, [esp+12]	; esi = address of source string
	mov	edi, [esp+16]	; edi = address of destination string
	cmp	edi, esi
	je	short equal	; address of destination string = address of source string?

;;	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	not	ecx		; ecx = length of destination string (including L'\0')
	sub	edi, ecx
	sub	edi, ecx	; edi = address of destination string
;;	xor	eax, eax	; eax = 0
	repe	cmpsw
	seta	al		; eax = (*source > *destination)
	sbb	eax, 0		; eax = (*source > *destination)
				;     - (*source < *destination)
				;     = {1, 0, -1}
equal:
	pop	edi
	pop	esi
	ret

wcscmp	endp

; https://msdn.microsoft.com/en-us/library/a7cwbx4t.aspx

; NOTE: wcscoll() is another implementation of wcscmp()!

wcscoll	proc	public		; int wcscoll(wchar_t const *source, wchar_t const *destination)

	mov	ecx, [esp+4]	; ecx = address of source string
	mov	edx, [esp+8]	; edx = address of destination string
	sub	edx, ecx
	jz	short equal	; address of destination string = address of source string?
compare:
	mov	ax, [ecx]
	cmp	ax, [ecx+edx]
	jne	short different

	inc	ecx
	inc	ecx
	test	ax, ax
	jnz	short compare	; *source <> L'\0'?
equal:
	xor	eax, eax	; eax = 0
	ret
different:
	sbb	eax, eax	; eax = (*source < *destination) ? -1 : 0
	or	eax, 1		; eax = (*source < *destination)
				;     - (*source < *destination)
				;     = {-1, 0, 1}
	ret

wcscoll	endp

; https://msdn.microsoft.com/en-us/library/kk6xf663.aspx

wcscpy	proc	public		; wchar_t *wcscpy(wchar_t *destination, wchar_t const *source)

	mov	edx, edi
	mov	edi, [esp+8]	; edi = address of source string
	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	not	ecx		; ecx = length of source string (including L'\0')
	mov	eax, esi
	mov	esi, [esp+8]	; esi = address of source string
	mov	edi, [esp+4]	; edi = address of destination string
	rep	movsw
	mov	edi, edx
	mov	esi, eax
	mov	eax, [esp+4]	; eax = address of destination string
	ret

wcscpy	endp

; https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

wcslen	proc	public		; size_t wcslen(wchar_t const *string)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw		; ecx = -1 - (address of L'\0' + 2 - address of string)
				;     = -1 - (length of string + 1)
				;     = -2 - length of string
if 0
	mov	eax, -2
	sub	eax, ecx	; eax = -2 + 2 + length of string
				;     = length of string
else
	mov	eax, ecx	; eax = -1 - (length of string + 1)
	not	eax		; eax = length of string + 1
	dec	eax		; eax = length of string
endif
	mov	edi, edx
	ret

wcslen	endp

; https://msdn.microsoft.com/en-us/library/tbyd7s1y.aspx

wcsncat	proc	public		; wchar_t *wcsncat(wchar_t *destination, wchar_t const *source, size_t count)

	push	esi
	push	edi
	mov	esi, [esp+16]	; esi = address of source string
	mov	edx, [esp+20]	; edx = count
	mov	edi, esi	; edi = address of source string
	mov	ecx, edx	; ecx = count
	xor	eax, eax	; eax = L'\0'
	repne	scasw
	sub	edx, ecx	; edx = length of source string (including L'\0')
	mov	edi, [esp+12]	; edi = address of destination string
;;	xor	eax, eax	; eax = 'L\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	dec	edi
	dec	edi		; edi = address of L'\0'
				;     = end of destination string
	mov	ecx, edx	; ecx = length of source string (including L'\0')
	rep	movsw
;;	xor	eax, eax	; eax = L'\0'
	stosw
	mov	eax, [esp+12]	; eax = address of destination string
	pop	edi
	pop	esi
	ret

wcsncat	endp

; https://msdn.microsoft.com/en-us/library/eywx8zcx.aspx

wcsncmp	proc	public		; int wcsncmp(wchar_t const *source, wchar_t const *destination, size_t count)

	push	esi
	push	edi
	xor	eax, eax	; eax = 0
	mov	esi, [esp+12]	; esi = address of source string
	mov	edi, [esp+16]	; edi = address of destination string
	cmp	edi, esi
	je	short equal	; address of destination string = address of source string?

	mov	ecx, [esp+20]	; ecx = count
	test	ecx, ecx
	jz	short equal	; count = 0?

;;	xor	eax, eax	; eax = 0,
;;				; CF = 0, ZF = 1 (required when count is 0)
	repe	cmpsw
	seta	al		; eax = (*source > *destination)
	sbb	eax, 0		; eax = (*source > *destination)
				;     - (*source < *destination)
				;     = {1, 0, -1}
equal:
	pop	edi
	pop	esi
	ret

wcsncmp	endp

; https://msdn.microsoft.com/en-us/library/xdsywd25.aspx

wcsncpy	proc	public		; wchar_t *wcsncpy(wchar_t *destination, wchar_t const *source, size_t count)

	push	esi
	push	edi
	mov	esi, [esp+16]	; esi = address of source string
	mov	edx, [esp+20]	; edx = count
	mov	edi, esi	; edi = address of source string
	mov	ecx, edx	; ecx = count
	xor	eax, eax	; eax = L'\0'
	repne	scasw
	sub	ecx, edx
	neg	ecx		; ecx = length of source string (including L'\0')
	sub	edx, ecx	; edx = count - length of source string (including L'\0')
	mov	edi, [esp+12]	; edi = address of destination string
	rep	movsw
	mov	ecx, edx	; ecx = count - length of source string (including L'\0')
;;	xor	eax, eax	; eax = L'\0'
	rep	stosw
	mov	eax, [esp+12]	; eax = address of destination string
	pop	edi
	pop	esi
	ret

wcsncpy	endp

; https://msdn.microsoft.com/en-us/library/z50ty2zh.aspx

wcsnlen	proc	public		; size_t wcsnlen(wchar_t const *string, size_t count)

	mov	ecx, [esp+8]	; ecx = count
	test	ecx, ecx
	jz	short empty	; count = 0?

	xor	eax, eax	; eax = L'\0'
	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
;;	test	edi, edi	; ZF = 0 (required when count is 0)
	repne	scasw		; ecx = (length of string < count)
				;     ? count - (length of string + 1) : 0
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = L'\0')
				;    = (length of string < count),
				; ecx = (length of string < count)
				;     ? length of string + 1 - count : 0
	sbb	ecx, eax	; ecx = (length of string < count)
				;     ? length of string - count : 0
	add	ecx, [esp+8]	; ecx = (length of string < count)
				;     ? length of string : count
	mov	edi, edx
empty:
	mov	eax, ecx	; eax = (length of string < count)
				;     ? length of string : count
	ret

wcsnlen	endp

; https://msdn.microsoft.com/en-us/library/ms235495.aspx

wcsnset	proc	public		; wchar_t *wcsnset(wchar_t *string, wchar_t character, size_t count)

	mov	edx, [esp+4]	; edx = address of string
	mov	ecx, [esp+12]	; ecx = count
	test	ecx, ecx
	jz	short zero	; count = 0?

	xor	eax, eax	; eax = L'\0'
	push	edi
	mov	edi, edx	; edi = address of string
;;	test	edi, edi	; ZF = 0 (required when count is 0)
	repne	scasw		; ecx = (length of string < count)
				;     ? count - (length of string + 1) : 0
	mov	edi, edx	; edi = address of string
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = L'\0')
				;    = (length of string < count),
				; ecx = (length of string < count)
				;     ? length of string + 1 - count : 0
	sbb	ecx, eax	; ecx = (length of string < count)
				;     ? length of string - count : 0
	add	ecx, [esp+16]	; ecx = (length of string < count)
				;     ? length of string : count
	mov	eax, [esp+12]	; eax = wide character
	rep	stosw
	pop	edi
zero:
	mov	eax, edx	; eax = address of string
	ret

wcsnset	endp

; https://msdn.microsoft.com/en-us/library/ftw0heb9.aspx

wcsrchr	proc	public		; wchar_t *wcsrchr(wchar_t const *string, wchar_t character)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	not	ecx		; ecx = length of string (including L'\0')
	dec	edi
	dec	edi		; edi = address of L'\0'
				;     = end of string
	mov	eax, [esp+8]	; eax = wide character
	std
	repne	scasw
	cld
	inc	edi
	inc	edi		; edi = address of wide character
	neg	ecx		; CF = (ecx <> 0)
				;    = ([edi] = wide character)
	sbb	eax, eax	; eax = (ecx = 0) ? 0 : -1
	and	eax, edi	; eax = (ecx = 0) ? 0 : address of wide character
	mov	edi, edx
	ret

wcsrchr	endp

; https://msdn.microsoft.com/en-us/library/ms235404.aspx

wcsrev	proc	private		; wchar_t *wcsrev(wchar_t *string)

	ud2

wcsrev	endp

; https://msdn.microsoft.com/en-us/library/ms235359.aspx

wcsset	proc	public		; wchar_t *wcsset(wchar_t *string, wchar_t character)

	mov	edx, edi
	mov	edi, [esp+4]	; edi = address of string
	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	not	ecx		; ecx = length of string (including L'\0')
	sub	edi, ecx
	sub	edi, ecx	; edi = address of string
	dec	ecx		; ecx = length of string
	mov	eax, [esp+8]	; eax = wide character
	rep	stosw
	mov	edi, edx
	mov	eax, [esp+4]	; eax = address of string
	ret

wcsset	endp

; https://msdn.microsoft.com/en-us/library/z9da80kz.aspx

wcsstr	proc	public		; wchar_t *wcsstr(wchar_t const *haystack, wchar_t const *needle)

	push	edi
	mov	edi, [esp+12]	; edi = address of needle
	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	not	ecx		; ecx = length of needle (including L'\0')
	dec	ecx		; ecx = length of needle
	mov	eax, [esp+8]	; eax = address of haystack
	jz	short empty	; length of needle = 0?

	mov	edx, ecx	; edx = length of needle
ifdef SIMPLE
	push	esi
compare:
	mov	esi, eax	; esi = current address in haystack
	mov	edi, [esp+16]	; edi = address of needle
	mov	ecx, edx	; ecx = length of needle
	repe	cmpsw
	je	short match	; needle in haystack?

	inc	eax
	inc	eax		; eax = next address in haystack
	cmp	word ptr [esi-2], 0
	jne	short compare	; non-matching wide character in haystack <> L'\0'?

	xor	eax, eax
match:
else ; SIMPLE
	mov	edi, eax	; edi = address of haystack
	xor	eax, eax	; eax = L'\0'
	xor	ecx, ecx
	dec	ecx		; ecx = -1
	repne	scasw
	not	ecx		; ecx = length of haystack (including L'\0')
	sub	edi, ecx
	sub	edi, ecx	; edi = address of haystack
	dec	ecx		; ecx = length of haystack
	jz	short empty	; length of haystack = 0?

	cmp	ecx, edx
	jb	short empty	; length of haystack < length of needle?

	push	esi
	push	ebx
search:
	mov	esi, [esp+20]	; esi = address of needle
	mov	ax, [esi]	; ax = first wide character of needle
				; edi = current address in haystack
	repne	scasw		; edi = next address in haystack,
				; ecx = current length of haystack
	jne	short break	; (first wide character of) needle not in haystack?

	dec	ecx		; ecx = next length of haystack

	mov	ax, [esi+edx*2-2]
				; ax = last wide character of needle
	cmp	ax, [edi+edx*2-4]
	jne	short continue	; last wide character of needle not in haystack?
compare:
	mov	eax, edi	; eax = next address in haystack
	mov	ebx, ecx	; ebx = next length of haystack
if 0
	dec	edi
	dec	edi		; edi = current address in haystack
				;     = address of matching wide character
				; esi = address of needle
	mov	ecx, edx	; ecx = length of needle
else
				; edi = next address in haystack
	inc	esi
	inc	esi		; esi = address of needle + 2
	mov	ecx, edx
	dec	ecx		; ecx = length of needle - 1,
				; ZF = (ecx = 0)
;;	jz	short match	; needle in haystack?
endif
	repe	cmpsw
	je	short match	; needle in haystack?

	mov	edi, eax	; edi = current address in haystack
	mov	ecx, ebx	; ecx = current length of haystack
continue:
	cmp	ecx, edx
	jae	short search	; length of haystack >= length of needle?
break:
	xor	eax, eax
	pop	ebx
	pop	esi
	pop	edi
	ret
match:
	dec	eax
	dec	eax		; eax = address of needle in haystack
	pop	ebx
endif ; SIMPLE
	pop	esi
empty:
	pop	edi
	ret

wcsstr	endp
	end
<<NOKEEP
